/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Copyright (C) 2005 Xianfeng Li
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * $Id: ss_isa.c,v 1.2 2006/06/24 08:55:05 lixianfe Exp $
 *
 ******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../common.h"
#include "../isa.h"
#include "arm_isa.h"
#include "arm.h"

/* floating point register dependence decoders */
#define DFPR(N)			(((N)+18+1))

/* miscellaneous register dependence decoders */
#define DPSR			(0+26+1)
#define DFPSR			(1+26+1)


extern isa_t	*isa;
extern int	num_isa;

extern enum md_opcode_arm md_mask2op_arm[MD_MAX_MASK+1];

// BEGIN: from machine.h in SimpleScalar 3.0
// -----------------------------------------------------------------------------



#if 0

char *arm_md_op2name[OP_MAX] = {
  NULL, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3) NAME,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NAME,
#define CONNECT(OP)
#include "machine.def"
};



/* enum md_opcode_arm -> opcode flags, used by simulators */
unsigned int arm_md_op2flags[OP_MAX] = {
  NA, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3) FLAGS,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NA,
#define CONNECT(OP)
#include "machine.def"
};

#endif

/* enum md_opcode_arm -> description string */
char *arm_md_op2name[OP_MAX] = {
  NULL, /* DNA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) NAME,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3, I1,I2,I3,I4) NAME,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NAME,
#define CONNECT(OP)
#include "arm.def"
};

/* enum md_opcode_arm -> opcode flags, used by simulators */
unsigned int arm_md_op2flags[OP_MAX] = {
  DNA, /* DNA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) FLAGS,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) FLAGS,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) DNA,
#define CONNECT(OP)
#include "arm.def"
};

//extern char *arm_md_op2name[];
//extern unsigned int arm_md_op2flags[];



// -----------------------------------------------------------------------------
// END: from machine.h in SimpleScalar 3.0


// initiate SimpleScalar ISA info by reading its machine.def file
void
arm_init_isa()
{
    int	    i, len;

    num_isa = OP_MAX - 1;

    isa = (isa_t *) calloc(OP_MAX - 1, sizeof(isa_t));
    CHECK_MEM(isa);
    for (i = 1; i < OP_MAX - 1; i++) {
		isa[i].opcode = i;
		if (arm_md_op2name[i] != NULL) {
			len = strlen(arm_md_op2name[i]) + 1;
			isa[i].name = (char *) malloc(len);
			CHECK_MEM(isa[i].name);
			strcpy(isa[i].name, arm_md_op2name[i]);
		}
		
		// ignore non-instruction entry in DEF file
		if (arm_md_op2flags[i] == NA)
			continue;
		
		//classify instruction type
		if (arm_md_op2flags[i] & F_ICOMP)
			isa[i].type = INST_ICOMP;
		else if (arm_md_op2flags[i] & F_FCOMP)
			isa[i].type = INST_FCOMP;
		else if (arm_md_op2flags[i] & F_LOAD)
			isa[i].type = INST_LOAD;
		else if (arm_md_op2flags[i] & F_STORE)
			isa[i].type = INST_STORE;
		else if ((arm_md_op2flags[i] & (F_CTRL|F_DIRJMP))
			== (F_CTRL|F_DIRJMP)) 
			//only temporary! ARM ISA do not have explicit conditional instruction
			//unconditional branch will be checked when decoding individual instruction
			isa[i].type = INST_COND;
		else if (arm_md_op2flags[i] & F_CALL)
			isa[i].type = INST_CALL;
		else if ((arm_md_op2flags[i] & (F_CTRL|F_UNCOND|F_INDIRJMP))
			== (F_CTRL|F_UNCOND|F_INDIRJMP))
			isa[i].type = INST_RET;
		else if (arm_md_op2flags[i] & F_TRAP)
			isa[i].type = INST_TRAP;
		else {
			fprintf(stderr, "%s: unidentified instruction type!\n", isa[i].name);
			exit(1);
		}
    }
}

// decode the raw instr into decoded instr
void
arm_decode_inst(de_inst_t *de_inst, unsigned int/*md_inst_t*/ inst, int op)
{
    int	    type, offset, i, incr;
    int	    in[4], out[3];
    int	    num_in = 0, num_out = 0;
	char* inst_format;

	//cleekee: temporarily ignore ARM's .word assembly directive
	if (op == 0)
		return;

    de_inst->op_enum = op;
    de_inst->size = sizeof(unsigned int/*md_inst_t*/);
    de_inst->cond_field = inst>>28;
	
	//for branch instruction, check condition code to identify unconditional branch
	if (isa[op].type == INST_COND)
		type = (de_inst->cond_field == 14)? INST_UNCOND : INST_COND;
	else
		type = isa[op].type;
	
	/* pop {... , pc} should be return instruction */
	if (type == INST_STORE) {
		if (((inst << 4) >> 19) == 0b1000101111011)
			type = INST_RET;
	}	
	
	de_inst->type = type;

	/*if (inst == 0x812fff1e)
		printf("GUARR!!\n");*/
		
	/* Get the instruction format */
	inst_format = MD_OP_FORMAT(op);

#if 0	 
    /* sudiptac ::: Set the immediate field. Needed for 
	 * address analysis */
	 while(*inst_format) {
		  switch(*inst_format)
		  {
			  case 'o':
			  case 'i':
#ifdef _DEBUG
				printf("Immediate value = %d\n", IMM);
#endif
				de_inst->imm = IMM;
				break;
			  case 'H':
#ifdef _DEBUG
				printf("Immediate value = %d\n", SHAMT);
#endif
				de_inst->imm = SHAMT;
				break;
			  case 'u':
#ifdef _DEBUG
				printf("Immediate value = %u\n", UIMM);
#endif
				de_inst->imm = UIMM;
			   break;
			  case 'U':
#ifdef _DEBUG
				printf("Immediate value = %d\n", UIMM);
#endif
				de_inst->imm = UIMM;
				break;
			  default:
				/* Do nothing */	 
			   ;
			   /* sudiptac: Default code may need to be 
				 * modified */
		  }
		  inst_format++;
	  }	  
#endif    
	// if inst is a ctr transfer, compute the target addr
    if (type == INST_COND || type == INST_UNCOND || type == INST_CALL) {
		  offset = (inst & 0xffffff) << 2;
		  offset = ((offset >> 23) == 0)? offset : (offset|0xff000000);
		  de_inst->target = de_inst->addr + 2*sizeof(unsigned int/*md_inst_t*/) + offset;
    }

    // decode the input/output info
    switch(op) {
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) \
    case OP: \
	     in[0] = I1; in[1] = I2; in[2] = I3; in[3] = I4; \
		 out[0] = O1; out[1] = O2; out[2] = O3; \
		 break; 
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)	 \
		 ; 
#define CONNECT(OP) \
		 ; 
#include "arm.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
    default:
	     in[0] = in[1] = in[2] = in[3] = DNA;
	     out[0] = out[1] = out[2] = DNA;
    }
    incr = 0;
    for (i = 0; i <= 3; i++) {
		 if (in[i] != DNA) {
		   num_in++;
			//incr = 1;
		}	
    }
    incr = 0;
    for (i = 0; i <= 2; i++) {
		 if (out[i] != DNA) { 
		   num_out++;
			//incr = 1;
		}	
    }
	 if(!strcmp(isa[op].name, "ldf\%c") || !strcmp(isa[op].name, "ldp\%c") ||
		  !strcmp(isa[op].name, "ldr\%c"))
	 {
		  if(in[3] != DNA)
			 num_in = 3;
		  else	 
			 num_in = 2;
	 }		 
	 if(!strcmp(isa[op].name, "sb") || !strcmp(isa[op].name, "sh") ||
		  !strcmp(isa[op].name, "sw") || !strcmp(isa[op].name, "swl") ||
		  !strcmp(isa[op].name, "swr") || !strcmp(isa[op].name, "s.d") ||
		  !strcmp(isa[op].name, "s.s"))
	 {
		  if(in[2] != DNA)
			 num_in = 3;
		  else	 
			 num_in = 2;
	 }		 
    
	 de_inst->in = (int *) calloc(num_in, sizeof(int));
    CHECK_MEM(de_inst->in);
    de_inst->num_in = 0;
	 if(!strcmp(isa[op].name, "ldf\%c") || !strcmp(isa[op].name, "ldp\%c") ||
		  !strcmp(isa[op].name, "ldr\%c"))
	 {	  
		for (i = 0; i < num_in; i++) 
		  de_inst->in[de_inst->num_in++] = in[i+1];
	 }
	 else if(!strcmp(isa[op].name, "sb") || !strcmp(isa[op].name, "sh") ||
		  !strcmp(isa[op].name, "sw") || !strcmp(isa[op].name, "swl") ||
		  !strcmp(isa[op].name, "swr") || !strcmp(isa[op].name, "s.d") || 
		  !strcmp(isa[op].name, "s.s"))
	 { 	  
		for (i = 0; i < num_in; i++) 
		  de_inst->in[de_inst->num_in++] = in[i];
	 }	  
	 else
	 {
		for (i = 0; i <= 3; i++) {
		  if (in[i] != DNA)
			de_inst->in[de_inst->num_in++] = in[i];
		}
	 }

    de_inst->out = (int *) calloc(num_out, sizeof(int));
    CHECK_MEM(de_inst->out);
    de_inst->num_out = 0;
    for (i = 0; i <= 2; i++) {
			if (out[i] != DNA)
		  de_inst->out[de_inst->num_out++] = out[i];
    }
}

int
arm_inst_fu(de_inst_t *inst)
{
   // return MD_OP_FUCLASS(inst->op_enum);
		return md_op2fu_arm[inst->op_enum];
}



extern range_t fu_lat[];
int
arm_max_inst_lat(de_inst_t *inst)
{
    int	    fu;

    fu = arm_inst_fu(inst);
    return fu_lat[fu].hi;
}
