#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>  
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <libelf.h>
#include <gelf.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include "../cfg.h"
#include "../common.h"
#include "arm.h"

#define MD_OP_ENUM(MSK)		(md_mask2op_arm[MSK])

//typedef unsigned int md_addr_t;
//typedef unsigned int md_inst_t;

// variables from readfile.c
extern prog_t	    prog;
addr_t		*procs_addr; 
int		num_procs;

/* store initial stack pointer value (0x8000 is just an arbitary default value) */
int sp_init_value = 0x4000;

unsigned 	text_offset;
unsigned	text_entry;

symbol_i *interpret_flag;
int total_interpret_flag = 0;

Elf *elf;
Elf32_Ehdr *elf_header;		/* ELF header */


/* opcode mask -> enum md_opcode_arm, used by decoder (MD_OP_ENUM()) */
enum md_opcode_arm md_mask2op_arm[MD_MAX_MASK+1];
unsigned int md_opoffset[OP_MAX];

/* enum md_opcode_arm -> mask for decoding next level */
unsigned int md_opmask_arm[OP_MAX] = {
  0, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) 0,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) 0,
#define DEFLINK(OP,MSK,NAME,SHIFT,MASK) MASK,
#define CONNECT(OP)
#include "arm.def"
};

/* enum md_opcode_arm -> shift for decoding next level */
unsigned int md_opshift_arm[OP_MAX] = {
  0, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) 0,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) 0,
#define DEFLINK(OP,MSK,NAME,SHIFT,MASK) SHIFT,
#define CONNECT(OP)
#include "arm.def"
};

/* enum md_opcode_arm -> description string */
char *md_op2name_arm[OP_MAX] = {
  NULL, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) NAME,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3, I1,I2,I3,I4) NAME,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NAME,
#define CONNECT(OP)
#include "arm.def"
};

/* enum md_opcode_arm -> opcode operand format, used by disassembler */
char *md_op2format_arm[OP_MAX] = {
  NULL, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) OPFORM,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) OPFORM,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NULL,
#define CONNECT(OP)
#include "arm.def"
};

/* enum md_opcode_arm -> enum md_fu_class, used by performance simulators */
enum md_fu_class md_op2fu_arm[OP_MAX] = {
  FUClamd_NA, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) RES,
#define DEFUOP(OP,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4) RES,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) FUClamd_NA,
#define CONNECT(OP)
#include "arm.def"
};

static unsigned long
arm_md_set_decoder(char *name,
	       unsigned long mskbits, unsigned long offset,
	       enum md_opcode_arm op, unsigned long max_offset)
{
  unsigned long msk_base = mskbits & 0xff;
  unsigned long msk_bound = (mskbits >> 8) & 0xff;
  unsigned long msk;

  msk = msk_base;
  do {
    if ((msk + offset) >= MD_MAX_MASK) {
      panic("MASK_MAX is too small, inst=`%s', index=%d",
	    name, msk + offset);
    }
    // cleekee: caused by duplicated opcode in arm.def
    /*if (md_mask2op_arm[msk + offset]) {
      warn("doubly defined opcode, inst=`%s', index=%d",
	    name, msk + offset);
    }*/

    md_mask2op_arm[msk + offset] = op;
    msk++;
  } while (msk <= msk_bound);

	//printf("%i %i %i \(mask = %i, offset = %i\)\n", max_offset, (msk-1) + offset, max(max_offset, (msk-1) + offset), msk, offset);

  return max(max_offset, (msk-1) + offset);
}

/* intialize the inst decoder, this function builds the ISA decode tables */
void
arm_md_init_decoder(void)
{
  unsigned long max_offset = 0;
  unsigned long offset = 0;

#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,O3,I1,I2,I3,I4)	\
  max_offset = arm_md_set_decoder(NAME, (MSK), offset, (OP), max_offset);
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
  max_offset = arm_md_set_decoder(NAME, (MSK), offset, (OP), max_offset);
#define CONNECT(OP)							\
    offset = max_offset+1;						\
    if (md_opoffset[OP])					\
      warn("doubly defined opoffset, inst=`%s', op=%d, offset=%d",	\
	    #OP, (int)(OP), offset);					\
    md_opoffset[OP] = offset;

#include "arm.def"

  if (max_offset >= MD_MAX_MASK)
    panic("MASK_MAX is too small, index==%d", max_offset);
}



/* amount of tail padding added to all loaded text segments */
#define TEXT_TAIL_PADDING 128

/* program text (code) segment base */

#if 0
extern struct	sym_sym_t **sym_textsyms;
extern int	sym_ntextsyms;



// tell from the name of the function whether it is a library function, if so, it
// means no subsequent user functions
static int
arm_is_lib_func(char *func_name)
{
    if ((strcmp(func_name, "__do_global_dtors") == 0)
	|| (strcmp(func_name, "__libc_init") == 0))
	return 1;
    else
	return 0;
}
#endif

int check_interpret_mode(unsigned addr)
{
	int i;
	for (i=0; i<total_interpret_flag; i++) {
		if (interpret_flag[i].addr == addr) {
			/* words starting from this address is just data bytes, not ARM instruction */
			if (strcmp(interpret_flag[i].name, "$d") == 0)
				return 0;
			/* words starting from this address is to be interpreted as ARM instruction */
			else if (strcmp(interpret_flag[i].name, "$a") == 0)
				return 1;
		}
	}
	return -1;
}

// read an instruction from the object file, decode it and store the decoded instr in
// the ith entry of the decode instruction sequence
static void
arm_read_inst(int fd, addr_t addr, int i)
{
    md_inst_t	    inst;
    int		    op;

	read(fd, &inst, sizeof(inst));
	
    prog.code[i].addr = addr;
    prog.code[i].size = sizeof(md_inst_t);
    prog.code[i].value = inst;
    MD_SET_OPCODE_ARM(op, inst);
    prog.code[i].op_enum = op;
    prog.code[i].interpret_mode = check_interpret_mode(addr);
    arm_decode_inst(&prog.code[i], inst, op);

	// check for NOP (mov r0,r0) instruction
	if (inst == 0xe1a00000)
		prog.code[i].type = INST_NOP;
	
	//printf("opcode = %i , addr = %x, inst = %x\n", op, addr, inst);
}

// locate the start & end addr of the user code by excluding library functions,
// in addition, locate the main() entrance

/* HBK: modify to read more information of procs & vars */
static void
arm_lookup_addr(char *fname)
{
	int     dbg = 0;
	int		i=0, size = 0; 
	int     max_num_vars = 1024, max_num_procs = 1024;
	int text_sect_idx = 0;

	Elf_Scn *scn = NULL;            /* Section Descriptor */
	GElf_Shdr shdr;                 /* Section Header */
	Elf_Data    *data;
	int         count;
	Elf32_Word sym_sh_link;
	
	// initialize prog structure
	procs_addr = (md_addr_t *) calloc(max_num_procs, sizeof(md_addr_t));
	prog.p_info = (symbol_i *) calloc(max_num_procs,sizeof(symbol_i));
	
	// find .text section
	/* Iterate through section headers */
	while((scn = elf_nextscn(elf, scn)) != NULL) {
		// point shdr at this section header entry
		gelf_getshdr(scn, &shdr);
		
		/* found a symbol table, go print it. */
		if (shdr.sh_type == SHT_SYMTAB || strcmp(elf_strptr(elf, elf_header->e_shstrndx, shdr.sh_name), ".symtab") == 0) {
			data = elf_getdata(scn, NULL);
			count = shdr.sh_size / shdr.sh_entsize;
			sym_sh_link = shdr.sh_link;
		}

		// this will store section header table index of text section 
		i++;
		
		// the shdr name is in a string table, libelf uses elf_strptr() to find it
		// using the e_shstrndx value from the elf_header
		if (strcmp(elf_strptr(elf, elf_header->e_shstrndx, shdr.sh_name), ".text") == 0
			/*|| strcmp(elf_strptr(elf, elf_header->e_shstrndx, shdr.sh_name), "P1 ro") == 0*/) {
			prog.code_size = shdr.sh_size;
			text_sect_idx = i;
		}
	}	

	prog.start_addr = UINT_MAX;
	prog.total_procs = 0;
	interpret_flag = (symbol_i *) calloc(count, sizeof(symbol_i));
	// go through each symbol to find procedure information
	for (i = 0; i < count; i++) {
		GElf_Sym sym;
		gelf_getsym(data, i, &sym);

		char* symbol_name = elf_strptr(elf, sym_sh_link, sym.st_name);

		/* keeps track of symbols which indicate how to interpret instructions at a specific address */
		if (strcmp(symbol_name, "$d") == 0 && sym.st_shndx == text_sect_idx) {
			interpret_flag[total_interpret_flag].addr = sym.st_value;
			interpret_flag[total_interpret_flag++].name = symbol_name;
		} else if (strcmp(symbol_name, "$a") == 0 && sym.st_shndx == text_sect_idx) {
			interpret_flag[total_interpret_flag].addr = sym.st_value;
			interpret_flag[total_interpret_flag++].name = symbol_name;
		}

		if (strcmp(symbol_name, "__stack_start__") == 0) {
			sp_init_value = sym.st_value;
		}

		// magical conditions to filter out symbols except names of procedures that we need to analyze
		if (sym.st_shndx == text_sect_idx
				&& (sym.st_info == 18 || sym.st_info == 2)
				&& sym.st_size > 0
				&& (symbol_name[0]!='_'
						|| strcmp(symbol_name, "__subsf3") == 0
						|| strcmp(symbol_name, "__mulsf3") == 0
						|| strcmp(symbol_name, "__divsf3") == 0
						|| strcmp(symbol_name, "__addsf3") == 0
						|| strcmp(symbol_name, "__aeabi_i2f") == 0
						|| strcmp(symbol_name, "__aeabi_f2iz") == 0
				   )
		   ) {
			// identify starting address for first instruction
			if (sym.st_value < prog.start_addr)
				prog.start_addr = sym.st_value;
				
			// identify 'main' procedure
			if (strcmp(symbol_name, "main") == 0)
				prog.main_addr = sym.st_value;

			prog.p_info[prog.total_procs].name = symbol_name;
			prog.p_info[prog.total_procs].addr = sym.st_value;
			prog.p_info[prog.total_procs].size = sym.st_size;
			procs_addr[prog.total_procs++] = sym.st_value;
			
			//printf("symbol name = %s, addr = %x, #ins = %i, ", symbol_name, (uint)(sym.st_value), (int)(sym.st_size/4));
			//printf("st_info = %i, st_shndx = %x\n", sym.st_info, sym.st_shndx);
		}
	}
	
	prog.end_addr = prog.start_addr + prog.code_size - 4;
}

// locate & read the text header of the program
static void
arm_read_text_head(int fd)
{
	unsigned		    text_size;

	Elf_Scn *scn = NULL;            /* Section Descriptor */
	GElf_Shdr shdr;                 /* Section Header */

	//md_addr_t	ld_text_base = 0;
	/* program text (code) size in bytes */
	unsigned int	ld_text_size = 0;
		
	arm_md_init_decoder();

	/* Iterate through section headers */
	while((scn = elf_nextscn(elf, scn)) != 0)
	{
		// point shdr at this section header entry
		gelf_getshdr(scn, &shdr);

		// the shdr name is in a string table, libelf uses elf_strptr() to find it
		// using the e_shstrndx value from the elf_header
		//printf("%s\n", elf_strptr(elf, elf_header->e_shstrndx, shdr.sh_name));

		if (strcmp(elf_strptr(elf, elf_header->e_shstrndx, shdr.sh_name), ".text") == 0
		    /*|| strcmp(elf_strptr(elf, elf_header->e_shstrndx, shdr.sh_name), "P1 ro") == 0*/)
			break;
	}

	text_size = shdr.sh_size;
	text_offset = shdr.sh_offset;
	text_entry = shdr.sh_addr;
	/*ld_text_size = ((shdr.s_vaddr + shdr.s_size) - MD_TEXT_BASE) 
			+ TEXT_TAIL_PADDING;

  text_offset += prog.start_addr - text_entry;
  ld_text_base = MD_TEXT_BASE;*/
}

// read the obj code, decode the inst and store them in prog
void
arm_read_code(char *fname)
{
	int canary = 0;
	int fd;
	int		    i;
	addr_t	    addr;
	GElf_Ehdr elfhdr;

	char *base_ptr;		// ptr to our object in memory
	struct stat elf_stats;	// fstat struct

	fd = open(fname, O_RDONLY);

 	if((fstat(fd, &elf_stats)))
  {
		printf("could not fstat %s\n", fname);
				      close(fd);
		exit(-1);
  }

  if((base_ptr = (char *) malloc(elf_stats.st_size)) == NULL)
  {
		printf("could not malloc\n");
				      close(fd);
		exit(-1);
  }

  if((read(fd, base_ptr, elf_stats.st_size)) < elf_stats.st_size)
  {
		printf("could not read %s\n", fname);
				      free(base_ptr);
				      close(fd);
		exit(-1);
  }

	if (elf_version(EV_CURRENT) == EV_NONE ) {
		/* library out of date */
		fprintf(stderr, "Elf library out of date!\n");
		close(fd);
		exit(-1);
	}

	elf_header = (Elf32_Ehdr *) base_ptr;	// point elf_header at our object in memory

	//associate file with ELF object
	if ((elf = elf_begin(fd, ELF_C_READ, NULL)) == NULL) {
		/*error*/
		fprintf(stderr, "Binary not in ELF format?\n");
		close(fd);
		exit(-1);
	}  

	if (gelf_getehdr(elf, &elfhdr) == 0) {
		fprintf(stderr, "Cannot read elf header. This usually means that is not a valid elf file\n");
	}

	arm_lookup_addr(fname);

	arm_read_text_head(fd);
	addr = prog.start_addr;
	int start_offset = addr - text_entry;
	prog.num_inst = (prog.code_size-start_offset) / sizeof(md_inst_t);
	prog.code = (de_inst_t *) calloc(prog.num_inst, sizeof(de_inst_t));
	CHECK_MEM(prog.code);

	// read the text
	lseek(fd, text_offset + start_offset, SEEK_SET);
	for (i = 0; i < prog.num_inst; i++) {
		arm_read_inst(fd, addr, i);
		addr += sizeof(md_inst_t);
	}
	
	close(fd);
}
