/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Copyright (C) 2005 Xianfeng Li
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * $Id: cfg.c,v 1.2 2006/06/24 08:54:56 lixianfe Exp $
 *
 ******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "cfg.h"
#include "task.h"
#include "bitarr/bitarr.h"

enum BINARY_TARGET binaryTarget;
extern isa_t	*isa;
extern prog_t	prog;

extern int is_task_graph_given;

extern worklist_p components;
extern char is_compositional;

// lookup addr in code, return the index if found, otherwise return -1
// since instr. addresses are in sorted order, we can use binary search
static int
lookup_addr(de_inst_t *code, int num, addr_t addr)
{
    int	    hi = num - 1, lo = 0, i;

    while (hi >= lo) {
	i = (hi + lo) / 2;
	if (code[i].addr == addr)
	    return i;
	else if (code[i].addr > addr)
	    hi = i - 1;
	else
	    lo = i + 1;
    }
    //printf("Couldn't find target addr=%x in code with addr range (%x to %x)\n",
    //		addr, code[0].addr, code[num-1].addr);
    return -1;
}



// scan the decoded instr. and mark the locations in proc_entries if the
// corresponding instr. is the proc entry; it returns proc id of main proc
static int
scan_procs(int *proc_ent)
{
    int	    main_id, i, tid;
#if 1
    prog.num_procs = 0;
    for (i = 0; i < prog.num_inst; i++) {
		// check whether this instr. is the main entrance; mark it if so
		if (prog.code[i].addr == prog.main_addr) {
			if (proc_ent[i] == 0) {
				proc_ent[i] = 1;
				//printf("proc_ent[%i] = 1; addr = %x\n", i, prog.code[i].addr);
				prog.num_procs++;
				main_id = i;
			}
		} else if (inst_type(&prog.code[i]) == INST_CALL) {
			// check the target addr of a call; mark it as proc entrance if not
			// marked yet.
			tid = lookup_addr(prog.code, prog.num_inst, prog.code[i].target);
			if (tid == -1) {
			   fprintf(stderr, "no match for call: %x\n", prog.code[i].target);
			   //exit(1);
			   continue;
			}
			if (proc_ent[tid] == 0) {
				proc_ent[tid] = 1;
				//printf("proc_ent[%i] = 1; addr %x calls addr %x\n", tid, prog.code[i].addr, prog.code[tid].addr);
				prog.num_procs++;
			}
		}
    }
#else
	int j;
	
	// cleekee: mark proc_ent based on procs_addr
	for (i = 0; i < prog.num_inst; i++) {
		for (j = 0; j < prog.num_procs; j++) {
			if (prog.code[i].addr == prog.p_info[j].addr) {
				proc_ent[i] = 1;
				break;
			}
		}
			
		if (prog.code[i].addr == prog.main_addr)
			main_id = i;
	}
#endif
    return main_id;
}

/* liangyun */
extern int bjptb;

// scan the code of a proc and mark the location if it corresponds to a block entry
static void
scan_blocks(int *bb_ent, proc_t *proc)
{
    int		i, type, tid;
    de_inst_t	*inst;

    bb_ent[0] = 1;
    proc->num_bb = 1;

    addr_t adr;
    int bfind;
    int num;

    for (i = 0; i < proc->num_inst; i++) {
		inst = &proc->code[i];
		type = inst_type(inst);
			
#if 1
		if ((type == INST_COND) || (type == INST_UNCOND)) {
			// check the branch target addr, mark it as block entry if not marked yet
			tid = lookup_addr(proc->code, proc->num_inst, inst->target);

			if (tid == -1) {
				//fprintf(stderr, "[P%i] no match for branch target: %x\n", proc->id, inst->target);
				//cleekee: ignore missing branch target for now
				continue;//exit(1);
			}
			if (bb_ent[tid] == 0) {
				bb_ent[tid] = 1;
				proc->num_bb++;
			}
			// if the fall-through addr has not been marked as a block entry yet,
			// mark it as block entry
			if (((i+1) < proc->num_inst)
			    && bb_ent[i+1] == 0
			    && proc->code[i+1].interpret_mode != 0
			    && proc->code[i+1].type != INST_NOP) {
				bb_ent[i+1] = 1;
				proc->num_bb++;
			}
		} else if (type == INST_CALL) {
			// for a call instr, simply mark the next instr as block entry
			if (((i+1) < proc->num_inst)
			    && bb_ent[i+1] == 0
			    && proc->code[i+1].interpret_mode != 0
			    && proc->code[i+1].type != INST_NOP) {
				bb_ent[i+1] = 1;
				proc->num_bb++;
			}
		} else if (type == INST_RET) {
			// for a return instr, if it is not the last instr in the proc, mark the
			// next instr as block entry
			if (((i+1) < proc->num_inst)
			    && (bb_ent[i+1] == 0)
			    && proc->code[i+1].interpret_mode != 0
			    && proc->code[i+1].type != INST_NOP) {
				bb_ent[i+1] = 1;
				proc->num_bb++;                
			}
			/* liangyun: indirect jump, also the target of this jump*/
					 
			if(bjptb && i < proc->num_inst - 1){
			   bfind = lookup_jptable(inst->addr);
			   // if bind == 0, can be jr $31,succeeded by nop
			   if(bfind > 0){
				  // check the jump target addr, mark it as block entry if not marked yet
				  num = 0;

				  bb_ent[i + 1] += (bfind - 1);       // flag turn 
				  proc->num_bb  += (bfind - 1); // add dummy block                  

				  while( num < bfind ){
						get_jptable(inst->addr, num, &adr);
						tid = lookup_addr(proc->code, proc->num_inst, adr);
						if (tid == -1) {
							printf("no id for jump target: %x\n", adr);
							//exit(1);
						}
						if (bb_ent[tid] == 0) {
							bb_ent[tid] = 1;
							proc->num_bb++;
						}
						num++;
				   }
			   }
			}
		}
#endif
    }
}


// lookup a basic block in a proc with a match of its start addr to the searched addr
static cfg_node_t *
lookup_bb(cfg_node_t *cfg, int num, addr_t addr)
{
    int	    hi = num - 1, lo = 0, i;

    while (hi >= lo) {
	i = (hi + lo) / 2;
	if (cfg[i].sa == addr)
	    return &cfg[i];
	else if (cfg[i].sa > addr)
	    hi = i - 1;
	else
	    lo = i + 1;
    }
    return NULL;
}



// create a new cfg edge from src->dst, "taken" specifies whether this edge is a
// taken edge or a fall-through edge; note since the number of incoming edges of the
// dst block cannot be statically determined, the storage for its incoming edges is
// dynamically scaled.
static void
new_edge(cfg_node_t *src, cfg_node_t *dst, int taken)
{
    cfg_edge_t	*e = (cfg_edge_t *) malloc(sizeof(cfg_edge_t));
    CHECK_MEM(e);

	//cleekee: ignore duplicate edge that is already created
	if (dst->in && dst->in[0]->src == src)
		return;

    e->src = src; 
	e->dst = dst;
	e->is_backedge = 0;
	
    if (taken == NOT_TAKEN)
	src->out_n = e;
    else
	src->out_t = e;
    dst->num_in++;
    dst->in = (cfg_edge_t **) realloc(dst->in, dst->num_in * sizeof(cfg_edge_t *));
    CHECK_MEM(dst->in);
    dst->in[dst->num_in - 1] = e;
}

/* liangyun */
extern  int bjptb;

// create outgoing cfg edges for each basic block in the following way:
// - first check the type of the block's last instr
// - create fall-through edge or/and a control transfer edge according to its type
static void
create_cfg_edges(proc_t *proc)
{
    int		i, type;
    cfg_node_t	*bb, *bb1;
 
    addr_t adr;
    int    bfind;

    for (i = 0; i < proc->num_bb; i++) {
		bb = &proc->cfg[i];
		type = inst_type(&bb->code[bb->num_inst-1]);
		if (type == INST_COND) {
			// create fall-through edge
			bb1 = &proc->cfg[i+1];
			new_edge(bb, bb1, NOT_TAKEN);
			// create taken edge
			bb1 = lookup_bb(proc->cfg, proc->num_bb, bb->code[bb->num_inst-1].target);
			if (bb1 == NULL) {
				//fprintf(stderr, "[P%i] cannot find block start with addr: %x\n",
				//	proc->id, bb->code[bb->num_inst-1].target);
				//exit(1);
				//cleekee: if no target found in proc->cfg, could be a procedure call
				bb->type = CTRL_CALL;
				continue;
			}
			new_edge(bb, bb1, TAKEN);
			bb->type = CTRL_COND;
		} else if (type == INST_UNCOND) {
			// create taken edges
			bb1 = lookup_bb(proc->cfg, proc->num_bb, bb->code[bb->num_inst-1].target);
			if (bb1 == NULL) {
				//fprintf(stderr, "[P%i] cannot find block start with addr: %x\n",
				//	proc->id, bb->code[bb->num_inst-1].target);
				//exit(1);
				//cleekee: if no target found in proc->cfg, could be a procedure call
				bb->type = CTRL_CALL;
				continue;
			}
			new_edge(bb, bb1, TAKEN);
			bb->type = CTRL_UNCOND;
		} else if (type == INST_RET) {
			// do not create any edge
			bb->type = CTRL_RET;

#ifdef ARCH_ARM
			if (bb->code[bb->num_inst-1].cond_field != 14) {
				/* create fall through edge for conditional return inst */
				if (i < proc->num_bb - 1) {
					bb1 = &proc->cfg[i+1];
					new_edge(bb, bb1, NOT_TAKEN);
				}
				bb->type = CTRL_RET_COND;
			}
#endif
				
			/* liangyun */
			// if it is return jr $31, then it must be the last basic block.
			if(bjptb && i != proc->num_bb - 1 ){
			   /*
			   //create taken edges
			   bfind = lookup_jptable(bb->code[bb->num_inst-1].addr);
			   if(bfind == 0){
				   printf("cannot find target addr for %x\n",bb->code[bb->num_inst-1].addr);
				   //exit(1);
			   }
			   bb1 = lookup_bb(proc->cfg,proc->num_bb,adr);
			   if (bb1 == NULL) {
				   printf("cannot find block start with addr: %x\n", adr);
				   //exit(1);
			   }
			   if(bb1->id == bb->id + 1)
				  new_edge(bb,bb1,TAKEN);
			   else
				  new_edge(bb, bb1, TAKEN);
			   bb->type = CTRL_UNCOND; // make it unconditional
			   */
			   bfind = get_jptable_static(bb->code[bb->num_inst - 1].addr, &adr);
			   if(bfind >= 0){
				  //printf("%08x -> %08x\n",bb->code[bb->num_inst - 1].addr, adr);
				  bb1 = lookup_bb(proc->cfg,proc->num_bb,adr);
				  if (bb1 == NULL){
					printf("cannot find block start with addr: %x\n",adr);
				  }
				  new_edge(bb,bb1,TAKEN);
				  //printf("edge %d %d\n",bb->id,bb1->id);
				  if(bfind){ //not tail;
					  bb1 = &proc->cfg[i+1];
				  new_edge(bb, bb1, NOT_TAKEN);
					  //printf("edge %d %d\n",bb->id,bb1->id);
				  }
				  bb->type = CTRL_UNCOND;
			   }
			}
		} else {
			// create fall-through edge (for seqencial block and call block) if
			// current block is not the last one in the proc
			if (i < proc->num_bb - 1) {
				bb1 = &proc->cfg[i+1];
				new_edge(bb, bb1, NOT_TAKEN);
			}
			if (type == INST_CALL)
				bb->type = CTRL_CALL;
			else
				bb->type = CTRL_SEQ;
		}
    }
}


// cleekee: DEPRECATED, as proc must be ordered from low to high starting address for this to work
// lookup a proc with a match of its start addr to the searched addr
static proc_t *
lookup_proc2(addr_t addr)
{
    int	    hi = prog.num_procs - 1, lo = 0, i;

    while (hi >= lo) {
	i = (hi + lo) / 2;
	if (prog.procs[i].sa == addr)
	    return &prog.procs[i];
	else if (prog.procs[i].sa > addr)
	    hi = i - 1;
	else
	    lo = i + 1;
    }
    return NULL;
}

// lookup a proc with a match of its start addr to the searched addr
static proc_t *
lookup_proc(addr_t addr)
{
	int i;
    for (i = 0; i < prog.num_procs; i++) {
		if (prog.procs[i].sa == addr)
			return &prog.procs[i];
	}

    return NULL;
}



inline int
bb_is_loop_head(cfg_node_t *bb)
{
    return (bb->loop_role & LOOP_HEAD);
}



inline int
bb_is_loop_tail(cfg_node_t *bb)
{
    return (bb->loop_role & LOOP_TAIL);
}



static void
loop_check(proc_t *proc, int start, int end)
{
    int		is_loop;
    int		bb_id;
    cfg_node_t	*bb;
    cfg_edge_t	*e;
    Queue	worklist;

    bb = &proc->cfg[start];
    if (bb_is_loop_head(bb))
	return;
    if (start == end) {
	proc->cfg[start].loop_role = LOOP_HEAD | LOOP_TAIL;
	return;
    }

    init_queue(&worklist, sizeof(cfg_node_t *));
    enqueue(&worklist, &bb);
    bb->flags = end;
    while (!queue_empty(&worklist)) {
	bb = *((cfg_node_t **) dequeue(&worklist));
	if (bb->id == end) {
	    proc->cfg[start].loop_role = LOOP_HEAD;
	    proc->cfg[end].loop_role = LOOP_TAIL;
	    break;
	} 
	if ((bb->out_n != NULL) && (bb->out_n->dst->flags != end)) {
	    enqueue(&worklist, &bb->out_n->dst);
	    bb->out_n->dst->flags = end;
	}
	if ((bb->out_t != NULL) && (bb->out_t->dst->flags != end)) {
	    enqueue(&worklist, &bb->out_t->dst);
	    bb->out_t->dst->flags = end;
	}
    }
    free_queue(&worklist);
}



static void
identify_loops(proc_t *proc)
{
    int	i, j;
    cfg_node_t	*bb;	
    cfg_edge_t	*e;
	int dom_is_changed;
	int dom_count;
	bit *compare, *intersect;
	int N;
	int tail_nodes[128], tail_nodes_dest[128], num_tail_nodes = 0;
	
	//cleekee: DEPRECATED, replaced by dominator analysis method below
#if 0
    for (i = proc->num_bb - 1; i >= 0; i--) {
	bb = &proc->cfg[i];
	if (bb->out_t == NULL)
	    continue;
	if (bb->out_t->dst->id <= i)
	    loop_check(proc, bb->out_t->dst->id, i);
    }
	for (i = 0; i < proc->num_bb; i++)
	proc->cfg[i].flags = 0;
#endif
	
	//perform dominator analysis
	ba_init();
	N = proc->num_bb;
	bit **dom = (bit**) calloc(N, sizeof(bit*));
	for (i = 0; i < N; i++)
		dom[i] = ba_new(N);
	compare = ba_new(N);
	intersect = ba_new(N);

	//STEP 1: dominator of the start node is the start itself
	ba_assign(dom[0], 0, TRUE);

	//STEP 2: for all other nodes, set all nodes as the dominators
	for (i = 1; i < N; i++)
		ba_all_assign(dom[i], N, TRUE); 

	//STEP 3: iteratively eliminate nodes that are not dominators from the
	//        set of dominators of each node
	do {
		dom_is_changed = 0;
		
		for (i = 1; i < N; i++) {
			//save a copy of the set of dominators for node i for comparison later
			ba_copy(compare, dom[i], N);
			
			//node n dominates node i if all paths to i go through n
			//if n dominates all predecessor nodes of i, it dominates i as well
			for (j = 0; j < proc->cfg[i].num_in; j++) {
				e = proc->cfg[i].in[j];
				
				ba_intersection(dom[i], dom[e->src->id], &intersect, N, N);
				ba_copy(dom[i], intersect, N);
			}
			ba_assign(dom[i], i, TRUE);
			
			//detect any change in dominators for node i
			//by comparing dot product value, could have a simpler way...
			int dom_count = ba_count(compare, N);
			if (ba_dotprod(dom[i], compare, N, N) != dom_count*dom_count)
				dom_is_changed = 1;
		}
	} while (dom_is_changed);
	
	//debugging purpose, print the set of dominators for each node
#if 0
	for (i = 0; i < N; i++) {
		printf("cfg ID %i : ", i);
		ba_print(dom[i], N, stdout); 
		printf("\n");
	}
#endif
	
	//identify loop head/tail using dominator analysis outcome
	for (i = 0; i < proc->num_bb; i++) {
		bb = &proc->cfg[i];
		
		if (bb->out_t != NULL && ba_value(dom[i], bb->out_t->dst->id) == TRUE) {
			bb->out_t->dst->loop_role |= LOOP_HEAD;
			bb->loop_role |= LOOP_TAIL;
			tail_nodes[num_tail_nodes] = bb->id;
			tail_nodes_dest[num_tail_nodes++] = bb->out_t->dst->id;
			bb->out_t->is_backedge = 1;
		}
		
		if (bb->out_n != NULL && ba_value(dom[i], bb->out_n->dst->id) == TRUE) {
			bb->out_n->dst->loop_role |= LOOP_HEAD;
			bb->loop_role |= LOOP_TAIL;
			tail_nodes[num_tail_nodes] = bb->id;
			tail_nodes_dest[num_tail_nodes++] = bb->out_n->dst->id;
			bb->out_n->is_backedge = 1;
		}
	}
	
	// An identified tail node is not a real tail node if:
	// (1) it dominates another tail node, and
	// (2) both dominatee and dominator tail nodes points to the same loop header.
	// This step removes the fake tail nodes.
	for (i = 0; i < num_tail_nodes; i++) {
		for (j = 0; j < num_tail_nodes; j++) {
			if (i==j)
				continue;
			
			if (tail_nodes_dest[i] == tail_nodes_dest[j]
				&& ba_value(dom[tail_nodes[j]], tail_nodes[i]) == TRUE) {
				proc->cfg[tail_nodes[i]].loop_role &= ~LOOP_TAIL;
				break;
			}
		}
	}
	
	free(dom);
}



// create a CFG for a proc in three steps:
// - find basic block entries and create basic blocks
// - set basic info for blocks
// - finish up the construction of CFG by connecting blocks with edges
static void
create_cfg(proc_t *proc)
{
    int		*bb_ent, i, bb_id = 0;
    int          num;
    int is_skipping_code = 0;

    cfg_node_t	*bb;
    addr_t	addr;
    proc_t	*callee;
    
    

    bb_ent = (int *) calloc(proc->num_inst, sizeof(int));
    CHECK_MEM(bb_ent);
    
    // find & create blocks
    scan_blocks(bb_ent, proc);
    proc->cfg = (cfg_node_t *) calloc(proc->num_bb, sizeof(cfg_node_t));
    CHECK_MEM(proc->cfg);

    // set basic info for blocks
    for (i = 0; i < proc->num_inst; i++) {
		if (proc->code[i].interpret_mode == 0)
			is_skipping_code = 1;
		else if (proc->code[i].interpret_mode == 1)
			is_skipping_code = 0;
		if (is_skipping_code)
			continue;

		if (bb_ent[i]) {
			// start of a new block
				num = bb_ent[i];
				while(num > 1){
					 // add dummy node
					 bb = &proc->cfg[bb_id];
					 bb->id = bb_id;
					 bb->proc = proc;
					 bb->sa   = proc->code[i - 1].addr; // previous instruction 
					 bb->size = proc->code[i - 1].size;
					 bb->num_inst = 1;
					 bb->code  = &proc->code[i - 1];
					 bb_id++;
					 num--;
				}
			bb = &proc->cfg[bb_id];
			bb->id = bb_id;
			bb->proc = proc;
			bb->sa = proc->code[i].addr;
			bb->code = &proc->code[i];
			bb->size = proc->code[i].size;
			bb->num_inst = 1;
			bb_id++;
		} else { // continuation of current block
			/* cleekee: ignore NOP instruction */
			if (proc->code[i].type == INST_NOP)
				continue;

			bb->size += proc->code[i].size;
			bb->num_inst++;
		}
    }
    free(bb_ent);

    //TODO: cleekee: remove prior modifications to proc->num_bb
    proc->num_bb = bb_id;

#if 1
    // connect block with control transfer edges
    create_cfg_edges(proc);
#endif
    // build links from callers (basic blocks) to its callees (procs)
    for (i = 0; i < proc->num_bb; i++) {
		bb = &proc->cfg[i];
		if (bb->type == CTRL_CALL) {
			addr = bb->code[bb->num_inst-1].target;
			callee = lookup_proc(addr);
			if (callee == NULL) {
				//fprintf(stderr, "no proc mathces the callee addr: %x\n", addr);
				//exit(1);
				bb->type = CTRL_SEQ;
			} else
			   bb->callee = callee;
		}
    }

    // identify loop heads/tails
    identify_loops(proc);
}

boolean_t is_proc_task_or_component(int p_info_id, int proc_id, int* proc_ent, int first_ins_id) {
	int i;

	if (is_task_graph_given) {
		/* set flag if this procedure is a task in given task graph */
		for (i = 0; i < ntasks; i++) {
			if (tasks[i]->proc == NULL && strcmp(prog.p_info[p_info_id].name, tasks[i]->name) == 0) {
				tasks[i]->proc = &prog.procs[proc_id];

				/* readjust num_procs to include this proc if not already included */
				if (!proc_ent[first_ins_id])
					prog.num_procs++;

				return 1;
			}
		}
	}

	if (is_proc_separately_analyzed(prog.p_info[p_info_id].name)) {
		/* readjust num_procs to include this proc if not already included */
		if (!proc_ent[first_ins_id])
			prog.num_procs++;

		return 1;
	}

	return 0;
}

// identify procedures, and construct a control flow graph for each
void
build_cfgs()
{
    int	    *proc_ent, i, j, proc_id = 0, main_id;
    proc_t  *proc = NULL;
	int started = 0;

    proc_ent = (int *) calloc(prog.num_inst, sizeof(int));
    CHECK_MEM(proc_ent);

    // find & create procs
    main_id = scan_procs(proc_ent);
    prog.procs = (proc_t *) calloc(prog.total_procs, sizeof(proc_t));
    CHECK_MEM(prog.procs);

    // set basic info for procs
	if (binaryTarget == PISA) {
		for (i = 0; i < prog.num_inst; i++) {
			if (proc_ent[i]) {
				// start of a new proc
				proc = &prog.procs[proc_id];
				started = 1;
				proc->id = proc_id;
				proc->sa = prog.code[i].addr; 
				proc->size = prog.code[i].size;
				proc->num_inst = 1;
				proc->code = &prog.code[i];
				if (i == main_id)
					prog.main_proc = proc_id;
				proc_id++;
			} else if (proc != NULL) { //cleekee: checking 'proc' for NULL for debugging
				int temp;
				if (started == 0)
				    abort();
				temp = prog.code[i].size;
				// continuation of current proc
				proc->size += temp;
				proc->num_inst++;
			}
		}
		//free(proc_ent); //cleekee: disable for debugging
	} else if (binaryTarget == ARM)  {
		for (i = 0; i < prog.total_procs; i++) {
			int first_ins_id = lookup_addr(prog.code, prog.num_inst, prog.p_info[i].addr);
			assert(first_ins_id != -1 && "proc not found in prog!");
						
			/* skip procedure which is not called by any instructions in program
			 * AND is not a task or component */
			if (!is_proc_task_or_component(i, proc_id, proc_ent, first_ins_id) && !proc_ent[first_ins_id])
				continue;

			/* ignore certain procedures */
			if (strcmp(prog.p_info[i].name, "OSTimeTickHook")==0)
				continue;

			assert(prog.p_info[i].size > 0 && "Procedure should contain instruction(s)!");
			
			proc = &prog.procs[proc_id]; 
			proc->id = proc_id;
			proc->name = prog.p_info[i].name;
			proc->sa = prog.p_info[i].addr;
			proc->size = prog.p_info[i].size;
			proc->num_inst = proc->size/4;
			proc->code = &prog.code[first_ins_id];
			
			if (first_ins_id == main_id) {
				prog.main_proc = proc_id;
			}

			/* set system call as a separately analyzed component */
			if (is_proc_separately_analyzed(prog.p_info[i].name)) {
				component_t *component = (component_t*) calloc (1, sizeof(component_t));
				component->proc = proc;
				addToWorkList(&components, component);
			}

			proc_id++;
			
			/* the rest of the procedures are not analyzed */
			if (proc_id >= prog.num_procs) break;
		}
		prog.num_procs = proc_id;
	}

	FILE* fCfg = fopen("dbg_cfg_dump", "w");
    // create CFG for each proc
    for (i = 0; i < prog.num_procs; i++) {
		//cleekee: temporary measure only
		//if (prog.procs[i].size == 0) continue;
    	assert(prog.procs[i].size > 0 && "proc size should not be zero!");
		
		create_cfg(&prog.procs[i]);
		dump_cfg(fCfg, &prog.procs[i]);
    }
	
    fclose(fCfg);
	free(proc_ent);
}

boolean_t is_proc_separately_analyzed(char* proc_name) {
	int i;

	char *sys_call_name[10] = {
			"OSTimeDly",
			"OSTimeGet",
			"OSSemPend",
			"OSSemPost",
			"OS_Sched",
			"OSIntExit",
			"OSTaskSuspend",
			"OSTaskResume",
			"OSMemGet",
			"OSMemPut"
	};
#if 1
	for (i=0; i<10; i++) {
		if (strcmp(proc_name, sys_call_name[i]) == 0)
			return 1;
	}
#else
	if (proc_name[0] == 'O' && proc_name[1] == 'S' && proc_name[2] != '_')
		return 1;
#endif
	return 0;
}

void
dump_cfg(FILE *fp, proc_t *proc)
{
    cfg_node_t  *bb;
    int         i;

    fprintf(fp, "\nP%d \"%s\" cfg:\n", proc->id, proc->name);
    for (i = 0; i < proc->num_bb; i++) {
	bb = &proc->cfg[i];
	fprintf(fp, " %d : %08x : [ ", bb->id, bb->sa);
	if (bb->out_n != NULL)
	    fprintf(fp, " %d",  bb->out_n->dst->id);
	else
	    fprintf(fp, " ");
	fprintf(fp, " , ");
	if (bb->out_t != NULL)
	    fprintf(fp, " %d ",  bb->out_t->dst->id);
	else
	    fprintf(fp, " ");
	fprintf(fp, " ] ");
	if (bb->callee != NULL) {
	    fprintf(fp, " P%d \"%s\"", bb->callee->id, bb->callee->name);
	}
#if 0
	if (bb_is_loop_head(bb))
	    fprintf(fp, "/");
	if (bb_is_loop_tail(bb))
	    fprintf(fp, "\\");
#endif
	fprintf(fp, "\n");
    }
}



