/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Copyright (C) 2005 Xianfeng Li
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * $Id: cfg.h,v 1.2 2006/06/24 08:54:56 lixianfe Exp $
 *
 ******************************************************************************/

#ifndef CFG_H
#define CFG_H


#include <stdio.h>
#include "isa.h"
#include "common.h"
#include "unicache.h"

#define NOT_TAKEN	0
#define TAKEN		1
#define BOTH_BRANCHES	2
#define LOOP_HEAD	1
#define LOOP_TAIL	2


typedef enum {CTRL_SEQ, CTRL_COND, CTRL_UNCOND, CTRL_CALL, CTRL_RET, CTRL_RET_COND} bb_type_t;
 

typedef struct cfg_edge_t   cfg_edge_t;
typedef struct proc_t	    proc_t;

typedef int *damage_p;

// component info for compositional analysis
struct component_t {
	proc_t* proc; // component is essentially a procedure
	int wcet, wcrt, wcet_with_crpd;
	int *crpd;
	int crpd_cost;
	int num_preempt; // bound on number of preemptions by interrupts

	worklist_p *miss_info; // miss info for the data preemption. TODO: list by set
	damage_p data_damage;
	damage_p cumulative_data_damage;
	instdamage_p inst_damage;
	instdamage_p cumulative_inst_damage;
};
typedef struct component_t component_t;

// control flow graph node (basic block) type
typedef struct {
    int		id;		// basic block id (per procedure)
    proc_t  	*proc;		// up-link to the procedure containing it
    addr_t	sa;		// block start addr
    int		size;		// size (in bytes)
    int		num_inst;	// number of instructions
    de_inst_t   *code;		// pointer to the first instruction

    bb_type_t	type;		
    cfg_edge_t	*out_n, *out_t;	// outgoing edges (non-taken/taken) 
    int		num_in;		// number of incoming edges
    cfg_edge_t	**in;		// incoming edges

    proc_t	*callee;	// points to a callee if callee_addr not NULL

    int		loop_role;	// whether it is a loop head, tail, or neither

    int		flags;		// for traverse usage

              						/* A data structure holding abstract register */  
	 ric_p* in_abs_reg_value;	/* values. Register numbers can directly be */
	 ric_p* out_abs_reg_value; /* indexed in this structure */

	 abs_mem_p in_abs_mem_value;  /* A data structure holding value in abstract */
	 abs_mem_p out_abs_mem_value; /*  memory locations */


     /*** HBK: for scope-aware data cache analysis ***/
     void   *d_instlist;
     int    num_d_inst;

     /* cleekee: If node type is CTRL_CALL, this field points to the
      * component being called. Value is NULL if callee is not a component. */
     component_t* called_component;

} cfg_node_t;



// control flow graph edge type
struct cfg_edge_t {
    cfg_node_t  *src, *dst;	// src -> dst
	char is_backedge;
};



// procedure type
struct proc_t {
    int		id;		// proc id
    char*	name;	// proc name
    addr_t	sa;		// proc start addr
    int		size;		// size (in bytes)
    int		num_inst;	// number of instructions
    de_inst_t   *code;		// instructions

    int		num_bb;		// number of basic blocks
    cfg_node_t	*cfg;		// cfg nodes with num_bb nodes

    int		flags;
};

// symbol table (address range of global variable
typedef struct {
    char* name;
    addr_t addr;
    int size;
} symbol_i;


// program type
typedef struct {
    de_inst_t	*code;		// decoded program text
    int		    code_size;	// code size (in bytes)
    int		    num_inst;	// number of instructions
    addr_t	    start_addr, end_addr, main_addr;
    proc_t	    *procs;		// procedures
    symbol_i    *p_info;    // procedure symbol info
    symbol_i    *v_info;    // global variable symbol info
    int         num_vars;   // number of global variables
    int		    num_procs;	// number of procedures to analyze
	int			total_procs;// total number of procedures extracted from ARM binary
    int		    main_proc;	// index of the main proc
} prog_t;

#define MAX_OVRL_NODES 2048
#define OPT 1

struct col_data
{
	union {
		 proc_t* proc;
		 de_inst_t* inst;
	}u;
	int type;
	struct col_data* next;
};

typedef struct col_data* col_data_p;
typedef struct col_data col_data_s;

/* Describes the data structures used for layout formation */
struct ovrl_graph {
	union {
		 proc_t* proc;
		 de_inst_t* inst;
		 col_data_p pref;
	}u;
	int type;
	int invalid;
	/* execution count by WCET */
	int freq;
	struct ovrl_graph* next;
};
typedef struct ovrl_graph* ovrl_graph_p;
typedef struct ovrl_graph ovrl_graph_s;

boolean_t is_proc_separately_analyzed(char* proc_name);

void dump_cfg(FILE *fp, proc_t *proc);

#endif
