/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Infeasible path analysis for Chronos
 * Vivy Suhendra
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * 03/2007 infeasible.c
 *  v4.1:
 *      _ rewrite disassembly parser, change from bash shell to C, improve speed
 *      _ separate symbolic execution data type from infeasible.h
 *
 *  v4.0: as in Chronos-4.0 distribution
 *
 ******************************************************************************/

#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "isa.h"
#include "cfg.h"
#include "tcfg.h"
#include "loops.h"
#include "infeasible.h"
#include "task.h"

#define MD_INST_SIZE 0x8

extern prog_t prog;
extern isa_t  *isa;

extern int num_tcfg_nodes;
extern tcfg_node_t **tcfg;

extern tcfg_nlink_t ***bbi_map;
extern loop_t **loop_map;

extern char is_analyzing_component;

extern task_p* tasks;

enum BINARY_TARGET binaryTarget;

/*
 * Determines a reverse topological ordering of the procedure call graph.
 */
int topo_call( int **callees, int *num_callee ) {

  int i, currp;

  int  *comefrom;     // keep the proc from which we came, to return to
  int  comefromsize;
  char *markfin;      // markfin: 0: unvisited, 1: visited but unfinished, 2: finished
  char fin;

  int countdone;

  comefrom = NULL;
  comefromsize = 0;
  markfin = (char*) calloc( prog.num_procs, sizeof(char) );
  callgraph = (int*) calloc( prog.num_procs, sizeof(int) );

  countdone = 0;

  // start from main procedure
  currp = prog.main_proc;

  while( countdone < prog.num_procs ) {

    fin = 1;
    for( i = 0; fin && i < num_callee[currp]; i++ )
      if( markfin[ callees[currp][i] ] != 2 )
	fin = 0;

    if( num_callee[currp] <= 0 || fin ) {  // finished
      markfin[currp] = 2;
      callgraph[countdone++] = currp;

      // returning
      if( !comefromsize )
	break;

      currp = comefrom[--comefromsize];
      continue;
    }

    markfin[currp] = 1;
    comefrom = (int*) realloc( comefrom, (comefromsize+1) * sizeof(int) );
    comefrom[comefromsize++] = currp;

    // go to next unvisited procedure call
    fin = 1;
    for( i = 0; fin && i < num_callee[currp]; i++ ) {
      if( markfin[callees[currp][i]] != 2 ) {
	currp = callees[currp][i];
	fin = 0;
      }
    }
    if( fin )
      printf( "Unvisited called procedure not detected.\n" ), exit(1);

  } // end while( countdone < num_bb )

  free( comefrom );
  free( markfin );

  return 0;
}

//HBK: rewrite readInstr() for faster implementation
void editAsmStr(char *tmp) {
    int dbg = 0;
    int j, flag;
    FILE *dbgExec=stdout;
    flag = 0;
    if (dbg) fprintf(dbgExec,"\nProcess %s ->",tmp);
    for (j=0; j<strlen(tmp); j++) {
		/* cleekee: strip ARM's constant prefix */
		if (tmp[j]=='#')
			tmp[j]=' ';
			
		/* cleekee: assume offset addressing for load/store instruction, 
		 * currently does NOT handle post-indexed addressing */
		if (tmp[j]=='[' || tmp[j]==']')
			tmp[j]=' ';
			
        if (tmp[j]==',' || tmp[j]=='\\' || tmp[j]=='(' || tmp[j]==')')
            tmp[j]=' ';
        if (tmp[j]=='<') { flag=1; tmp[j]=' '; }
        if (tmp[j]=='>') { flag=0; tmp[j]=' '; }
        if (flag==1) tmp[j]=' ';
    }
    if (dbg) fprintf(dbgExec,"%s",tmp);
}


/* reading binary, constructing procedures and so on */
void readInsn ( char *obj_file ) {
    int dbg = 0;
    int i,j,k, flag;
    FILE *f;
    char tmp[120],str[120];
    int tmpAddr, curAddr;
    int curInsn;
    proc_t     *p;
    cfg_node_t *b;
    de_inst_t  *d;

    inf_proc_t *ip;
    inf_node_t *ib;
    insn_t     *is, *tmpIs;

    int n, len;
    int *num_callee = (int*) calloc(prog.num_procs, sizeof(int));
    int **callees = (int**) malloc(prog.num_procs*sizeof(int*));

    sprintf(tmp,"%s.dis",obj_file);
    f = fopen(tmp,"r");
    insn_t  *isList;

    char dummyArg[64];
	
		int timeout;
	
		isList = calloc(prog.num_inst,sizeof(insn_t));

    /* Read preprocessing */
    /* num_insn_st = 0;
    insnlist_st = NULL; */
    
		if (dbg) printf("\nPreprocessing instructions");

    while (fgets(tmp,120,f)) {
        if (strstr(tmp,(binaryTarget == PISA)? "0040" : ":	") == NULL) 
					continue; /* skip non-inst line */
			
        editAsmStr(tmp);

        /*num_insn_st++;
        insnlist_st = (insn_t*) realloc( insnlist_st, num_insn_st * sizeof(insn_t) );
				if (insnlist_st == NULL) {
				perror("Error (re)allocating memory");
				exit (1);
				}

        is = &(insnlist_st[num_insn_st-1]);
        is->r1[0]='\0';is->r2[0]='\0';is->r3[0]='\0';
        if (binaryTarget == PISA)
            sscanf( tmp, "%x %s %s %s %s\n",
                    &tmpAddr, is->op, is->r1, is->r2, is->r3 );
        else
						sscanf( tmp, "%x: %s %s %s %s %s\n",
                    &tmpAddr, dummyArg, is->op, is->r1, is->r2, is->r3 );
						sprintf(is->addr,"%x",tmpAddr);
        		if (dbg) printf("\n\tInsn: addr:%s op:%s r1:%s r2:%s r3:%s",
            				is->addr, is->op, is->r1, is->r2, is->r3);*/
			
				is = &(isList[0]);
				
				if (binaryTarget == PISA)
      		sscanf( tmp, "%x %s %s %s %s\n",
                    &tmpAddr, is->op, is->r1, is->r2, is->r3 );
    		else
					sscanf( tmp, "%x: %s %s %s %s %s\n",
                    &tmpAddr, dummyArg, is->op, is->r1, is->r2, is->r3 );
				sprintf(is->addr,"%x", prog.start_addr);

				/* Stop parsing at the address of the first instruction */
        if (tmpAddr == prog.start_addr) 
					break;
    }
	
    /* Read instructions into "isList" */
    if (dbg) printf("\nProgram instructions");
    
		/* curAddr = prog.start_addr; */
    for (i=1; i<prog.num_inst; i++) {
        is = &(isList[i]);
		
				#if 0
				/* TODO: is curAddr needed? */
        sprintf(is->addr,"%x",curAddr);
        if (0) printf("\nScan inst %x: ",curAddr);
				#endif
		
        while (fgets(tmp,120,f)) {
            if (strstr(tmp,(binaryTarget == PISA)? "0040" : ":	") == NULL) 
							continue; /* skip non-inst line */
            else 
							break;
     		}
		
        editAsmStr(tmp);
        is->r1[0]='\0';is->r2[0]='\0';is->r3[0]='\0';
		
        if (binaryTarget == PISA)
            sscanf( tmp, "%x %9s %9s %9s %9s\n",
                    &tmpAddr, is->op, is->r1, is->r2, is->r3 );
        else
						sscanf( tmp, "%x: %9s %9s %9s %9s %9s\n",
                    &tmpAddr, dummyArg, is->op, is->r1, is->r2, is->r3 );
        sprintf(is->addr,"%x",tmpAddr);
		
        if (dbg) printf("\n%i\tInsn: addr:%s op:%s r1:%s r2:%s r3:%s",i,
            is->addr, is->op, is->r1, is->r2, is->r3);
        //curAddr = curAddr+sizeof(MD_INST_SIZE);
    }

    for (i=0; i<prog.num_procs; i++) 
		callees[i] = NULL;

		// Map proc_t to inf_proc_t, cfg_node_t to inf_node_t, de_inst_t to insn_t
    curInsn = 0;

#ifdef _INF
    inf_procs = (inf_proc_t*) malloc(prog.num_procs*sizeof(inf_proc_t));

    for (i = 0; i < prog.num_procs; i++) {
        p = &(prog.procs[i]);

        ip= &(inf_procs[i]);
        ip->proc = p;
        ip->num_bb = p->num_bb;
        ip->inf_cfg = (inf_node_t*) malloc(ip->num_bb*sizeof(inf_node_t));

        for (j=0; j<ip->num_bb; j++) {
            b = &(p->cfg[j]);
            ib= &(ip->inf_cfg[j]);
            ib->bb = b;
            ib->num_insn = b->num_inst;
            ib->insnlist = (insn_t*) malloc (ib->num_insn*sizeof(insn_t));
            ib->num_assign = 0;
            ib->assignlist = NULL;
            ib->branch = NULL;
            ib->loop_id = -1;
            ib->exec_count = -1;
            ib->loop_role = b->loop_role;// copy loop_role of cfg block to inf block

            for (k=0; k<ib->num_insn; k++) {
                d = &(b->code[k]);
                is= &(ib->insnlist[k]);
                if (0) printf("\nFind insn %x: ",d->addr);
				
								// Iterate instructions in "isList" to find match for instruction "d"
								timeout = 0;

								do {
												tmpIs = &(isList[curInsn]);
												curInsn = (curInsn+1)%prog.num_inst;
												sscanf(tmpIs->addr,"%x",&curAddr);

												timeout++;
												if (timeout > prog.num_inst) break;

												//printf("i=%i, j=%i, k=%i, curInsn= %i, d->addr= %x, curAddr= %x\n", 
												//       i, j, k, curInsn, d->addr, curAddr);
								} while (curAddr < d->addr);

								// cleekee: this shouldn't happen, it indicates some instructions 
								// missing in .dis file. Put some default values.
								if (timeout > prog.num_inst) {
												strcpy(is->addr, d->addr);
												strcpy(is->op, '\0');
												strcpy(is->r1, '\0');
												strcpy(is->r2, '\0');
												strcpy(is->r3, '\0');
												continue;
								}

								strcpy(is->addr, tmpIs->addr);
								strcpy(is->op, tmpIs->op);
								strcpy(is->r1, tmpIs->r1);
								strcpy(is->r2, tmpIs->r2);
								strcpy(is->r3, tmpIs->r3);
								if (dbg) printf("\n\tInsn: addr:%s op:%s r1:%s r2:%s r3:%s",
																is->addr, is->op, is->r1, is->r2, is->r3);
						}
				}
		}
#else
    for (i = 0; i < num_tcfg_nodes; i++) {
            tcfg_node_t* bbi = tcfg[i];
						
            for (k = 0; k < bbi->bb->num_inst; k++) {
                d = &(bbi->bb->code[k]);
                is= &(bbi->insnlist[k]);
                if (0) printf("\nFind insn %x: ",d->addr);
				
								// Iterate instructions in "isList" to find match for instruction "d"
								timeout = 0;

    						curInsn = 0;
								
								do {
												tmpIs = &(isList[curInsn]);
												curInsn = (curInsn + 1) % prog.num_inst;
												sscanf(tmpIs->addr,"%x",&curAddr);

												timeout++;
												if (timeout > prog.num_inst) break;

												//printf("i=%i, j=%i, k=%i, curInsn= %i, d->addr= %x, curAddr= %x\n", 
												//       i, j, k, curInsn, d->addr, curAddr);
								} while (curAddr < d->addr);

								// cleekee: this shouldn't happen, it indicates some instructions 
								// missing in .dis file. Put some default values.
								if (timeout > prog.num_inst) {
												strcpy(is->addr, d->addr);
												strcpy(is->op, '\0');
												strcpy(is->r1, '\0');
												strcpy(is->r2, '\0');
												strcpy(is->r3, '\0');
												continue;
								}

								strcpy(is->addr, tmpIs->addr);
								strcpy(is->op, tmpIs->op);
								strcpy(is->r1, tmpIs->r1);
								strcpy(is->r2, tmpIs->r2);
								strcpy(is->r3, tmpIs->r3);

								is->bbi = bbi;

								if (dbg) printf("\n\tInsn: addr:%s op:%s r1:%s r2:%s r3:%s",
																is->addr, is->op, is->r1, is->r2, is->r3);
						}
				}
#endif /* sudiptac: INF_ related structure initialization ends */
		
		topo_call( callees, num_callee );

		/*  produced callgraph is in reverse order, so flip the array */
		n = prog.num_procs - 1;
		len = prog.num_procs / 2;
		for( i = 0; i < len; i++ ) {
						int tmp        = callgraph[i];
						callgraph[i]   = callgraph[n-i];
						callgraph[n-i] = tmp;
		}

		free(isList);
		free( num_callee );
		for( i = 0; i < prog.num_procs; i++ )
						free( callees[i] );
		free( callees );
		if (dbg) fflush(stdout);
}

#if 0
void readInstr( char *obj_file ) {

				int i, j, k;

				FILE *f;
				char tmp[120];
				int addr;

				proc_t     *p;
				cfg_node_t *b;
				de_inst_t  *d;

				inf_proc_t *ip;
				inf_node_t *ib;
				insn_t     *is;

				// for call graph construction
				int n, len;
				int *num_callee = (int*) calloc( prog.num_procs, sizeof(int) );
				int **callees = (int**) malloc( prog.num_procs * sizeof(int*) );
				for( i = 0; i < prog.num_procs; i++ )
								callees[i] = NULL;

				// transfer information that is already parsed
				inf_procs = (inf_proc_t*) malloc( prog.num_procs * sizeof(inf_proc_t) );
				for( i = 0; i < prog.num_procs; i++ ) {
								p  = &(prog.procs[i]);
								ip = &(inf_procs[i]);

								ip->proc = p;
								ip->num_bb = p->num_bb;
								ip->inf_cfg = (inf_node_t*) malloc( ip->num_bb * sizeof(inf_node_t) );

								for( j = 0; j < ip->num_bb; j++ ) {
												b  = &(p->cfg[j]);
												ib = &(ip->inf_cfg[j]);

												ib->bb = b;
												ib->num_insn = b->num_inst;
												ib->insnlist = (insn_t*) malloc( ib->num_insn * sizeof(insn_t) );
												ib->num_assign = 0;
												ib->assignlist = NULL;
												ib->branch = NULL;
												ib->loop_id = -1;
												ib->exec_count = -1;

												for( k = 0; k < ib->num_insn; k++ ) {
																d  = &(b->code[k]);
																is = &(ib->insnlist[k]);

																sprintf( is->addr, "%x", d->addr );
																strcpy( is->op, isa[d->op_enum].name );
																is->r1[0] = '\0';
																is->r2[0] = '\0';
																is->r3[0] = '\0';

																// for operands, read anew from .dis file, because these are not stored in de_inst_t
																sprintf( tmp, "sed -n '/^00%s/p' %s.dis | awk '{print $4}' > tline", is->addr, obj_file );
																system( tmp );
																system( "cat tline | tr ',' '\\ ' | tr '(' '\\ ' | tr ')' '\\ ' > topr" );

																f = fopen( "topr", "r" );
																fscanf( f, "%s %s %s", is->r1, is->r2, is->r3 );
																fclose( f );
												}
								}
				}

				// read program preprocessing
				// these are not part of CFG but needed for symbolic execution later
				sprintf( tmp, "grep __start %s.dis | awk '{print $1,$3,$4}' > st.dis", obj_file );
				system( tmp );
				sprintf( tmp, "cat st.dis | tr ',' '\\ ' | tr '(' '\\ ' | tr ')' '\\ ' > pre", obj_file );
				system( tmp );

				num_insn_st = 0;
				insnlist_st = NULL;

				f = fopen( "pre", "r" );
				while( fgets( tmp, INSN_LEN, f )) {
								num_insn_st++;
								insnlist_st = (insn_t*) realloc( insnlist_st, num_insn_st * sizeof(insn_t) );

								is = &(insnlist_st[num_insn_st-1]);
								is->r1[0] = '\0';
								is->r2[0] = '\0';
								is->r3[0] = '\0';
								sscanf( tmp, "%x %s %s %s %s\n", &addr, is->op, is->r1, is->r2, is->r3 );
								sprintf( is->addr, "%x", addr );  // to ensure same format (not depending on 0x, 00)
				}
				fclose( f );

				system( "rm -f tline topr pre st.dis" );

				topo_call( callees, num_callee );

				// produced callgraph is in reverse order, so flip the array
				n = prog.num_procs - 1;
				len = prog.num_procs / 2;
				for( i = 0; i < len; i++ ) {
								int tmp        = callgraph[i];
								callgraph[i]   = callgraph[n-i];
								callgraph[n-i] = tmp;
				}

				free( num_callee );
				for( i = 0; i < prog.num_procs; i++ )
								free( callees[i] );
				free( callees );
}
#endif

/*
 * Returns 1 if paft is pbef in the call graph; i.e. it is deeper in the chain of calls.
 * Assume there is no cyclic calls.
 */
char isDeeper( int paft, int pbef ) {

				int i;
				for( i = 0; i < prog.num_procs; i++ ) {
								if( callgraph[i] == paft )
												return 0;
								if( callgraph[i] == pbef )
												return 1;
				}
				return 0;
}


int setCountRec( tcfg_node_t *bbi, int count, int return_pid ) {

				tcfg_edge_t *e;

				if( bbi->exec_count >= count ) {
								printf( "Existing count tcfg %d(%d::%d) = %d\n", bbi->id, bbi->bb->proc->id, bbi->bb->id, bbi->exec_count );
								return 1;
				}
				bbi->exec_count = count;
				printf( "Set count tcfg %d(%d::%d) = %d\n", bbi->id, bbi->bb->proc->id, bbi->bb->id, count );

				for( e = bbi->out; e != NULL; e = e->next_out ) {
								if( e->dst->bb->proc->id != return_pid )
												setCountRec( e->dst, count, return_pid );
				}
				return 0;
}


int setCount( tcfg_node_t *bbi, int count ) {

				if( bbi->exec_count >= count ) {
								printf( "Existing count tcfg %d(%d::%d) = %d\n", bbi->id, bbi->bb->proc->id, bbi->bb->id, bbi->exec_count );
								return 1;
				}
				bbi->exec_count = count;
				//printf( "Set count tcfg %d(%d::%d) = %d\n", bbi->id, bbi->bb->proc->id, bbi->bb->id, count );

				// propagate to called procedure
				if( bbi->bb->callee != NULL && bbi->out != NULL )
								setCountRec( bbi->out->dst, count, bbi->bb->proc->id );

				return 0;
}


int setLoopIDRec( tcfg_node_t *bbi, int lpid, int return_pid ) {
				int dbg = 0;
				tcfg_edge_t *e;
				inf_loop_t *lp = &(inf_loops[lpid]);

				if( bbi->loop_id == lpid || isDeeper( inf_loops[bbi->loop_id].pid, lp->pid ))
								return 1;

				bbi->loop_id = lpid;
				if (dbg) printf( "Set tcfg %d(%d::%d)  loop:%d  entry:(%d::%d)  bound:%d\n",
												bbi->id, bbi->bb->proc->id, bbi->bb->id, lpid, lp->pid, lp->entry, lp->bound );

				for( e = bbi->out; e != NULL; e = e->next_out ) {
								if( e->dst->bb->proc->id != return_pid )
												setLoopIDRec( e->dst, lpid, return_pid );
				}
				return 0;
}


int setLoopID( tcfg_node_t *bbi, int lpid ) {
				int dbg = 0;

				inf_loop_t *lp = &(inf_loops[lpid]);

				if( bbi->loop_id == lpid || isDeeper( inf_loops[bbi->loop_id].pid, lp->pid ))
								return 1;

				bbi->loop_id = lpid;
				if (dbg) printf( "Set tcfg %d(%d::%d)  loop:%d  entry:(%d::%d)  bound:%d\n",
												bbi->id, bbi->bb->proc->id, bbi->bb->id, lpid, lp->pid, lp->entry, lp->bound );

				// propagate to called procedure
				if( bbi->bb->callee != NULL && bbi->out != NULL )
								setLoopIDRec( bbi->out->dst, lpid, bbi->bb->proc->id );

				return 0;
}


/*
 * Recursively set loop id for reachable nodes.
 */
void markLoop( inf_proc_t *ip, inf_node_t *head, inf_node_t *ib, int lpid, char **checked ) {

				int i;
				tcfg_nlink_t *ndlink;
				inf_loop_t *lp = &(inf_loops[lpid]);

				// In case of nested loops, we rely on the assumption that the inner loop 
				// will have larger entry ID. Update only if the new information is for 
				// the deeper level.
				if( ib->loop_id == -1 || inf_loops[ib->loop_id].entry < lp->entry ) {
								//printf( "Set (%d.%d)  loop:%d  entry:%d  bound:%d\n", ip->proc->id, \
								// ib->bb->id, lpid, lp->entry, lp->bound );
								ib->loop_id = lpid;

								// propagate to tcfg nodes
								for( ndlink = bbi_map[ip->proc->id][ib->bb->id]; ndlink != NULL; ndlink = ndlink->next )
										setLoopID( ndlink->bbi, lpid );
				}
				else {
								printf( "Nested (%d.%d)  loop:%d  entry:%d  (not updated)\n", ip->proc->id, \
									ib->bb->id, ib->loop_id, inf_loops[ib->loop_id].entry );
				}
				(*checked)[ib->bb->id] = 1;
				//if( ib->bb->loop_role == LOOP_HEAD )
				if( ib == head ) {
								return;
				}

				for( i = 0; i < ib->bb->num_in; i++ ) {
								int pre = ib->bb->in[i]->src->id;
								//HBK: > or >= , can entry node considered in-loop?
								//Consider entry node in loop because it also executed LB times
								if( !(*checked)[pre] && pre >= lp->entry )
												markLoop( ip, head, &(ip->inf_cfg[pre]), lpid, checked );
				}
}


/*
 * Parses loop bound information and stores known block execution counts,
 * to be used in constraint formulation.
 */
int readBlockCounts( char *obj_file ) {
				FILE *f;
				FILE *cons_file;
				int fake_bounds = 0;
				char tmp[256];
				int i, pid, bid, pid2, entry, count;
				
#ifdef _INF
				inf_proc_t *ip;
				inf_node_t *ib;
#endif

				for( i = 0; i < num_tcfg_nodes; i++ ) {
								tcfg[i]->loop_id = -1;
								tcfg[i]->exec_count = -1;
				}

				// absolute counts
				sprintf( tmp, "cat %s.cons | tr 'c' '\\ ' | awk '$2 ~ \"=\" {print $1,$NF}' | tr '.' '\\ ' > counts", obj_file );
				system( tmp );

				f = fopen( "counts", "r" );
				
				while( fscanf( f, "%d %d %d", &pid, &bid, &count ) != EOF ) {

								tcfg_nlink_t *ndlink;

								//printf( "Detected absolute count %d::%d = %d\n", pid, bid, count );

								if( !include_proc[pid] ) {
												printf( "Ignoring count for excluded function.\n" );
												continue;
								}
#ifdef _INF
								inf_procs[pid].inf_cfg[bid].exec_count = count;
#endif
								// propagate to tcfg nodes
								for( ndlink = bbi_map[pid][bid]; ndlink != NULL; ndlink = ndlink->next )
												setCount( ndlink->bbi, count );
				}
				fclose( f );

				// loopbounds
				sprintf( tmp, "grep \\< %s.cons | sed '/+/d' | tr '-' '\\ ' | tr '=' '\\ ' | tr '\\<' '\\ ' > tmp", obj_file );
				system( tmp );
				system( "awk '{if(NF==4) print $1,$3,$2}' tmp | tr 'c' '\\ ' | tr '.' '\\ ' > bounds" );

#ifdef _INF
				num_inf_loops = 0;
				inf_loops = NULL;

				for( pid = 0; pid < prog.num_procs; pid++ ) {
								if( include_proc[pid] ) {
												ip = &(inf_procs[pid]);
												for( bid = 0; bid < ip->num_bb; bid++ )
																ip->inf_cfg[bid].loop_id = -1;
								}
				}
#endif /* sudiptac: end of INF_ loop initialization */
	
				char buf[512];
				strncpy(buf, obj_file, 512);
				strncat(buf, ".cons", 512);
				struct stat st;

				if (stat(buf, &st) == -1){
								fake_bounds = 1;
								perror("Enabling fake loop bounds, because during .cons load");
				} 
				f = fopen( "bounds", "r" );
				i = -1;

				if (fake_bounds) 
								cons_file = fopen( buf, "w" );

				for (;;) {
								if (fake_bounds) {
												i++;
												if (i >= num_tcfg_nodes)
																break;
												if ((tcfg[i]->bb->loop_role & LOOP_HEAD) == 0)
																continue;
												pid2 = bbi_pid(tcfg[i]->in->src);
												entry = bbi_bid(tcfg[i]->in->src);
												bid = bbi_bid(tcfg[i]);
												pid = bbi_pid(tcfg[i]);
												count = FAKE_BOUND;  

												fprintf(cons_file, "c%d.%d - %d c%d.%d <= 0\n", pid, bid, count, pid2, entry);
								} else {
												if (fscanf( f, "%d %d %d %d %d", &pid, &bid, &pid2, &entry, &count ) == EOF)
																break;
								}

								if (count <= 0) {
									perror("Negative or zero loop bound given in .cons file. Abort.\n");
									exit(1);
								}
							
								//printf("-----printing loop bound information-----\n");
								//printf("%d %d %d %d %d\n", pid, bid, pid2, entry, count);

								int lpid, head, tail;
								loop_t *lpn;

#ifdef _INF
								inf_loop_t *lp;
								inf_node_t *hd, *ta;
#endif

								char *checked;
								tcfg_nlink_t *ndlink;

								//printf( "Detected loop bound for %d::%d (entry: %d) = %d\n", pid, bid, entry, count );

								if( !include_proc[pid] || !include_proc[pid2] ) {
												printf( "Ignoring bound for excluded function.\n" );
												continue;
								}

#ifdef _INF
								ip = &(inf_procs[pid]);
								ib = &(ip->inf_cfg[bid]);

								lpid = num_inf_loops++;
								inf_loops = (inf_loop_t*) realloc( inf_loops, num_inf_loops * sizeof(inf_loop_t) );
								lp = &(inf_loops[lpid]);
								lp->pid = pid;
								lp->entry = entry;
								lp->bound = count;
								/* 
									mark all blocks in the loop body
									traverse from tail upwards until 
									head is found; all traversed block 
									is in the loop 
								*/
								lpn = loop_map[bbi_map[pid][bid]->bbi->id];
								lp->loop_id = lpn->id; //HBK: to backtrace the loop_t of this inf_loop
								
								//printf("Flag h:%d t:%d\n",lpn->head, lpn->tail);fflush(stdout);
								head = bbi_bid( lpn->head );
								/*HBK: shouldn't the entry be counted in the loop?*/
								head = bbi_bid( lpn->head );
								if (lpn->tail!=NULL)
												tail = bbi_bid( lpn->tail );
								else
												tail = bbi_bid(tcfg[num_tcfg_nodes-1]);
								hd = &(ip->inf_cfg[head]);
								ta = &(ip->inf_cfg[tail]);
								checked = (char*) calloc( ip->num_bb, sizeof(char) );
								markLoop( ip, hd, ta, lpid, &checked );
								free( checked );
#else
								/* 
									mark all blocks in the loop body
									traverse from tail upwards until 
									head is found; all traversed block 
									is in the loop 
								*/
								//printf("setting loop bound.....\n");
								for( ndlink = bbi_map[pid][bid]; ndlink != NULL; ndlink = ndlink->next ) {
										lpn = loop_map[ndlink->bbi->id];
										lpn->bound = count;
										//printf("loop bound (%d) set for tcfg loop id (%d)@0x%x\n", count, lpn->id, lpn);
								}
#endif
				}

				fclose( f );
				if (fake_bounds)
								fclose( cons_file );

				//system( "rm tmp counts bounds" );
				return 0;
}


/* sudiptac: merge leekee changes into infeasible analysis with task_id */
void infeas_analysis( char *obj_file, int task_id ) { 
  const int DEBUG = 0;
  char filename[256];
  if (DEBUG) {
    printf("\nAnalyze file |%s|\n",obj_file);fflush(stdout);
    printf( "Read Instr...\n" ); fflush(stdout);
    printf( "Read Block Counts...\n" ); fflush(stdout);
  }

  if(DEBUG) {printf("[infeas_analysis] Reading binary...\n");}
  readInsn( obj_file );

  if(DEBUG) {printf("[infeas_analysis] Parsing given loop bounds and known BB execution counts...\n");}
  if (task_id == -1) {
    readBlockCounts(obj_file);
  } else {
	if (is_analyzing_component)
		sprintf(filename, "%s_comp_%s", obj_file, prog.procs[task_id].name);
	else
		sprintf(filename, "%s_task_%i", obj_file, task_id);

    readBlockCounts(filename);
  }

  //printInstructions();

  if(DEBUG) {printf("[infeas_analysis] Starting symbolic execution...\n");}

  execute();
  //printEffects(0);

  //printf( "Starting conflict detection...\n" ); fflush(stdout);
  //detectConflicts();
  //printf( "Detected %d BB and %d BA\n", num_BB, num_BA ); fflush(stdout);
  //printEffects(1);
}
