/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Copyright (C) 2005 Xianfeng Li
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * $Id: main.c,v 1.3 2006/07/15 03:22:50 lixianfe Exp $
 *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "cfg.h"
#include "bpred.h"
#include "cache.h"
#include "address.h"
#include "scp_cache.h"
#include "task.h"
#include "util.h"

#define CEIL(VARIABLE) ( (VARIABLE - (int)VARIABLE)==0 ? (int)VARIABLE : (int)VARIABLE+1 )

#define CPU_FREQ 454000000

#define INF_ENABLED 1


extern int nsets_dl1, assoc_dl1;
extern int nsets, assoc;

char DEBUG_INFEAS = 0;

int PROGRESS_STEP = 10000;
extern int num_tcfg_edges;
extern int num_tcfg_nodes;
extern tcfg_node_t **tcfg;
extern tcfg_edge_t** tcfg_edges;
extern tcfg_nlink_t*** bbi_map;

extern char *run_opt;
extern FILE *filp, *fusr;
extern loop_t* getIbLoop(inf_node_t *ib);

int bpred_scheme;
int enable_icache = 0;
int enable_scp_dcache = 0; /* Enable scope-aware dcache analysis */
int enable_scp_dl2cache = 0;/* Enable scope-aware level 2 dcache analysis */
//int enable_dcache = 0; /* For anabling dcache analysis */
int enable_il2cache = 0; /* For enabling level 2 icache analysis */
int enable_ul2cache = 0; /* For enabling level 2 ucache analysis */

int enable_abs_inst_cache = 0; /* For anabling abstract icache analysis */

int is_task_graph_given = 0;

int ISR_cost = 0;

int X, Y, B, l1, l2;
extern worklist_p*** scp_addrset_l1;
extern worklist_p*** scp_addrset_l2;

prog_t prog;

FILE* fTaskStat;
struct timespec start, end, start2, end2;

enum ILPSolver {
	ILP_CPLEX,
	ILP_LPSOLVE
};

// vivy: infeasible path analysis
#include "infeasible.h"

// vivy: marker for procedures to include in estimation
char *include_proc;

// cleekee: set zero to disable compositional analysis of system calls
char is_compositional = 1;

// cleekee: store components that need to be analyzed separately
worklist_p components;
char is_analyzing_component;

/* sudiptac:: For performance measurement */
int findproc(int addr) {
	int i;
	for (i = 0; i < prog.num_procs; i++)
		if (prog.procs[i].sa == addr)
			return i;
	return -1;
}

// vivy: read list of functions to include in analysis
// Some analysis steps are expensive so avoid processing unnecessary functions.
int read_functions(char *obj_file) {
	FILE *fptr;
	char fname[80];

	int id;
	int addr;
	char name[80];

	include_proc = (char*) calloc(prog.num_procs, sizeof(char));

	sprintf(fname, "%s.fnlist", obj_file);
	fptr = fopen(fname, "r");

	if (!fptr) {
		if (DEBUG_INFEAS)
			printf(
					"%s.fnlist not found -- all procedures will be included in estimation.\n",
					obj_file);
		for (id = 0; id < prog.num_procs; id++)
			include_proc[id] = 1;
		return -1;
	}

	while (fscanf(fptr, "%x %s", &addr, name) != EOF) {
		id = findproc(addr);
		if (id != -1)
			include_proc[id] = 1;
		else
			printf("Warning: procedure [0x%x] %s not found.\n", addr, name);
	}
	fclose(fptr);
	return 0;
}

void my_dump_tcfg(int* unrolled_loop_map) {
	extern tcfg_node_t** tcfg;
	extern int num_tcfg_nodes;
	printf("\nnumber of tcfg node:%d\n", num_tcfg_nodes);
	int i;
	for (i = 0; i < num_tcfg_nodes; i++) {
		tcfg_node_t* bbi = tcfg[i];
		printf("tcfg node:%d (%d) lp:%d", i, bbi->id, loop_map[bbi->id]->id);
		printf("\tout: (");
		tcfg_edge_t* e;
		for (e = bbi->out; e != NULL; e = e->next_out) {
			printf(" %d", e->dst->id);
		}
		printf(")");
		printf("\tin: (");
		for (e = bbi->in; e != NULL; e = e->next_in) {
			printf(" %d", e->src->id);
		}
		printf(")\n");

	}
	printf("\n DUMP loop\n");
	for (i = 0; i < num_tcfg_loops; i++) {
		loop_t*lp = loops[i];
		printf("loop %d bound:%d", lp->id, lp->bound);
		if (lp->parent != NULL
		)
			printf(" parent:%d", lp->parent->id);
		if (unrolled_loop_map != NULL
		)
			printf(" mapping:%d", unrolled_loop_map[lp->id]);
		tcfg_node_t* head = lp->head;
		tcfg_node_t* tail = lp->tail;
		if (tail == NULL) {
			tail = tcfg[num_tcfg_nodes - 1];
		}
		printf("\n");
		printf("\thead: %d\n", head->id);
		printf("\ttail: %d\n", tail->id);
		printf("\n");
	}
//	printf("\nDUMP topo\n");
//	for (i = 0; i < num_tcfg_nodes; i++) {
//		tcfg_node_t* bbi = tcfg[topo_tcfg[i]];
//		printf("topo_id %d: %d\n", i, bbi->id);
//	}
#if 1

#endif
#if 0
	printf("checking instructions\n");
	if (enable_icache) {
		extern cache_t cache;
		for (i = 0; i < num_tcfg_nodes; i++) {
			tcfg_node_t* bbi = tcfg[i];
			printf("basic block %d\n", bbi->id);
			int inst_id;
			for (inst_id = 0; inst_id < bbi->bb->num_inst; inst_id++) {
				de_inst_t*inst = &(bbi->bb->code[inst_id]);
				int mblk = MBLK_ID(bbi->bb->sa,inst->addr);
				int offset = CACHE_LINE(inst->addr);
				printf("\tinst %d: 0x%x (%d) offset=%d\n", inst_id, inst->addr,
						mblk, offset);
			}
		}
	}
#endif
}

char* get_component_name(int task_id) {
	if (is_analyzing_component) {
		return "SYSCALL";
	} else if (tasks[task_id]->is_interrupt) {
		return "ISR";
	} else {
		return "TASK";
	}
}

// construct control flow graphs from objective code
static void read_code_and_build_cfg(char *fName) {
	int i, j;
	char obj_file[256];
	// read object code, decode it
	strcpy(obj_file, fName);
	read_code(obj_file);

	// create procs and their CFGs from the decoded text
	build_cfgs();

	// vivy: read list of functions to include in estimation
	// do this after prog.procs are established
	strcpy(obj_file, fName);
	read_functions(obj_file);
}

// program flow analysis
static void path_analysis(char *fName, int task_id) {
	const int DEBUG = 0;

	char obj_file[256];

	// transform the CFGs into a global CFG called tcfg (transformed-cfg);
	if (is_task_graph_given && is_analyzing_component)
		prog_tran(task_id);
	else if (is_task_graph_given && !is_analyzing_component)
		prog_tran(tasks[task_id]->proc->id);
	else
		prog_tran(prog.main_proc);
		
	FILE* fvtcfg = fopen("visual-tcfg.dot", "w");
	dump_visual_tcfg(fvtcfg);
	fclose(fvtcfg);	
	// identify loop levels as well as block-loop mapping
	loop_process();
#if 0
	int i;
	for (i=0;i<num_tcfg_loops;i++) {
		loop_t* lp = loops[i];
		tcfg_node_t* lpHead= lp->head;
		printf("\nloop %d  lpHead bb:%d\n",lp->id,lpHead->bb->id);
		if (lp->parent!=NULL) printf("\tparent:%d\n",lp->parent->id);
	}
	exit(0);
#endif

#if 0
	// find topological order of all tcfg nodes
	// topo_tcfg[topo_id] = tcfg_id
	set_topological_tcfg();
#endif
	/********************************************/
	/* vivy: infeasible path analysis */
#if 0
	my_dump_tcfg(NULL);
#endif
	strcpy(obj_file, fName);
#if INF_ENABLED
	if(DEBUG) {printf("Performing infeasibility analysis...\n");}
	infeas_analysis(obj_file, task_id);
#endif
#if 0
	exit(0);
#endif
	/********************************************/
	if (enable_scp_dcache) {
		/* sudiptac: do only when icache is not enabled, otherwise after the virtual unrolling */
		if (!enable_icache) {
		    extern int assoc_dl1, nsets_dl1, bsize_dl1, mem_lat[2];
		    X = assoc_dl1, Y = nsets_dl1, B = bsize_dl1, l1 = mem_lat[0], l2 = 0;
		    scp_pre_address_analysis(fName, &scp_addrset_l1);
		    if (enable_scp_dl2cache || enable_ul2cache) {
				extern int assoc_dl2, nsets_dl2, bsize_dl2, cache_dl2_lat,
						cache_dl1_lat;
				X = assoc_dl2, Y = nsets_dl2, B = bsize_dl2, l1 = cache_dl2_lat, l2 =
						mem_lat[0];

				scp_pre_address_analysis(fName, &scp_addrset_l2);
		    }
		    /*
		     * TODO: perform address analysis here, store the addresses in ...
		     */
		    scp_store_address_set();
    	}
	} else {
		/*
		 * set loop-bound for loops
		 * NOTE: If scp_enable_dcache is enabled ,
		 * the loop bound is set during address analysis
		 */
/* #if INF_ENABLED */
		int i;
#ifdef _INF /* sudiptac: loop bound has already been set */
		for (i = 0; i < prog.num_procs; i++) {
			inf_proc_t* ip = &(inf_procs[i]);
			int j;
			for (j = 0; j < ip->num_bb; j++) {
				inf_node_t* ib = &(ip->inf_cfg[j]);
				loop_t* lp = (loop_t*) getIbLoop(ib);
				if (lp) {
					lp->bound = getIbLB(ib);
					lp->rId = lp->rType = lp->rBound = -1;
				}
			}
		}
#else
		for (i = 0; i < num_tcfg_loops; i++) {
				loop_t* lp = loops[i];
				if (lp) 	
					lp->rId = lp->rType = lp->rBound = -1;
		}
#endif
		loops[0]->bound = 1;
/* #endif */
	}
	/********************************************/

	if (enable_icache) {

                /*
		 * TODO: store loop_id, loop_bound
		 */
		
                /* sudiptac: we no longer create inf_loops, so changing 
		 * the following implementation */

		int index;
		int old_num_tcfg_loops = num_tcfg_loops;

#ifdef _INF
		int* old_loop_id = calloc(num_tcfg_loops, sizeof(int));
#else
		int** old_loop_id = calloc(prog.num_procs, sizeof(int *));	

		for (index = 0; index < prog.num_procs; index++) {
			old_loop_id[index] = calloc(num_tcfg_nodes, sizeof(int));
		}
#endif
		
		int* old_loop_bound = calloc(old_num_tcfg_loops, sizeof(int));
		int i;
		
#ifdef _INF
		for (i = 0; i < num_tcfg_loops; i++) {
			old_loop_id[i] = loops[i]->id;
			old_loop_bound[i] = loops[i]->bound;
		}
#else
		for (i = 0; i < num_tcfg_loops; i++) {
			proc_t* proc = loops[i]->head->bb->proc;
			cfg_node_t* head = loops[i]->head->bb;
			old_loop_id[proc->id][head->id] = loops[i]->id;
			old_loop_bound[i] = loops[i]->bound;
		}
#endif

		/******************************************/
		if(DEBUG) {printf("Performing virtual loop unrolling\n");}
		virtual_unroll();
	
		/* recollect the tcfg edges as the transformed CFG has been
		 * augmented after virtual unrolling */
		collect_tcfg_edges();
		build_bbi_map();
		loop_process();
		/*
		 *  find topological order of all tcfg nodes
		 *  topo_tcfg[topo_id] = tcfg_id
		 */
		set_topological_tcfg();
		
		if (is_task_graph_given && !is_analyzing_component)
			tasks[task_id]->topo_tcfg = topo_tcfg;
		
		// EXAMPLE load damage
		
		// Load cache damage from file
/*		FILE *d  = fopen("cache.dmg", "r");
		damage_p dmg = load_damage(d);
		fclose(d);
		print_damage(dmg); */
		
		//Remove CFG from node 13 to node 18, replace by damage
		// virtual_bypass(13, 18, dmg);
		
		FILE *cfg;
		cfg = fopen("tcfg.map", "w");
		dump_tcfg(cfg);
		fclose(cfg);

#if 0 
		/*****************************************/
		/*
		 * TODO: calculate loop_bound (only in 
                 * depricated mode)
		 */
		for (i = 0; i < num_tcfg_loops; i++) {
			loop_t* lp = loops[i];
			tcfg_node_t* head = lp->head;
			if (head == NULL)
				continue;
			cfg_node_t* head_bb = head->bb;

			inf_proc_t* iproc = &(inf_procs[head_bb->proc->id]);
			inf_node_t* ihead = &(iproc->inf_cfg[head_bb->id]);
			lp->bound = getIbLB(ihead) - 1;
		}
#endif
            	/*****************************************/
		/*
		 * sudiptac: calculate loop_bound 
		 * (only in depricated mode)
		 */

#ifdef _INF /* sudiptac: loop bound has already been set */
		for (i = 0; i < num_tcfg_loops; i++) {
			loop_t* lp = loops[i];
			tcfg_node_t* head = lp->head;
			if (head == NULL)
				continue;
			cfg_node_t* head_bb = head->bb;

			inf_proc_t* iproc = &(inf_procs[head_bb->proc->id]);
			inf_node_t* ihead = &(iproc->inf_cfg[head_bb->id]);
			lp->bound = getIbLB(ihead) - 1;
		}
#else
		for (i = 0; i < num_tcfg_loops; i++) {
			loop_t* lp = loops[i];
			proc_t* proc = loops[i]->head->bb->proc;
			cfg_node_t* head = loops[i]->head->bb;
			int old_lpid = old_loop_id[proc->id][head->id];

			lp->bound = old_loop_bound[old_lpid] - 1;
		}
#endif

		loops[0]->bound = 1; //TODO: test here

		/*****************************************/
		if (enable_scp_dcache) {
			//TODO: recalculate temporal scopes
			/*
			 * TODO: create a mapping from new loops to old loops
			 */
			int* unrolled_loop_map = calloc(num_tcfg_loops, sizeof(int));
			for (i = 0; i < num_tcfg_loops; i++) {
				tcfg_node_t* head = loops[i]->head;
				if (head == NULL)
					continue;
				cfg_node_t* head_bb = head->bb;

#ifdef _INF
				inf_proc_t* iproc = &(inf_procs[head_bb->proc->id]);
				inf_node_t* ihead = &(iproc->inf_cfg[head_bb->id]);
				if (ihead->loop_id == -1) {
					unrolled_loop_map[loops[i]->id] = 0;
				} else {
					unrolled_loop_map[loops[i]->id] =
							old_loop_id[inf_loops[ihead->loop_id].loop_id];
				}
#else
				/* sudiptac: changed the implementation for computing 
				 * unrolled_loop_map */
				unrolled_loop_map[0] = 0;
				unrolled_loop_map[loops[i]->id] = old_loop_id[head_bb->proc->id][head_bb->id];
#endif
			}
			
			if(DEBUG) {printf("Performing address analysis for scope-aware data cache analysis\n");}

			extern int assoc_dl1, nsets_dl1, bsize_dl1, mem_lat[2];
			X = assoc_dl1, Y = nsets_dl1, B = bsize_dl1, l1 = mem_lat[0], l2 = 0;
			scp_pre_address_analysis(fName, &scp_addrset_l1);
			if (enable_scp_dl2cache || enable_ul2cache) {
				extern int assoc_dl2, nsets_dl2, bsize_dl2, cache_dl2_lat, cache_dl1_lat;
				X = assoc_dl2, Y = nsets_dl2, B = bsize_dl2, l1 = cache_dl2_lat, l2 = mem_lat[0];

				scp_pre_address_analysis(fName, &scp_addrset_l2);
			}
		       /*
		 	* TODO: perform address analysis here, store the addresses in ...
		 	*/
			scp_store_address_set();
			/*
			 * TODO: filter the temporal scope
			 */
			scp_recalculate_temporal_scope(unrolled_loop_map, old_loop_bound);
		}
	}
	/**********************************************/
#if 0
	/*
	 * NOTE: checking the consistency between loop id(s) of temporal scope(s)
	 * and the loops of basic block
	 * If there are any problems, program is terminated with error flag.
	 */
	// my_dump_tcfg(NULL);
//	scp_dump_address();
	int i;
	/*for (i=0;i<num_tcfg_nodes;i++) {
	 tcfg_node_t* bbi = tcfg[i];
	 printf("Basic block %d\n",bbi->id);
	 int j;
	 for (j= 0; j< bbi->bb->num_inst;j++) {
	 de_inst_t* inst = &(bbi->bb->code[j]);
	 printf("\tinst[%d]: 0x%x",j,inst->addr);
	 if (inst_type(inst) == INST_LOAD )
	 printf("[LOAD]");
	 else if (inst_type(inst) == INST_STORE)
	 printf("[STORE]");
	 printf("\n");
	 }
	 }*/
	for (i = 0; i < num_tcfg_nodes; i++) {
		tcfg_node_t* bbi = tcfg[i];
		if (bbi->bb->num_d_inst == 0)
		continue;
		printf("Basic block:%d bb:%d(%d)\n", bbi->id,bbi->bb->id,bbi->bb->num_d_inst);
		loop_t* innerLoop = loop_map[bbi->id];
		worklist_p innerList = NULL;
		worklist_p outerList = NULL;
		loop_t* lpPointer = innerLoop;
		for (; lpPointer != NULL; lpPointer = lpPointer->parent) {
			addToWorkList(&outerList, lpPointer);
		}
		worklist_p lNode;
		for (lNode = outerList; lNode != NULL; lNode = lNode->next) {
			addToWorkList(&innerList, lNode->data);
		}
		int j;
		int bugs=0;
		for (j = 0; j < bbi->bb->num_d_inst; j++) {
			dat_inst_t* dat_inst = bbi->bb->d_instlist;
			dat_inst = dat_inst + j;
			insn_t* insn = dat_inst->insn;
			worklist_p addrnode = bbi->addrset_l1[j];
			for (; addrnode != NULL; addrnode = addrnode->next) {
				saddr_p saddr = addrnode->data;
				worklist_p tsNode = saddr->tsList;
				for (;tsNode!=NULL;tsNode=tsNode->next) {
					ts_p ts = tsNode->data;
					loop_t* lpNode = innerLoop;
					int flag =0;
					for (;lpNode!=NULL;lpNode=lpNode->parent) {
						if (lpNode->id == ts->loop_id) {
							flag=1;
							break;
						}
					}
					if (!flag) {
						bugs=1;
						printf("\tdinst[%d] 0x%s: 0x%x ",j,insn->addr,saddr->blkAddr);
						scp_print_ts(saddr->tsList);
						printf("\t|");
						for (lNode = innerList;lNode!=NULL;lNode = lNode->next) {
							loop_t* tmp = lNode->data;
							printf(" l%d",tmp->id);
						}
						printf("\n");
					}
				}
			}
		}
		if (bugs) {
			printf("inconsistency BUGGG\n");
			exit(1);
		}
	}

	//int i;
	int * visited = calloc(num_tcfg_nodes, sizeof(int));
	worklist_p qHead, qTail;
	for (i = 0; i < num_tcfg_loops; i++) {
		loop_t* lup = loops[i];
		tcfg_node_t* lpHead = lup->head;
		int j;
		for (j = 0; j < num_tcfg_nodes; j++) {
			visited[j] = 0;
		}
		qHead = qTail = NULL;
		scp_enqueue(lpHead, &qHead, &qTail);
		visited[lpHead->id] = 1;
		printf("loop level %d\n",lup->id);
		while (scp_empty_queue(qHead, qTail) == 0) {
			tcfg_node_t* curNode = scp_dequeue(&qHead, &qTail);
			printf("\tbbi:%d\n",curNode->id);
			tcfg_edge_t* out_e;
			for (out_e = curNode->out; out_e != NULL; out_e = out_e->next_out) {
				if (scp_cmpLpOrder(lup->id, loop_map[out_e->dst->id]->id) >= 0) { //TODO: change here
					if (visited[out_e->dst->id] == 0) {
						scp_enqueue(out_e->dst, &qHead, &qTail);
						visited[out_e->dst->id] = 1;
					}
				}
			}
		}
	}
	//exit(0);
#endif

	if(DEBUG) {printf("Completed path analysis of task %i\n", task_id);}
}

static void microarch_modeling() {
	if (bpred_scheme != NO_BPRED)
		bpred_analysis();

	pipe_analysis();
}

static int do_ilp(char *obj_file, int task_id) {
	printf("Solving ILP of %s %d\n", get_component_name(task_id), task_id);

	int exit_val;
	int	wcet_dag;
	char s[256], filename[256];

	if (task_id == -1)
		sprintf(filename, "%s", obj_file);
	else if (is_analyzing_component)
		sprintf(filename, "%s_comp_%s", obj_file, prog.procs[task_id].name);
	else
		sprintf(filename, "%s_task_%i", obj_file, task_id);

	sprintf(s, "%s.lp", filename);
	filp = fopen(s, "w");
	sprintf(s, "%s.cons", filename);
	fusr = fopen(s, "r");
	if ((filp == NULL) || (fusr == NULL)) {
		fprintf(stderr, "fail to open ILP/CONS files for writing/reading\n");
		exit(1);
	}

	constraints();

	fclose(filp);
	fclose(fusr);

	/* vivy: print a cplex version */
	sprintf(s, "%s.ilp", filename);
	filp = fopen(s, "w");
	fprintf(filp, "enter Q\n");
	fclose(filp);

	// same with lp_solve format but no comment supported
	// sprintf( s, "sed '/\\\\/d' %s.lp >> %s.ilp", obj_file, obj_file );
	// EDIT: cplex supports the same comment format as lp_solve
	sprintf(s, "cat %s.lp >> %s.ilp", filename, filename);
	system(s);

	sprintf(s, "%s.ilp", filename);
	filp = fopen(s, "a");
	fprintf(filp, "optimize\n");
	fprintf(filp, "set logfile %s.sol\n", filename);
	fprintf(filp, "display solution objective\n");
	fprintf(filp, "display solution variables -\n");
	fprintf(filp, "quit\n");
	fclose(filp);

	sprintf(s, "%s.lp", filename);
	wcet_dag = (int) solve_ILP( s, ILP_LPSOLVE );

	if (task_id != -1 && is_analyzing_component) {
		sprintf(s, "%s.lp", filename);
		remove(s);
		sprintf(s, "%s.ilp", filename);
		remove(s);
		//sprintf(s, "%s.cons", filename);
		//remove(s);
	}

	if (is_task_graph_given && !is_analyzing_component) {
		tasks[task_id]->wcet = wcet_dag;
	}

	//fprintf(stdout, "\n------------------<Printing WCET of DAG>------------------------------\n");
	if (task_id >= 0 && !is_analyzing_component) {
		//fprintf(stdout, "task name = %s\n", tasks[task_id]->name);
		fprintf(fTaskStat, "WCET of DAG = %d\n", wcet_dag);
	}
	//fprintf(stdout, "WCET of DAG = %d\n", wcet_dag);
	//fprintf(stdout, "-----------------------<END>------------------------------------------\n\n");

	/* Command:
	 * rm -f %s.sol; cplex < %s.ilp >/dev/null 2>/dev/null; cat %s.sol | sed '/^/s/Obj/obj/'
	 */

	return wcet_dag;
}

static int run_est(char *obj_file, int task_id) {
    printf("Running bpred and pipeline analysis of %s %d\n",
    		get_component_name(task_id), task_id);
	microarch_modeling();

	return (do_ilp(obj_file, task_id));
}

static void run_cfg(char *obj_file) {
	int i;
	char s[128];
	FILE *fcfg;

	sprintf(s, "%s.cfg", obj_file);
	//printf("dumping control flow graphs to file:%s\n", s);
	fcfg = fopen(s, "w");
	if (fcfg == NULL) {
		fprintf(stderr, "fail to create file: %s.cfg\n", s);
		exit(1);
	}

	for (i = 0; i < prog.num_procs; i++) {
		dump_cfg(fcfg, &prog.procs[i]);
	}
	fclose(fcfg);

	//printf("done.\n");
}

extern int fetch_width;

/* modification to indirect jump */
/* liangyun */

typedef struct jptb {
	addr_t adr;
	int ntarget;
	addr_t *ptarget;
	int index;
} jptb;

jptb * pjptb;
int bjptb = 0;
int njp;

int lookup_jptable(addr_t adr) {
	int i;
	int num = 0;
	for (i = 0; i < njp; i++) {
		if (pjptb[i].adr == adr) {
			num = pjptb[i].ntarget;
		}
	}
	return num;
}

void get_jptable(addr_t src, int index, addr_t * target) {
	int i;
	for (i = 0; i < njp; i++) {
		if (pjptb[i].adr == src) {
			*target = pjptb[i].ptarget[index];
			break;
		}
	}
	assert(i < njp);
}

int get_jptable_static(addr_t src, addr_t * target) {
	int i, j;
	int tail = -1;
	for (i = 0; i < njp; i++) {
		if (pjptb[i].adr == src) {
			j = pjptb[i].index;
			*target = pjptb[i].ptarget[j];
			pjptb[i].index++;
			if (pjptb[i].index == pjptb[i].ntarget)
				tail = 0;
			else
				tail = 1;
			break;
		}
	}
	//assert(i < njp);
	return tail;
}

void read_injp(char * objfile) {
	int tablesize, i, j;
	char file[100];
	FILE *ftable;
	sprintf(file, "%s.jtable", objfile);
	ftable = fopen(file, "r");
	if (!ftable) {
		// no indirect jump
		bjptb = 0;
	} else {
		bjptb = 1;
		fscanf(ftable, "%d", &tablesize);
		njp = tablesize;
		pjptb = (jptb *) calloc(tablesize, sizeof(jptb));
		for (i = 0; i < tablesize; i++) {
			fscanf(ftable, "%08x", &pjptb[i].adr);
			fscanf(ftable, "%d", &pjptb[i].ntarget);
			pjptb[i].index = 0;
			pjptb[i].ptarget = (addr_t *) calloc(pjptb[i].ntarget,
					sizeof(addr_t));
			for (j = 0; j < pjptb[i].ntarget; j++)
				fscanf(ftable, "%08x", &pjptb[i].ptarget[j]);
		}
		fclose(ftable);
	}
	//printf("bool: %d\n",bjptb);
}

int *pdepth;
int bdepth = 0;

int test_depth(int pid, int depth) {
	if (depth < pdepth[pid])
		return 1;
	else
		return 0;
}

void read_recursive(char * objfile) {
	char file[100];
	FILE *ftable;
	int size, i;
	sprintf(file, "%s.recursive", objfile);
	ftable = fopen(file, "r");
	if (!ftable) {
		bdepth = 0;
	} else {
		bdepth = 1;
		fscanf(ftable, "%d", &size);
		pdepth = (int *) calloc(size, sizeof(pdepth));
		for (i = 0; i < size; i++)
			fscanf(ftable, "%d", &pdepth[i]);
		fclose(ftable);
	}
}
#if 0
/* HBK: scope-aware data cache analysis */
int X,Y,B,l1,l2;
extern int assoc_dl1, nsets_dl1,bsize_dl1,mem_lat[2];
static void scp_aware_datacache_analysis(char *bin_fname) {
	//printf("\nADDRESS ANALYSIS: %s\n", bin_fname);
	ticks a, b;
	//X = 2; Y = 32; B = 32; l1 = 6; l2 = 0;//no L2 cache
	//X = assoc_dl1; Y = nsets_dl1;B = bsize_dl1;l1=mem_lat[0];
	/*
	 * TODO: unnecessary to perform address analysis, just copy
	 * approriate temporal scope from scp_addrset_l1,scp_addreset_l2
	 */
	/*
	 a = getticks();
	 //classified_address_analysis(bin_fname);
	 b = getticks();
	 printf("\n===================================================\n");
	 printf("Address analysis time = %lf secs\n", (b - a) / ((1.0) * CPU_MHZ));
	 printf("===================================================\n");
	 */
	a = getticks();

	//NOTE: need to repair Vivy's AB BB infeasible detection
	enable_dcache = 0;
	printf("\nCACHE ANALYSIS: %s\n", bin_fname);
	fflush(stdout);
	//mpa_datacache();
	mpaex_datacache();
	b = getticks();
	printf("\n===================================================\n");
	printf("Cache analysis time = %lf secs\n", (b - a) / ((1.0) * CPU_MHZ));
	printf("===================================================\n");
}
#endif
/* sudiptac :::: analyze two level cache hierarchies */
static void analyze_cache_hierarchy() {
	const int DEBUG = 0;

	inst_chmc_l1 = inst_chmc_l2 = NULL;
	inst_age_l1 = inst_age_l2 = NULL;

	if(DEBUG) {printf("=========== Instruction cache analysis ===========\n");}
	/**********************************************/
	ticks a, b;

	a = getticks();
#ifdef _DEBUG_CRPD
	printf("Starting cache analysis........\n");
#endif

	if (enable_icache) {
		/* FIXME: this flag need to be removed in final version */
		//enable_abs_inst_cache = 1;
		/* cleekee: duplicated here from run_est() to access mp instructions */
		if (bpred_scheme != NO_BPRED)
			collect_mp_insts();

		/* sudiptac : analyze instruction cache (abstract interpretation approach) */
		analyze_abs_instr_cache_all();
	}

	b = getticks();
#ifdef _DEBUG_CRPD
	printf("Finished cache analysis........\n");
#endif

	//printf("===================================================\n");
	//printf("Maximum cache analysis time = %lf msecs\n",
	//		1000 * (b - a) / ((1.0) * CPU_MHZ));
	if(DEBUG) {printf("======= Finished instruction cache analysis =======\n");}
}

static void scp_aware_analyze_cache_hierarchy() {
	const int DEBUG = 0;

	if (enable_scp_dcache == 1) {
		extern int assoc_dl1, nsets_dl1, bsize_dl1, mem_lat[2];
		if(DEBUG) {printf("================= L1 data cache ===================\n");}
		/*initACS();*/
		X = assoc_dl1, Y = nsets_dl1, B = bsize_dl1, l1 = mem_lat[0], l2 = 0;
		mpaex_datacache(L1_DCACHE_ANALYSIS);
		if (enable_scp_dl2cache == 1 || enable_ul2cache == 1) {
			extern int assoc_dl2, nsets_dl2, bsize_dl2, cache_dl2_lat;
			if (enable_scp_dl2cache)
				if(DEBUG) {printf(
						"================= L2 data cache ===================\n");}
			else {
				if(DEBUG) {printf(
						"=================== UL2 cache ======================\n");}
			}
			X = assoc_dl2, Y = nsets_dl2, B = bsize_dl2, l1 = cache_dl2_lat, l2 =
					mem_lat[0];
			if (enable_scp_dl2cache == 1)
				mpaex_datacache(L2_DCACHE_ANALYSIS);
			else if (enable_ul2cache == 1) {
				set_cache_l2();
				mpaex_datacache(UNIFIED_CACHE_ANALYSIS);
			}
		}
		//	scp_dump_address();
		if(DEBUG) {printf("============ Finished L1 data cache ===============\n");}
	}
}

void build_component(char *filename, component_t *component) {
	int set;
	int i,j;
	int time_inside_interrupts = 0;
	int wcrt_old;
	int *total_ageing = malloc(sizeof(int)*nsets_dl1);
	int *num_preempt = malloc(sizeof(int) * ntasks);
	
	printf("Performing path analysis of system call %i\n", component->proc->id);

	path_analysis(filename, component->proc->id);
	
	analyze_cache_hierarchy();
	//compute instruction damage
	component->inst_damage = compute_instdamage();

	//scope-aware PS analysis for dcache
	scp_aware_analyze_cache_hierarchy();

	component->wcet = run_est(filename, component->proc->id);

/*
	for (i = 0; i < num_tcfg_nodes; i++) {
		for (j = 0; j < tcfg[i]->bb->num_d_inst ; j++) {
			if (tcfg[i]->address_set[j] == NULL)
				continue;
			worklist_p addr_in;
			for (addr_in = tcfg[i]->address_set[j]; addr_in; addr_in = addr_in->next) {
				saddr_p saddr = addr_in->data;
				saddr->blkAddr = 0x42;
			}
		}

	}
*/

	//compute data damage
	if (enable_scp_dcache) {
		assert(Y == nsets_dl1);
		assert(Y == MAX_CACHE_SET);
	}
	component->data_damage = compute_all(&component->miss_info);
}

void update_component(component_t *component) {
	int set;
	int i,j;
	int wcrt_old;
	int *total_ageing = malloc(sizeof(int)*nsets_dl1);
	int *num_preempt = malloc(sizeof(int) * ntasks);

	printf("Performing interrupt analysis of system call %i\n", component->proc->id);

	//compute instruction CRPD
	component->crpd = malloc(sizeof(int) * ntasks);
	memset(component->crpd, 0, sizeof(int) * ntasks);
	for (i = 0; i < ntasks; i++) {
		if (tasks[i]->is_interrupt) {
			component->crpd[i] = compute_crpd_fifo(-1, i)*L1_MISS_PENALTY;
		}
	}
	
	component->wcrt = component->wcet;
	
	do {
		wcrt_old = component->wcrt;
		component->wcrt = component->wcet;
		component->wcet_with_crpd = component->wcet;
		component->num_preempt = 0;
		component->crpd_cost = 0;
		memset(total_ageing, 0, sizeof(int) * nsets_dl1 ); 
		
		for (i=0; i<ntasks; i++) {
			if (tasks[i]->is_interrupt) {
				num_preempt[i] = CEIL((double) wcrt_old/tasks[i]->period);
				component->num_preempt += num_preempt[i];
				int preempt_cost = component->crpd[i] + pipeline_delay_cost();// + ISR_cost;
				component->crpd_cost += num_preempt[i] * component->crpd[i];
				component->wcrt += num_preempt[i] * (preempt_cost + tasks[i]->wcet);
				component->wcet_with_crpd += num_preempt[i] * preempt_cost;

				if (tasks[i]->data_damage == NULL) {
					fprintf(stderr, "WARNING: Data damage was not computed for task %u (interrupt)\n", i);
					continue;
				}
				
				//data ageing
				for (set = 0; set < nsets_dl1; set++) {
					total_ageing[set] += tasks[i]->data_damage[set] * num_preempt[i];
				}

					
			}
			
		}
		
		//compute crpd data
		for (set = 0; set < nsets_dl1; set++) {
			int crpd_contrib = ageing2crpd(component->miss_info[set], total_ageing[set]);
			component->wcrt += crpd_contrib;
			component->wcet_with_crpd += crpd_contrib;
		}

	} while (component->wcrt != wcrt_old);

	//account for additionnal data damage due to interrupts
/*
	component->cumulative_data_damage = malloc(sizeof(int) * nsets_dl1);
	component->cumulative_inst_damage = malloc(sizeof(instdamage_t));
	component->cumulative_inst_damage->lru_damage = malloc(sizeof(int) * nsets);

	for (set = 0; set < nsets_dl1; set++) {
		component->cumulative_data_damage[set] = component->data_damage[set];
		for (i = 0; i < ntasks; i++) {
			if (tasks[i]->data_damage == NULL) {
				fprintf(stderr, "WARNING: Data damage was not computed for task %u (interrupt)\n", i);
			} else  component->cumulative_data_damage[set] += tasks[i]->data_damage[set] * num_preempt[i];
		}
	}
	for (set = 0; set < nsets; set++) {
		component->cumulative_inst_damage->lru_damage[set] = component->inst_damage->lru_damage[set];
		for (i = 0; i < ntasks; i++) {
			if (tasks[i]->inst_damage == NULL) {
				fprintf(stderr, "WARNING: Instruction damage was not computed for task %u (interrupt)\n", i);
			} else  component->cumulative_inst_damage->lru_damage[set] += tasks[i]->inst_damage->lru_damage[set] * num_preempt[i];
		}
	}
*/

	//account for additionnal instruction damage due to interrupts


				
	//first, merge used blocks for interrupt and component. 
		
	//current component block
/*
	worklist_p wl = component->inst_damage->blocks;
	component->cumulative_inst_damage->blocks = NULL;
	for (; wl ; wl = wl->next) {
		int *block = wl->data;
		addToWorkList(&component->cumulative_inst_damage->blocks, block);
	}
*/	
	//merge blocks from interrupts
	/*
	for (i = 0; i < ntasks; i++) {

		//interrupt
		wl = tasks[i]->inst_damage->blocks;
		component->cumulative_inst_damage->blocks = NULL;
		for (; wl ; wl = wl->next) {
			int *block = wl->data;
			addToWorkList(&component->cumulative_inst_damage->blocks, block);
		}
		
		
		
	}*/
	
		
	//compute block count
	/*
	int blockcount = 0;
	component->cumulative_inst_damage->block_count = 0;
	wl = component->cumulative_inst_damage->blocks;
	for (; wl ; wl = wl->next)
		component->cumulative_inst_damage->block_count ++;
		
	// conservative values, because we are not sure that any interrupt will run
	component->cumulative_inst_damage->fifo_exit = component->inst_damage->fifo_exit;
	component->cumulative_inst_damage->must_exit = component->inst_damage->must_exit; */ 
	
	/* cleekee: Cumulative damage is not needed, because the potential damage caused by ISR
	 * on the running task when it preempts a system call (invoked by the running task), would
	 * already be taken into account in the WCRT computation. */
	component->cumulative_inst_damage = component->inst_damage;
	component->cumulative_data_damage = component->data_damage;
	
}

void read_and_analyze_one_task(char *argv, int task_id) {
	int set;
	int crpd_data;
	int i,j;
	tcfg_node_t *node;

	char *component_name = get_component_name(task_id);

	printf("Performing path analysis of %s %i\n", component_name, task_id);

	path_analysis(argv, (task_id==-1)? -1 : task_id);

	printf("Performing instruction cache analysis of %s %i\n", component_name, task_id);
	analyze_cache_hierarchy();
	
	if ((task_id != -1) && (tasks[task_id]->is_interrupt)) {
		tasks[task_id]->inst_damage = compute_instdamage();
	}

	if (enable_scp_dcache == 1) {
		printf("Performing scope-aware persistence data cache analysis of %s %i\n",
				component_name, task_id);
		scp_aware_analyze_cache_hierarchy();
	}

	/*
	for (i = 0; i < num_tcfg_nodes; i++) {
		for (j = 0; j < tcfg[i]->bb->num_d_inst ; j++) {
			if (tcfg[i]->address_set[j] == NULL)
				continue;
			worklist_p addr_in;
			for (addr_in = tcfg[i]->address_set[j]; addr_in; addr_in = addr_in->next) {
				saddr_p saddr = addr_in->data;
				saddr->blkAddr = 0x42;
			}
		}

	}
*/
	//fprintf(stderr, "analyzed tas: %u", task_id);

	printf("Computing data cache damage for %s %i\n", component_name, task_id);

	if (task_id != -1) {
	
		tasks[task_id]->data_damage = compute_all(&tasks[task_id]->miss_info);
		
		//This is normally needed only for interrupts because.

//cumulative data damage is only for components, because delays due to interrupt from (non-component) tasks will be handled by CRPD

//		tasks[task_id]->cumulative_data_damage = NULL;
		/* malloc(sizeof(int) * X);

		for (set = 0; set < X; set++) {
			tasks[task_id]->cumulative_data_damage[set] = tasks[task_id]->data_damage[set];
		}
		*/
		
	}

#if 0
	//scp_dump_instruction();
	scp_dump_address();
	//my_dump_tcfg(NULL);
//	exit(0);
#endif
}

/* set analysis for this_task */
void set_task(task_p this_task)
{
	/* set TCFG related parameters */
	num_tcfg_nodes = this_task->num_tcfg_nodes;
	num_tcfg_edges = this_task->num_tcfg_edges;
	tcfg = this_task->tcfg;
	tcfg_edges = this_task->tcfg_edges;
	bbi_map = this_task->bbi_map;
	topo_tcfg = this_task->topo_tcfg;

	/* set loop related parameters */
	num_tcfg_loops = this_task->num_tcfg_loops;
	loops = this_task->loops;
	loop_map = this_task->loop_map;
	loop_comm_ances = this_task->loop_comm_ances;

	/* set analysis related paramters */
	inst_chmc_l1 = this_task->inst_chmc_l1;
	inst_chmc_l2 = this_task->inst_chmc_l2;
	inst_age_l1 = this_task->inst_age_l1;
	inst_age_l2 = this_task->inst_age_l2;
#if 0
	ecb_log_l1 = this_task->ecb_log_l1;
	ecb_log_l2 = this_task->ecb_log_l2;
#endif
}

/* sudiptac :::: read and analyze each task in the task graph */
static void read_and_analyze_task_graph(FILE* tfp, char* app_name)
{
	int tid = 0, i;
	double period_in_ms;
	char tname[MAX_TASKNAME_LEN];
	Queue queue;

	fscanf(tfp, "%d", &ntasks);

	/* allocate memory for all the tasks */
	tasks = (task_p *) calloc(ntasks, sizeof(task_p));
	CHECK_MEM(tasks);

	while (tid < ntasks)
	{
		//reset();

		fscanf(tfp, "%s", tname);

		/* allocate memory for current task */
		tasks[tid] = (task_p) calloc(1, sizeof(task_s));
		CHECK_MEM(tasks[tid]);

		tasks[tid]->name = (char *) calloc(strlen(tname) + 1, sizeof(char));
		strcpy(tasks[tid]->name, tname);

		tasks[tid]->proc = NULL;

		tasks[tid]->tid = tid;

		/* task parameters */
		fscanf(tfp, "%d", &(tasks[tid]->priority));
		//printf("Task %s....priority - %d\n", tname, tasks[tid]->priority);
		
		fscanf(tfp, "%lf", &(period_in_ms));
		/* convert from milliseconds to CPU cycles */
		tasks[tid]->period = period_in_ms / 1000 * CPU_FREQ;
		
#if 0
		tasks[tid]->start_time_low = tasks[tid]->start_time_high = 0;

		fscanf(tfp, "%d", &(tasks[tid]->ncore));
		fscanf(tfp, "%d", &(tasks[tid]->mscid));

		printf("Task %s....assigned core - %d\n", tname, tasks[tid]->ncore);
		printf("Task %s....assigned MSC - %d\n", tname, tasks[tid]->mscid);

		/* successor list */
		fscanf(tfp, "%d", &(tasks[tid]->nsucc));
		for (i = 0; i < tasks[tid]->nsucc; i++) {
			if (!tasks[tid]->succ)
				tasks[tid]->succ = (int *) calloc(tasks[tid]->nsucc, sizeof(int));
			fscanf(tfp, "%d", &(tasks[tid]->succ[i]));
			printf("Reading successor %d.....\n", tasks[tid]->succ[i]);
		}
#endif

		// FIXME : fill correctly
		tasks[tid]->is_syscall = 0;
		fscanf(tfp, "%d", &(tasks[tid]->is_interrupt));

		tasks[tid]->wcet_analysis_time = 0;
		tasks[tid]->wcrt_analysis_time = 0;

		tid++;
	}

	/* read binary and create a CFG for each procedure */
	printf("\n********** Read binary and create CFG for each procedure **********\n");
	read_code_and_build_cfg(app_name);

	printf("\n********** Construct TCFG and analyze each ISR **********\n");
	for (tid=0; tid<ntasks; tid++) {
		/* Only need to compute summary for ISR in this round */
		if (!tasks[tid]->is_interrupt) {
			continue;
		}

		get_clock_time(&start);

		/* read and construct TCFG of one task */
		read_and_analyze_one_task(app_name, tid);

		get_clock_time(&end);
		tasks[tid]->wcet_analysis_time += (end.tv_sec+(double)end.tv_nsec/1000000000
				- start.tv_sec+(double)start.tv_nsec/1000000000);

		/* set TCFG related parameters */
		tasks[tid]->num_tcfg_nodes = num_tcfg_nodes;
		tasks[tid]->num_tcfg_edges = num_tcfg_edges;
		tasks[tid]->tcfg = tcfg;
		tasks[tid]->tcfg_edges = tcfg_edges;
		/* used for mapping CFG node to all TCFG nodes */
		tasks[tid]->bbi_map = bbi_map;

		/* set loop related parameters */
		tasks[tid]->num_tcfg_loops = num_tcfg_loops;
		tasks[tid]->loops = loops;
		tasks[tid]->loop_map = loop_map;
		tasks[tid]->loop_comm_ances = loop_comm_ances;

		/* set cache related parameters (after intra-core cache analysis) */
		tasks[tid]->inst_chmc_l1 = inst_chmc_l1;
		tasks[tid]->inst_chmc_l2 = inst_chmc_l2;
		tasks[tid]->inst_age_l1 = inst_age_l1;
		tasks[tid]->inst_age_l2 = inst_age_l2;			 
	}

	/* update_component() need to be done after read_and_analyze_one_task(),
	since it needs damage info from interrupts. */

	/* cleekee: pre-compute WCET and cache damage of components(system calls) */
	printf("\n********** Compute WCET and cache damage of system calls **********\n");
	is_analyzing_component = 1;
	worklist_p component;
	for (component = components; component != NULL; component = component->next) {
		component_t* component_data = (component_t*) component->data;
		build_component(app_name, component_data);
		update_component(component_data);

		/* cleekee: TODO: ISR cost only consider WCET of OSIntExit(), not quite safe */
		if (strcmp(component_data->proc->name, "OSIntExit") == 0)
			ISR_cost = component_data->wcet;
	}

#if 1
	/* cleekee: for debugging only */
	fprintf(fTaskStat, "-------------- System call WCET --------------\n");
	for (component = components; component != NULL; component = component->next) {
		component_t* component_data = (component_t*) component->data;
		fprintf(fTaskStat, "%s: WCET = %i\n", component_data->proc->name, component_data->wcet);
	}
#endif
	is_analyzing_component = 0;

#if 0
	/* cleekee: pre-compute WCET and cache damage of components(system calls) */
	/*printf("\n********** Compute WCET and cache damage of system calls **********\n");
	is_analyzing_component = 1;
	worklist_p component;*/
	for (component = components; component != NULL; component = component->next) {
		component_t* component_data = (component_t*) component->data;
		//build_component(app_name, component_data);
		update_component(component_data);

		/* cleekee: TODO: ISR cost only consider WCET of OSIntExit(), not quite safe */
		/*if (strcmp(component_data->proc->name, "OSIntExit") == 0)
			ISR_cost = component_data->wcet;*/
	}
	//is_analyzing_component = 0;
#endif

	printf("\n********** Construct TCFG and analyze each ISR **********\n");
	for (tid=0; tid<ntasks; tid++) {
		/* Only need to compute summary for ISR in this round */
		if (!tasks[tid]->is_interrupt) {
			continue;
		}

		get_clock_time(&start);

		/* read and construct TCFG of one task */
		read_and_analyze_one_task(app_name, tid);

		get_clock_time(&end);
		tasks[tid]->wcet_analysis_time += (end.tv_sec+(double)end.tv_nsec/1000000000
				- start.tv_sec+(double)start.tv_nsec/1000000000);

		/* set TCFG related parameters */
		tasks[tid]->num_tcfg_nodes = num_tcfg_nodes;
		tasks[tid]->num_tcfg_edges = num_tcfg_edges;
		tasks[tid]->tcfg = tcfg;
		tasks[tid]->tcfg_edges = tcfg_edges;
		/* used for mapping CFG node to all TCFG nodes */
		tasks[tid]->bbi_map = bbi_map;

		/* set loop related parameters */
		tasks[tid]->num_tcfg_loops = num_tcfg_loops;
		tasks[tid]->loops = loops;
		tasks[tid]->loop_map = loop_map;
		tasks[tid]->loop_comm_ances = loop_comm_ances;

		/* set cache related parameters (after intra-core cache analysis) */
		tasks[tid]->inst_chmc_l1 = inst_chmc_l1;
		tasks[tid]->inst_chmc_l2 = inst_chmc_l2;
		tasks[tid]->inst_age_l1 = inst_age_l1;
		tasks[tid]->inst_age_l2 = inst_age_l2;
	}

	printf("\n********** Construct TCFG and analyze each task **********\n");
	for (tid=0; tid<ntasks; tid++) {
		/* No need to compute summary for ISR in this round */
		if (tasks[tid]->is_interrupt) {
			continue;
		}

		/* read and construct TCFG of one task */
		read_and_analyze_one_task(app_name, tid);

		/* set TCFG related parameters */
		tasks[tid]->num_tcfg_nodes = num_tcfg_nodes;
		tasks[tid]->num_tcfg_edges = num_tcfg_edges;
		tasks[tid]->tcfg = tcfg;
		tasks[tid]->tcfg_edges = tcfg_edges;
		/* used for mapping CFG node to all TCFG nodes */
		tasks[tid]->bbi_map = bbi_map;

		/* set loop related parameters */
		tasks[tid]->num_tcfg_loops = num_tcfg_loops;
		tasks[tid]->loops = loops;
		tasks[tid]->loop_map = loop_map;
		tasks[tid]->loop_comm_ances = loop_comm_ances;

		/* set cache related parameters (after intra-core cache analysis) */
		tasks[tid]->inst_chmc_l1 = inst_chmc_l1;
		tasks[tid]->inst_chmc_l2 = inst_chmc_l2;
		tasks[tid]->inst_age_l1 = inst_age_l1;
		tasks[tid]->inst_age_l2 = inst_age_l2;

		int i, total_inst = 0;
		for(i=0; i<num_tcfg_nodes; i++) {
			total_inst += tcfg[i]->bb->num_inst;
		}
	}
}

component_t* find_component(char* name_to_find) {
	worklist_p component;
	for (component = components; component != NULL; component = component->next) {
		component_t* component_data = (component_t*) component->data;
		if (strcmp(component_data->proc->name, name_to_find) == 0)
			return component_data;
	}

	return NULL;
}

int context_switch_cost() {
	/* cleekee: No additional cost needs to be considered as:
	 * (1) Task level context switch cost, OS_Sched() is already computed in OSTimeDly(),
	 *     which is part of a task's WCET.
	 * (2) Interrupt level context switch cost, OSIntExit() is already computed in
	 *     the time tick ISR, which is computed as a task.
	 * Possibly only need to consider the cost of pushing/popping registers from stack.
	 */
	return 0;
#if 0
	int wcet1, wcet2, context_switch_cost;

	/* Task preemption occurs from either task-level or interrupt level
	 * context switch, get the maximum cost. */
	wcet1 = find_component("OS_Sched")->wcet;
	wcet2 = find_component("OSIntExit")->wcet;

	if (wcet1 > wcet2) {
		context_switch_cost = wcet1;
	} else {
		context_switch_cost = wcet2;
	}

	return context_switch_cost;
#endif
}

void compute_wcrt(int t) {
	int i, k;
	int set;	int wcrt_old, wcrt;
	int *total_ageing = malloc(sizeof(int)*nsets_dl1);
	int *data_crpd = malloc(sizeof(int)*nsets_dl1);
	int *num_preempt = malloc(sizeof(int) * ntasks);
	int num_preempt2[ntasks][ntasks];
	int preempt_cost;
	
	memset(num_preempt, 0, sizeof(int) * ntasks);
	wcrt = tasks[t]->wcet;
	
	printf("Computing WCRT of %s %i (\"%s\")...\n", get_component_name(t), t, tasks[t]->proc->name);
	printf("WCET of %s %i = %i\n", get_component_name(t), t, tasks[t]->wcet);

	do {
		wcrt_old = wcrt;

		//printf("Iterating... WCRT of task %i = %i\n", t, wcrt_old);
		
		wcrt = tasks[t]->wcet;

		//memset(total_ageing, 0, sizeof(int) * X);
		memset(total_ageing, 0, sizeof(int) * nsets_dl1);
		memset(data_crpd, 0, sizeof(int) * nsets_dl1);

		/* Higher priority tasks (lower value is higher priority)
		 * add preemption costs to task being analyzed */
		for (i=0; i<ntasks; i++) {
			if (tasks[i]->priority >= tasks[t]->priority) {
				continue;
			}

			//printf("Task %i preempts!\n", i);
			num_preempt[i] = CEIL((double) wcrt_old/tasks[i]->period);

			// account for task preemption delay + instruction cache delay + pipeline delay
			preempt_cost = tasks[i]->wcet + tasks[t]->crpd[i] + pipeline_delay_cost();
			// account for ISR cost (for interrupt) or context switch cost (for task preemption)
			if (tasks[i]->is_interrupt) {
				preempt_cost += ISR_cost;
			} else {
				preempt_cost += context_switch_cost();
			}

			wcrt += num_preempt[i] * preempt_cost;

			fprintf(fTaskStat, "WCRT += %i (num_preempt[%i] = %i, preempt_cost = %i, tasks[t]->crpd[i]=%i)\n",
					num_preempt[i] * preempt_cost, i, num_preempt[i], preempt_cost, tasks[t]->crpd[i]);

			/* Also need to take into account modification of micro-architectural states due to
			 * indirect nested preemptions.
			 */
			for (k=0; k<ntasks; k++) {
				if (tasks[k]->priority >= tasks[t]->priority || tasks[k]->priority <= tasks[i]->priority) {
					continue;
				}

				num_preempt2[i][k] = CEIL((double) wcrt_old/tasks[k]->period)
						* CEIL((double) tasks[k]->wcrt/tasks[i]->period);
				int preempt_cost2 = tasks[k]->crpd[i] + pipeline_delay_cost();

				//printf("wcrt_old = %d, tasks[k]->period = %d, tasks[k]->wcrt = %d, tasks[i]->period = %d\n",
				//		wcrt_old, tasks[k]->period, tasks[k]->wcrt, tasks[i]->period);

				//printf("Task k = %d, tasks[k]->wcrt = %d, num_preempt2 = %d, preempt_cost2 = %d\n",
				//		k, tasks[k]->wcrt, num_preempt2[i][k], preempt_cost2);

				wcrt += num_preempt2[i][k] * preempt_cost2;
			}

			/* account for data ageing caused by preemption of task t by task i */
			//fprintf(fTaskStat, "total_ageing = ");
			for (set = 0; set < nsets_dl1; set++) {
				total_ageing[set] += tasks[i]->data_damage[set] * num_preempt[i];
				//fprintf(fTaskStat, "%i ", total_ageing[set]);
			}
			//fprintf(fTaskStat, "\n");

		/*	Not needed because syscall are handled by component_t

			if (tasks[t]->is_syscall && tasks[i]->is_interrupt) {
				for (set = 0; set < X; set++)
					tasks[t]->cumulative_data_damage[set] = tasks[t]->data_damage[set] + tasks[i]->data_damage[set] + num_preempt[i];
			}
		*/
		}

		int wcrt_before_ageing = wcrt;

		for (set = 0; set < nsets_dl1; set++) {
			//wcrt += ageing2crpd(tasks[t]->miss_info[set], &total_ageing[set])*L1_MISS_PENALTY;
			data_crpd[set] = ageing2crpd(tasks[t]->miss_info[set], total_ageing[set])*L1_MISS_PENALTY;
			wcrt += data_crpd[set];
		}

		fprintf(fTaskStat, "Data cache preemption delay: ageing2crpd() cost = %i\n", wcrt-wcrt_before_ageing);

		if (wcrt > tasks[t]->period) {
			printf("WCRT exceeds period of task %i!\n", t);
			break;
		}
	} while (wcrt != wcrt_old);

	tasks[t]->wcrt = wcrt;

	/* Print debug information */
	for (i=0; i<ntasks; i++) {
		if (tasks[i]->priority >= tasks[t]->priority) {
			continue;
		}

		printf("%s %i preempts %i times ", get_component_name(i), i, num_preempt[i]);
		printf("\n");
		//printf("I-CRPD = %i, ", tasks[t]->crpd[i]);
		int total_data_crpd = 0;
		for (set = 0; set < nsets_dl1; set++) {
			total_data_crpd += data_crpd[set];
		}
		//printf("D-CRPD = %i)\n", total_data_crpd);

		/*for (k=0; k<ntasks; k++) {
			if (tasks[k]->priority >= tasks[t]->priority || tasks[k]->priority <= tasks[i]->priority) {
				continue;
			}

			printf("    %s %i preempts through %s %i %i times (",
					get_component_name(i), i,
					get_component_name(k), k,
					num_preempt2[i][k]);
			printf("I-CRPD = %i)\n", tasks[k]->crpd[i]);
		}*/
	}

	free(total_ageing);
	free(num_preempt);

	printf("Final WCRT of %s %i (\"%s\") = %i (%.3f ms)\n", get_component_name(t), t, tasks[t]->proc->name, wcrt, (double)wcrt/CPU_FREQ*1000);
	fprintf(fTaskStat, "Final WCRT of task \"%s\" = %i (%.3f ms)\n", tasks[t]->proc->name, wcrt, (double)wcrt/CPU_FREQ*1000);
}

void dump_cache_config() {
	printf("\n");
	if (enable_icache) {
		extern int nsets, assoc, bsize;
		extern int nsets_l2, assoc_l2, bsize_l2;
		extern int nsets_dl2, assoc_dl2, bsize_dl2;
		printf("l1 instruction cache\n");
		printf("\tnset:%d assoc:%d bsize:%d\n", nsets, assoc, bsize);
		if (enable_il2cache) {
			printf("l2 instruction cache\n");
			printf("\tnset_l2:%d assoc_l2:%d bsize_l2:%d\n", nsets_l2, assoc_l2,
					bsize_l2);
		} else if (enable_ul2cache) {
			printf("ul2 cache\n");
			extern cache_t cache_l2;
			printf("\tnset_dl2:%d assoc_dl2:%d bsize_dl2:%d\n", nsets_dl2,
					assoc_dl2, bsize_dl2);
			printf("\tnset_dl2:%d assoc_dl2:%d bsize_dl2:%d\n", cache_l2.ns,
					cache_l2.na, cache_l2.ls);

		}
	}
	if (enable_scp_dcache) {
		extern int nsets_dl1, assoc_dl1, bsize_dl1;
		extern int nsets_dl2, assoc_dl2, bsize_dl2;
		printf("l1 data cache\n");
		printf("\tnset_dl1:%d assoc_dl1:%d bsize_dl1:%d\n", nsets_dl1,
				assoc_dl1, bsize_dl1);
		if (enable_scp_dl2cache) {
			printf("l2 data cache\n");
			printf("\tnset_dl2:%d assoc_dl2:%d bsize_dl2:%d\n", nsets_dl2,
					assoc_dl2, bsize_dl2);

		}
	}
}

void dump_component_usage_stats(int task_id) {
	int i;

	//if (tasks[task_id]->is_interrupt)
	//	return;

	set_task(tasks[task_id]);
	fprintf(fTaskStat, "Task \"%s\":\n", tasks[task_id]->proc->name);

	for (i=0; i<num_tcfg_nodes; i++) {
		if (tcfg[i]->bb->called_component != NULL)
			fprintf(fTaskStat, "%s (num_preempt = %i, crpd_cost=%i)\n",
					tcfg[i]->bb->called_component->proc->name,
					tcfg[i]->bb->called_component->num_preempt,
					tcfg[i]->bb->called_component->crpd_cost);
	}

	fprintf(fTaskStat, "\n");
}

int comp (const void * elem1, const void * elem2)
{
    int f = *((int*)elem1);
    int s = *((int*)elem2);
    if (tasks[f]->priority > tasks[s]->priority) return 1;
    if (tasks[f]->priority < tasks[s]->priority) return -1;
    return 0;
}

int main(int argc, char **argv) {
	FILE* f_task_graph;
	int dbg = 1;
	int i;
	int j;
	char fName[256], appDirPath[256], str[256], f_task_graph_name[256];
	char taskStatPath[256];
	
	strcpy(fName, argv[argc - 1]);
	strncpy (appDirPath,fName, strrchr(fName,'/')-fName+1);

	strcpy(taskStatPath, appDirPath);
	strcat(taskStatPath, "task_stat");

	fTaskStat = fopen(taskStatPath, "w");

	if (dbg) {
		printf("\nFile name %s", fName);
		fflush(stdout);
		/*
		printf(
				"\n***NOTICE: you need to manually inline all procedures for address analysis to work");
		printf(
				"\n***NOTICE: for triangular loop, you need to create file <binary file name>.econ and set extra loop conditions to help Chronos recognize them");
		*/
		printf("\n Contrainst format: <type> L1_id L2_id k");
		printf("\n type = \"eql\" : L1's loop bound <= L2's iteration + k");
		printf("\n type = \"inv\" : L1's loop bound <= k - L2's iteration\n");
	}
	if (argc <= 1) {
		fprintf(stderr, "Usage:\n");
		fprintf(stderr, "%s <options> <benchmark>\n", argv[0]);
		exit(1);
	}

	strcpy(str, fName);
	init_isa(str);

	/* read options including (1) actions; (2) processor configuration */
	read_opt(argc, argv);
	
	/* liangyun: read jump table if necessary */
	strcpy(str, fName);
	read_injp(str);

	/* liangyun: read depth table for recursive function */
	strcpy(str, fName);
	read_recursive(str);

	/* vivy: only these steps are needed to build CFG */
	if (strcmp(run_opt, "CFG") == 0) {
		strcpy(str, fName);
		read_code(str);
		build_cfgs();
		run_cfg(str);
		return 0;
	}
	
#if 0
	dump_cache_config();
	exit(0);
#endif
	
	strcpy(str, fName);
	/* load task graph file if available */
	sprintf(f_task_graph_name,"%s.tg",fName);
	f_task_graph = fopen(f_task_graph_name, "r");
	if (f_task_graph) {		
		get_clock_time(&start2);

		printf("Task graph file \"%s\" found\n", f_task_graph_name);

		is_task_graph_given = 1;
		
		read_and_analyze_task_graph(f_task_graph, str);
		
		printf("\n********** WCET computation for all tasks/ISRs **********\n");
		fprintf(fTaskStat, "\n-------------- WCET computation --------------\n");

		for (i = 0; i < ntasks; i++) {
			set_task(tasks[i]);

			get_clock_time(&start);

			tasks[i]->wcet = run_est(fName, i);

			get_clock_time(&end);
			tasks[i]->wcet_analysis_time += (end.tv_sec+(double)end.tv_nsec/1000000000
					- start.tv_sec+(double)start.tv_nsec/1000000000);
			fprintf(fTaskStat, "\"%s\" WCET analysis time = %.2f ms\n\n", tasks[i]->proc->name,
					tasks[i]->wcet_analysis_time);

			tasks[i]->wcrt_analysis_time = tasks[i]->wcet_analysis_time;

			get_clock_time(&start);

			tasks[i]->crpd = malloc(sizeof(int) * ntasks);
			memset(tasks[i]->crpd, 0, sizeof(int) * ntasks);
			for (j = 0; j < ntasks; j++) {
				if (tasks[j]->priority < tasks[i]->priority) {
					tasks[i]->crpd[j] = compute_crpd_fifo(i, j)*L1_MISS_PENALTY;
					//printf("CRPD(task %i preempted by task %i) = %i\n", i, j, tasks[i]->crpd[j]);
				}
			}

			get_clock_time(&end);
			tasks[i]->wcrt_analysis_time += (end.tv_sec+(double)end.tv_nsec/1000000000
					- start.tv_sec+(double)start.tv_nsec/1000000000);
		}
		printf("\n");

		printf("********** Fixed point computation of WCRT **********\n");
		fprintf(fTaskStat, "\n-------------- Fixed point computation of WCRT --------------\n");

		int task_order[ntasks];
		for (i = 0; i < ntasks; i++) {
			task_order[i] = i;
		}

		qsort (task_order, ntasks, sizeof(int), comp);

		printf("ID of task ordered from highest priority to least: ");
		for (i = 0; i < ntasks; i++) {
			printf("%d ", task_order[i]);
		}
		printf("\n");

		for (i = 0; i < ntasks; i++) {
			int tid = task_order[i];

			/* Do not need to compute WCRT of interrupt handlers */
			if (tasks[tid]->is_interrupt) {
				continue;
			}

			get_clock_time(&start);

			compute_wcrt(tid);

			get_clock_time(&end);
			tasks[tid]->wcrt_analysis_time += (end.tv_sec+(double)end.tv_nsec/1000000000
					- start.tv_sec+(double)start.tv_nsec/1000000000);
			fprintf(fTaskStat, "\"%s\" WCRT analysis time = %.2f ms\n\n", tasks[tid]->proc->name,
								tasks[tid]->wcrt_analysis_time);
		}

		get_clock_time(&end2);
		printf("\nWhole analysis time = %.2fs\n", (end2.tv_sec+(double)end2.tv_nsec/1000000000
				- start2.tv_sec+(double)start2.tv_nsec/1000000000));
		fprintf(fTaskStat, "\nWhole analysis time = %.2fs\n", (end2.tv_sec+(double)end2.tv_nsec/1000000000
				- start2.tv_sec+(double)start2.tv_nsec/1000000000));

		fprintf(fTaskStat, "\n-------------- Compositional analysis statistics --------------\n");
		for (i = 0; i < ntasks; i++)
			dump_component_usage_stats(i);
	} 
	/* for non multitasking application that contains 'main' function */
	else {
		read_code_and_build_cfg(str);

		read_and_analyze_one_task(str, -1);
	
		run_est(argv[argc - 1], -1);
	}

	fclose(fTaskStat);

	return 0;
}
