/*******************************************************************************

 *

 * Chronos: A Timing Analyzer for Embedded Software

 * =============================================================================

 * http://www.comp.nus.edu.sg/~rpembed/chronos/

 *

 * Copyright (C) 2005 Xianfeng Li

 *

 * This program is free software; you can redistribute it and/or modify it under

 * the terms of the GNU General Public License as published by the Free Software

 * Foundation; either version 2, or (at your option) any later version.

 *

 * This program is distributed in the hope that it will be useful, but WITHOUT

 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or

 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more

 * details.

 *

 * $Id: tcfg.c,v 1.2 2006/06/24 08:54:57 lixianfe Exp $

 *

 ******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "tcfg.h"
#include "loops.h"
#include "scp_cache.h"
#include "adapter.h"

/* Set flag to display debug printf statements for cache damage */
#if 0
#  define D_DMG(x) x
#else
#  define D_DMG(x) 
#endif

extern prog_t prog;


#if 0
tcfg_node_t **tcfg;

int num_tcfg_nodes = 0, tcfg_size = 0;

tcfg_edge_t **tcfg_edges;

int num_tcfg_edges = 0;

tcfg_nlink_t ***bbi_map;
#endif

extern worklist_p components;
extern char is_compositional, is_analyzing_component;

// return the proc id of a tcfg node (basic block instance, or bbi)

int

bbi_pid(tcfg_node_t *bbi)

{

	return bbi->bb->proc->id;

}

// return the basic block id of a tcfg node

int

bbi_bid(tcfg_node_t *bbi)

{

	return bbi->bb->id;

}

// return 1 if the tcfg node corresponds to a conditional basic block

int

cond_bbi(tcfg_node_t *bbi)

{

	return (bbi->bb->type == CTRL_COND);

}

// return 1 if the tcfg edge corresponds to a back edge in the cfg

int

bbi_backedge(tcfg_edge_t *edge)
{
	return edge->is_backedge;
/*
	tcfg_node_t *src, *dst;
	proc_t *proc1, *proc2;
	
	src = edge->src;
	dst = edge->dst;
	proc1 = src->bb->proc;
	proc2 = dst->bb->proc;

	if ((proc1 == proc2) 
	    && (src->bb->loop_role & LOOP_TAIL) 
	    && (dst->bb->loop_role & LOOP_HEAD)
	   )	
		return 1;
	else
		return 0;*/
}

// return 1 if the tcfg edge goes to the loop entry

int

bbi_is_loopback(tcfg_edge_t *edge, int head, int tail)

{

	if ((edge->dst->bb->id == head)

	&& (BETWEEN(edge->src->bb->id, head, tail))

	&& (edge->dst->bb->proc == edge->src->bb->proc))

		return 1;

	else

		return 0;

}

// return 1 if the tcfg node is return

int

bbi_is_return(tcfg_node_t *bbi)

{

	if ((bbi->bb->out_n == NULL) && (bbi->bb->out_t == NULL))

		return 1;

	else

		return 0;

}

int bbi_is_loop_tail(tcfg_node_t *bbi, loop_t* lp)
{
	int i;
    for (i=0; i<lp->num_tails; i++)
		if (bbi == lp->tails[i])
			return 1;
	return 0;
}


static tcfg_node_t *

new_tcfg_node(cfg_node_t *bb)

{

	tcfg_node_t *bbi;

	bbi = (tcfg_node_t *) calloc(1, sizeof(tcfg_node_t));

	CHECK_MEM(bbi);

	bbi->bb = bb;

	bbi->id = num_tcfg_nodes;
	bbi->marked = 0;
//	bbi->ucb = NULL;
	bbi->insnlist = (insn_t *) calloc (bbi->bb->num_inst, sizeof(insn_t));


	return bbi;

}
/* remove a previously created TCFG edge */
static void delete_tcfg_edge(tcfg_edge_t *edge)

{
	tcfg_edge_t *edge1;
	tcfg_node_t* src, *dst;

	src = edge->src;
	dst = edge->dst;

	if (src->out == edge) {
		src->out = src->out->next_out;
	} else {
		edge1 = src->out;

		while (edge1->next_out != edge)
			edge1 = edge1->next_out;

		edge1->next_out = edge->next_out;
	}

	if (dst->in == edge) {
		dst->in = dst->in->next_in;
	} else {
		edge1 = dst->in;

		while (edge1->next_in != edge)
			edge1 = edge1->next_in;

		edge1->next_in = edge->next_in;
	}

	free(edge);
}
static tcfg_edge_t *

new_tcfg_edge(tcfg_node_t *src, tcfg_node_t *dst, int branch)

{

	tcfg_edge_t *edge, *edge1;

	edge = (tcfg_edge_t *) calloc(1, sizeof(tcfg_edge_t));

	CHECK_MEM(edge);

	edge->src = src;
	edge->dst = dst;
	edge->branch = branch;
	edge->dmg = NULL;
	edge->inst_dmg = NULL;

	if (src->out == NULL
	)

		src->out = edge;

	else {

		edge1 = src->out;

		while (edge1->next_out != NULL)

			edge1 = edge1->next_out;

		edge1->next_out = edge;

	}

	if (dst->in == NULL
	)

		dst->in = edge;

	else {

		edge1 = dst->in;

		while (edge1->next_in != NULL)

			edge1 = edge1->next_in;

		edge1->next_in = edge;

	}

	return edge;

}

/* virtually unroll one loop */
/*****************CAUTION**************/
/* sudiptac: this function assumes that when you unroll loop
 * lp, all its inner loops have been unrolled before this.
 * Therefore, the virtual unrolling of loops must follow in
 * a reverse topological order.*/
static void vu_loop(loop_t* lp) {
	int i, j;
	loop_t* lpp;
	int old_tcfg_nodes = num_tcfg_nodes;
	int nnodes = 0;
	int backedge_found;
	int flag;
	tcfg_node_t** sub_tcfg; /* temp mapping between old and new(duplicated) nodes */
	cfg_node_t* bb;
	tcfg_edge_t* edge;
	tcfg_edge_t* src_out;

	sub_tcfg = (tcfg_node_t **) calloc(num_tcfg_nodes, sizeof(tcfg_node_t *));
	CHECK_MEM(sub_tcfg);

	for (i = 0; i < old_tcfg_nodes; i++) {
		lpp = loop_map[i];

		if (lpp != lp && loop_comm_ances[lp->id][lpp->id] != lp) {
			sub_tcfg[tcfg[i]->id] = tcfg[i];
			continue;
		}
#ifdef _DEBUG_CRPD
		printf("TCFG node %d is included in loop starting at 0x%x\n", i, lp->head->bb->sa);
#endif
		if (num_tcfg_nodes >= tcfg_size) {
			/* tcfg store is full, need to realloc memory for more tcfg nodes */
			tcfg_size += TCFG_STEP_SIZE;
			tcfg = (tcfg_node_t **) realloc(tcfg,
					tcfg_size * sizeof(tcfg_node_t *));
			CHECK_MEM(tcfg);
		}
		/* create new tcfg node */
		sub_tcfg[tcfg[i]->id] = tcfg[num_tcfg_nodes] = new_tcfg_node(
				tcfg[i]->bb);
		/* copy the unconditional branch flag */
		tcfg[num_tcfg_nodes]->is_unconditional = tcfg[i]->is_unconditional;

	        /* sudiptac: copy the address set for data cache analysis */
		tcfg[num_tcfg_nodes]->insnlist = tcfg[i]->insnlist;
		tcfg[num_tcfg_nodes]->d_instlist = tcfg[i]->d_instlist;
		
		num_tcfg_nodes++;
	}

	/* adjust the loop map */
	loop_map = (loop_t **) realloc(loop_map,
			(num_tcfg_nodes) * sizeof(loop_t *));
	CHECK_MEM(loop_map);

	for (i = old_tcfg_nodes; i < num_tcfg_nodes; i++) {
		loop_map[i] = lp;
	}

	/* Create the external edges from the pre-header of the loop */
	for (edge = lp->head->in; edge; edge = edge->next_in) {
		int edge_deleted = 0;
		
		src_out = edge->src->out;

		/* Skip the backedge */
		//if (edge->src == lp->tail)
		//	continue;
		/*backedge_found = 0;
		for (i=0; i<lp->num_tails; i++)
			if (edge->src == lp->tails[i])
				backedge_found = 1;	
		if (backedge_found)
			continue;*/
		if (edge->is_backedge)
			continue;

		while (src_out) {
			if (src_out->dst == lp->head) {
				tcfg_node_t* src_out_dst = src_out->src;
				char src_out_branch = src_out->branch;

				delete_tcfg_edge(src_out);
				edge_deleted = 1;
				
				new_tcfg_edge(src_out_dst, sub_tcfg[lp->head->id],
						src_out_branch);
				//printf("external edge from node %i to node %i\n", src_out_dst->id, sub_tcfg[lp->head->id]->id);
				break;
			}
			src_out = src_out->next_out;
		}
		
		if (edge_deleted)
			break;
	}

	/* Now adjust the edges in and out of newly created TCFG nodes */
	for (i = 0; i < old_tcfg_nodes; i++) {
		tcfg_edge_t *out, *in;
		lpp = loop_map[i];
		flag = 0;
		
		if (lpp != lp && loop_comm_ances[lp->id][lpp->id] != lp)
			continue;
		/* Special handling for the back edge */
		for (out = tcfg[i]->out; out != NULL; out = out->next_out) {
			if (out->is_backedge && out->dst == lp->head) {
				edge = new_tcfg_edge(sub_tcfg[tcfg[i]->id], lp->head, TAKEN);
				edge->is_backedge = 0; // unrolled edge is no longer a back edge
				//printf("BACK edge from node %i to node %i\n", sub_tcfg[tcfg[i]->id]->id, lp->head->id);
					
				/* set unconditional branch flag */
				//sub_tcfg[tcfg[i]->id]->is_unconditional = 1;
				
				/* sudiptac: create an edge to loop exit node */
				if (out->dst != lp->head && loop_map[out->dst->id] != lp) {
					new_tcfg_edge(sub_tcfg[tcfg[i]->id], out->dst, NOT_TAKEN);
					//printf("NT edge from node %i to node %i\n", sub_tcfg[tcfg[i]->id]->id, out->dst->id);
				}
				
				continue;
			}
			
			if (out->dst->id >= old_tcfg_nodes) 
				continue;
			
			edge = new_tcfg_edge(sub_tcfg[tcfg[i]->id], sub_tcfg[out->dst->id],
				out->branch);
			edge->is_backedge = out->is_backedge;
			
			//printf("internal edge from node %i to node %i\n", sub_tcfg[tcfg[i]->id]->id, sub_tcfg[out->dst->id]->id);
		}
	}

	free(sub_tcfg);
}

void set_topological_tcfg() {
	char* visited;
	int i, found;

	visited = (char *) calloc(num_tcfg_nodes, sizeof(char));
	CHECK_MEM(visited);
	topo_tcfg = (int*) calloc(num_tcfg_nodes,sizeof (int));

	/* visit the first node */
	visited[0] = 1;
	num_topo_tcfg = 0;
	topo_tcfg[num_topo_tcfg++] = 0;

	while (1) {
		found = 0;
		for (i = 0; i < num_tcfg_nodes; i++) {
			if (visited[i] == 0 && all_predecessors_visited(tcfg[i], visited)) {
				visited[i] = 1;
				topo_tcfg[num_topo_tcfg++] = tcfg[i]->id;
				found = 1;
			}
		}
		if (!found)
			break;
	}

	free(visited);
}


/* sudiptac: create topologically ordered TCFG loops */
void set_topological_tcfg_loops() {
	char* visited;
	int i, found;

	visited = (char *) calloc(num_tcfg_nodes, sizeof(char));
	CHECK_MEM(visited);

	/* visit the first node */
	visited[0] = 1;

	while (1) {
		found = 0;
		for (i = 0; i < num_tcfg_nodes; i++) {
			if (visited[i] == 0 && all_predecessors_visited(tcfg[i], visited)) {
				visited[i] = 1;
				/* is this a loop ? */
				if (loop_map[i] == NULL
				)
					printf("\nBUGGG! %d %d\n", i, num_tcfg_nodes);
				if (loop_map[i] != loops[0] && loop_map[i]->head == tcfg[i])
					topo_tcfg_loops[num_topo_tcfg_loops++] = loop_map[i]->id;
				found = 1;
			}
		}
		if (!found)
			break;
	}

	free(visited);
}

/* sudiptac : Virtual unrolling (VIVU), adding context information (first-next
 * iteration of loop) for better cache analysis */
/* This process must be followed after TCFG is constructed */
void virtual_unroll() {
	int i, topoidx;

	/* find loop information */
	/*	find_loops();
	 map_bbi_loop();
	 loop_relations();*/

	if (!num_tcfg_nodes) {
		fprintf(stderr, "Create the transformed CFG first\n");
		exit(-1);
	}

	/* topologically sort loops */
	num_topo_tcfg_loops = 0;
	topo_tcfg_loops = (int *) calloc(num_tcfg_loops, sizeof(int));
	CHECK_MEM(topo_tcfg_loops);
	set_topological_tcfg_loops();

//	for (i = num_topo_tcfg_loops - 1; i >= 0; i--) {
//		loop_t* lp = loops[topo_tcfg_loops[i]];
//		printf("topo loop %d: %d\n",i,lp->id);
//	}

	/* go reverse topological */
	for (i = num_tcfg_loops - 2; i >= 0; i--) {
		topoidx = topo_tcfg_loops[i];
#ifdef _DEBUG_CRPD
		printf("Loop id = %d with end address = 0x%x\n", i, loops[i]->tail->bb->sa);
#endif
		vu_loop(loops[topoidx]);
	}
	
	//cleekee: print out TCFG after virtual unrolling to a file
	FILE* ftcfg = fopen("tcfg_unrolled.map", "w");
	dump_tcfg(ftcfg);
	fclose(ftcfg);
	FILE* fvtcfg = fopen("visual-tcfg_unrolled.dot", "w");
	dump_visual_tcfg(fvtcfg);
	fclose(fvtcfg);	
	
	/* free out-dated loop information */
#if 0
	for (i = 0; i < num_tcfg_loops; i++) {
		free(loops[i]);
		free(loop_comm_ances[i]);
	}
	free(loops);
	free(loop_map);
	free(loop_comm_ances);
	free(topo_tcfg_loops);
#endif
}

extern int bdepth;
int test_depth(int pid, int depth);

// create a proc instance (virtual inlining), it may recursively call itself if proc

// calls are encountered in current proc, if this proc is called by some block,

// call_bb is the caller and ret_bb is the block to return to (thus correponding tcfg

// edges should be constructed); otherwise they are NULL

static void

proc_inline(proc_t *proc, tcfg_node_t *call_bbi, tcfg_node_t *ret_bbi,
		int depth)

{

	int i;
	cfg_node_t *bb;
	tcfg_node_t **sub_tcfg;
	tcfg_edge_t *edge;

	sub_tcfg = (tcfg_node_t **) calloc(proc->num_bb + 1, sizeof(tcfg_node_t *));

	CHECK_MEM(sub_tcfg);

	// 1. create tcfg nodes of the corresponding proc & remember them in the map

	for (i = 0; i < proc->num_bb; i++) {

		if (num_tcfg_nodes >= tcfg_size) {

			// tcfg store is full, need to realloc memory for more tcfg nodes

			tcfg_size += 64;

			tcfg = (tcfg_node_t **) realloc(tcfg,
					tcfg_size * sizeof(tcfg_node_t *));

		}
		sub_tcfg[i] = tcfg[num_tcfg_nodes] = new_tcfg_node(&proc->cfg[i]);

		num_tcfg_nodes++;

	}

	// 2. recursively create proc instance for callees

	for (i = 0; i < proc->num_bb; i++) {

		if (proc->cfg[i].type == CTRL_CALL) {
			// cleekee: some callee is deliberately set NULL
			// to avoid calling certain functions
			//if (proc->cfg[i].callee == NULL)
				//continue;

			if (is_compositional && !is_analyzing_component) {
				proc_t *callee = proc->cfg[i].callee;
				char is_callee_system_call = 0;
				worklist_p component;
				for (component = components; component != NULL; component = component->next) {
					component_t *data = (component_t*) (component->data);
					if (data->proc->id == callee->id) {
						is_callee_system_call = 1;
						proc->cfg[i].called_component = data;
						break;
					}
				}
				if (is_callee_system_call) {
					continue; /* cleekee: skip inlining component */
				} else {
					proc->cfg[i].called_component = NULL;
				}
			}

			/* vivy: guard against recursive function
			 * for now, ignore the call
			 * needs to refine later
			 */
			if (proc->cfg[i].callee != proc->cfg[i].proc) {
				proc_inline(proc->cfg[i].callee, sub_tcfg[i], sub_tcfg[i + 1],
						0);
			} else {
				//fprintf( stderr, "Recursive function detected: call to self at proc %d bb %d.\n", proc->cfg[i].proc->id, proc->cfg[i].id );
				if (!bdepth)
					continue;
				if (test_depth(proc->id, depth + 1)) {
					//printf("%d-> ",proc->id);
					proc_inline(proc->cfg[i].callee, sub_tcfg[i],
							sub_tcfg[i + 1], depth + 1);
				} else {
					//fprintf(stderr,"out of depth limit for procedure %d\n",proc->id);
				}
			}
		}
	}

	// 3. create tcfg nodes with the help of the map

	// first, create the edge from the caller, if any, to the first tcfg node

	if (call_bbi != NULL
	)

		new_tcfg_edge(call_bbi, sub_tcfg[0], TAKEN);

	for (i = 0; i < proc->num_bb; i++) {

		// if the block calls a proc, the tcfg edge has been created in the recursive

		// invokation of the callee, thus we do nothing here

		if (proc->cfg[i].type == CTRL_CALL) {
			/* cleekee: since component is not inlined, it must proceed to create edges */
			if (proc->cfg[i].called_component == NULL)
				continue;
		}

		// if the block is return, create an edge from it to ret_bbi if not NULL

		if (proc->cfg[i].type == CTRL_RET) {

			if (ret_bbi != NULL
			)

				new_tcfg_edge(sub_tcfg[i], ret_bbi, TAKEN);

			continue;

		}

		if (proc->cfg[i].type == CTRL_RET_COND) {

			if (ret_bbi != NULL
			)

				new_tcfg_edge(sub_tcfg[i], ret_bbi, TAKEN);

		}

		if (proc->cfg[i].out_n != NULL) {

			bb = proc->cfg[i].out_n->dst;

			edge = new_tcfg_edge(sub_tcfg[i], sub_tcfg[bb->id], NOT_TAKEN);
			edge->is_backedge = proc->cfg[i].out_n->is_backedge;

		}

		if (proc->cfg[i].out_t != NULL) {

			bb = proc->cfg[i].out_t->dst;

			edge = new_tcfg_edge(sub_tcfg[i], sub_tcfg[bb->id], TAKEN);
			edge->is_backedge = proc->cfg[i].out_t->is_backedge;
		}

		if (is_compositional && !is_analyzing_component) {
			if (proc->cfg[i].type == CTRL_CALL && proc->cfg[i].called_component != NULL) {
				//virtual_bypass(edge->src->id, edge->dst->id, proc->cfg[i].called_component->dcache_pers_dmg);
				edge->dmg = proc->cfg[i].called_component->cumulative_data_damage;
				edge->inst_dmg = proc->cfg[i].called_component->cumulative_inst_damage;
			}
		}

	}

	free(sub_tcfg);

}

/* check whether all predecessors are visited */
int all_predecessors_visited(tcfg_node_t* bbi, char* visited) {
	tcfg_edge_t* e;

	for (e = bbi->in; e; e = e->next_in) {
		/* skip back edges */
		if (loop_map[bbi->id] && loop_map[bbi->id] == loop_map[e->src->id]
				&& bbi == loop_map[bbi->id]->head)
			continue;
		if (!visited[e->src->id])
			return 0;
	}

	return 1;
}

/* Generates an array of all tcfg edges by visiting each tcfg node */
void collect_tcfg_edges() {

	int i, edge_id = 0;

	tcfg_edge_t *e;

	num_tcfg_edges = 0;

	for (i = 0; i < num_tcfg_nodes; i++) {

		for (e = tcfg[i]->out; e != NULL; e = e->next_out)

			num_tcfg_edges++;

	}

	tcfg_edges = (tcfg_edge_t **) calloc(num_tcfg_edges, sizeof(tcfg_edge_t *));

	CHECK_MEM(tcfg_edges);

	for (i = 0; i < num_tcfg_nodes; i++) {

		for (e = tcfg[i]->out; e != NULL; e = e->next_out) {

			e->id = edge_id++;

			tcfg_edges[e->id] = e;

		}

	}

}

static void

update_bbi_map(tcfg_node_t *bbi)

{

	tcfg_nlink_t *nl;

	int pid, bid;

	nl = (tcfg_nlink_t *) calloc(1, sizeof(tcfg_nlink_t));

	nl->bbi = bbi;

	pid = bbi_pid(bbi);
	bid = bbi_bid(bbi);

	nl->next = bbi_map[pid][bid];

	bbi_map[pid][bid] = nl;

}

void

build_bbi_map()

{

	int i;

	bbi_map = (tcfg_nlink_t ***) calloc(prog.num_procs,

	sizeof(tcfg_nlink_t **));

	CHECK_MEM(bbi_map);

	for (i = 0; i < prog.num_procs; i++) {

		bbi_map[i] = (tcfg_nlink_t **) calloc(prog.procs[i].num_bb,

		sizeof(tcfg_nlink_t *));

		CHECK_MEM(bbi_map[i]);

	}

	for (i = 0; i < num_tcfg_nodes; i++)

		update_bbi_map(tcfg[i]);

}
// transform the CFGs of the procs into a global flow graph (transformed-CFG)

void

prog_tran(int proc_id)

{

	proc_t *proc;

	FILE *ftcfg;
	num_tcfg_nodes = num_tcfg_edges = num_tcfg_loops = tcfg_size = 0;
	proc = &prog.procs[proc_id];
	tcfg = (tcfg_node_t **) malloc(64 * sizeof(tcfg_node_t *));
	
	proc_inline(proc, NULL, NULL, 0);

	collect_tcfg_edges();
	/* sudiptac :::: virtual unrolling, only if cache is enabled */
//	extern int enable_icache;
//	if (enable_icache) {
//		virtual_unroll();
//		/* recollect the tcfg edges as the transformed CFG has been
//		 * augmented after virtual unrolling */
//		collect_tcfg_edges();
//	}
	build_bbi_map();

	ftcfg = fopen("tcfg-no-unroll.map", "w");
	dump_tcfg(ftcfg);
	fclose(ftcfg);
}

void

clear_bbi_flags()

{

	int i;

	for (i = 0; i < num_tcfg_nodes; i++)

		tcfg[i]->flags = 0;

}

void

clear_tcfg_edge_flags()

{

	int i;

	for (i = 0; i < num_tcfg_edges; i++)

		tcfg_edges[i]->flags = 0;

}

int

bbi_type(tcfg_node_t *bbi)

{

	return bbi->bb->type;

}

void

dump_tcfg(FILE *fp)

{

	tcfg_node_t *bbi;

	tcfg_edge_t *edge;

	int i;

	fprintf(fp, "dump tcfg...\n");

	for (i = 0; i < num_tcfg_nodes; i++) {

		bbi = tcfg[i];
		if (!bbi)
			continue;

		//printf("dump block %u at addr 0x%x\n", bbi->id, bbi->bb->code->addr);
		fprintf(fp, "%d(%d.%d): [ ", bbi->id, bbi_pid(bbi), bbi_bid(bbi));

		for (edge = bbi->out; edge != NULL; edge = edge->next_out) {

			fprintf(fp, "%d ", edge->dst->id);

		}

		fprintf(fp, "]\n");

	}

	fprintf(fp, "\n");

}

void dump_visual_tcfg(FILE *fp) {
	tcfg_node_t *bbi;
	tcfg_edge_t *edge;
	int i;

	fprintf(fp, "digraph {\n");
	
	for (i = 0; i < num_tcfg_nodes; i++) {
		bbi = tcfg[i];
		if (!bbi)
			continue;

		for (edge = bbi->out; edge != NULL; edge = edge->next_out) {
			fprintf(fp, "\"%d (%d.%d)\" -> ", bbi->id, bbi_pid(bbi), bbi_bid(bbi));
			fprintf(fp, "\"%d (%d.%d)\";\n", edge->dst->id, bbi_pid(edge->dst), bbi_bid(edge->dst));
		}
	}
	
	fprintf(fp, "\n}");
}

int strictly_dominates(adnode_p n1, adnode_p n2) {
    return(bs_test(n2->dom_in, adnode_id(n1)));
}

int dominates(adnode_p n1, adnode_p n2) {
    return ((n1 == n2) || strictly_dominates(n1, n2));
}

int strictly_postdominates(adnode_p n1, adnode_p n2) {
    return(bs_test(n2->postdom_in, adnode_id(n1)));
}

int postdominates(adnode_p n1, adnode_p n2) {
    return ((n1 == n2) || strictly_postdominates(n1, n2));
}



/*
 * Compute dominance. 
 * TODO: This algorithm is slow. Need to use algorithm from http://dl.acm.org/citation.cfm?id=357071&dl=ACM&coll=DL&CFID=273379941&CFTOKEN=20777842
 */
void compute_dominance(adapter_p ad) {
    int change_flag = 1;
    int i,j;
    int topoidx;
	aditer_p iter;
    tcfg_edge_t *edge;
    bitset_p changetest;
    changetest = bs_alloc(adapter_numnodes(ad));

	for (iter = adapter_iter(ad); !aditer_ended(iter); iter = aditer_next(iter)) {
		adnode_p adnode = aditer_value(iter);
        adnode->dom_in = bs_alloc(adapter_numnodes(ad));
        adnode->dom_out = bs_alloc(adapter_numnodes(ad));
        D_DMG(printf("ALLOC: %u\n", adapter_numnodes(ad)));
        if (adapter_is_entry(adnode)) {
            bs_empty(adnode->dom_in);
        } else {
            bs_fill(adnode->dom_in);
        }
        bs_fill(adnode->dom_out);
    }
    while (change_flag) {
        change_flag = 0;
		for (iter = adapter_iter(ad); !aditer_ended(iter); iter = aditer_next(iter)) {
			adnode_p adnode = aditer_value(iter);
			aditer_p iter2;
			D_DMG(printf("traite bb %u\n", adnode_id(adnode)));
            
            /* merge predecessors */ 
			for (iter2 = adnode_inedge(adnode); !aditer_ended(iter2); iter2 = aditer_next(iter2)) {
				D_DMG(printf("merge pred: %u\n", adnode_id(aditer_value(iter2))));
				assert(adapter_exists_node(ad, adnode_id(aditer_value(iter2))));
				D_DMG(printf("%u\n", adnode->dom_in->numbits));
				D_DMG(printf("%x\n", aditer_value(iter2)));
                bs_inter(adnode->dom_in, aditer_value(iter2)->dom_out);
            }
			bs_print(changetest);
            bs_copy(changetest, adnode->dom_out);
            bs_copy(adnode->dom_out, adnode->dom_in);
            bs_add(adnode->dom_out, adnode_id(adnode));
			bs_print(adnode->dom_out);
			bs_print(changetest);
            
            change_flag |= !bs_equals(changetest, adnode->dom_out);
			D_DMG(printf("changeflag: %u\n", change_flag));
            
        }
	}
}

void compute_postdominance(adapter_p ad) {
    int change_flag = 1;
    int i,j;
    int topoidx;
	aditer_p iter;
    tcfg_edge_t *edge;
    bitset_p changetest;
    changetest = bs_alloc(adapter_numnodes(ad));

    D_DMG(printf("ajoute debut.\n"));
    fflush(stdout);
	for (iter = adapter_iter(ad); !aditer_ended(iter); iter = aditer_next(iter)) {
		adnode_p adnode = aditer_value(iter);
        adnode->postdom_in = bs_alloc(adapter_numnodes(ad));
        adnode->postdom_out = bs_alloc(adapter_numnodes(ad));
        D_DMG(printf("ALLOC (POSTDOM): %u\n", adapter_numnodes(ad)));
        D_DMG(printf("Ajoute pour: %u\n", adnode_id(adnode)));
        fflush(stdout);
        if (adapter_is_exit(adnode)) {
            bs_empty(adnode->postdom_in);
        } else {
            bs_fill(adnode->postdom_in);
        }
        bs_fill(adnode->postdom_out);
    }
    while (change_flag) {
        change_flag = 0;
		for (iter = adapter_reviter(ad); !aditer_ended(iter); iter = aditer_next(iter)) {
			adnode_p adnode = aditer_value(iter);
			aditer_p iter2;
			D_DMG(printf("traite bb %u\n", adnode_id(adnode)));
            
            /* merge predecessors */ 
			for (iter2 = adnode_outedge(adnode); !aditer_ended(iter2); iter2 = aditer_next(iter2)) {
				D_DMG(printf("merge succ: %u\n", adnode_id(aditer_value(iter2))));
                bs_inter(adnode->postdom_in, aditer_value(iter2)->postdom_out);
            }
            
            bs_copy(changetest, adnode->postdom_out);
            bs_copy(adnode->postdom_out, adnode->postdom_in);
            bs_add(adnode->postdom_out, adnode_id(adnode));
			bs_print(adnode->postdom_out);
            
            change_flag |= !bs_equals(changetest, adnode->postdom_out);
            
        }
	}
}




void print_dominance(adapter_p ad) {
	aditer_p iter1, iter2;
    for (iter1 = adapter_iter(ad); !aditer_ended(iter1) ; iter1 = aditer_next(iter1)) {
        printf("Node %u, liste des dominateurs = ", adnode_id(aditer_value(iter1)));
        for (iter2 = adapter_iter(ad); !aditer_ended(iter2) ; iter2 = aditer_next(iter2)) {
            if (strictly_dominates(aditer_value(iter2), aditer_value(iter1))) {
              printf("%u ", adnode_id(aditer_value(iter2)));
            }
        }
        printf("\n");

    }
}

void print_postdominance(adapter_p ad) {
	aditer_p iter1, iter2;
    for (iter1 = adapter_iter(ad); !aditer_ended(iter1) ; iter1 = aditer_next(iter1)) {
        printf("Node %u, liste des postdominateurs = ", adnode_id(aditer_value(iter1)));
        for (iter2 = adapter_iter(ad); !aditer_ended(iter2) ; iter2 = aditer_next(iter2)) {
            if (strictly_postdominates(aditer_value(iter2), aditer_value(iter1))) {
              printf("%u ", adnode_id(aditer_value(iter2)));
            }
        }
        printf("\n");

    }
}




int is_cycle_equiv(adnode_p n1, adnode_p n2) {
  return ((n1 == n2) || (dominates(n1, n2) && postdominates(n2, n1)) || (dominates(n2, n1) && postdominates(n1, n2)));
}


/*
 * Compute cycle-equivalence. Two nodes A and B are cycle-equivalent iff one cycle in the CFG which passes thru node A also passes thru node B
 * In this particular case, it is equivalent to ((A dom B) && (B postdom A)), because we work on graphs where all inner loops have been represented by a single node.
 * Computes a number in node->equiv, who represents class number ( == is equal for all nodes in the same equivalence class)
 */
void compute_cycle_equiv(adapter_p ad) {
  int i,j;
  aditer_p iter1, iter2;
  int next_class = 0;
  for (iter1 = adapter_iter(ad); !aditer_ended(iter1) ; iter1 = aditer_next(iter1)) {
    aditer_value(iter1)->equiv = -1;
  }
  for (iter1 = adapter_iter(ad); !aditer_ended(iter1) ; iter1 = aditer_next(iter1)) {
    if (aditer_value(iter1)->equiv == -1) {
      aditer_value(iter1)->equiv = next_class;
      next_class++;
      for (iter2 = adapter_iter(ad); !aditer_ended(iter2); iter2 = aditer_next(iter2)) {
        if (is_cycle_equiv(aditer_value(iter1), aditer_value(iter2)))
          aditer_value(iter2)->equiv = aditer_value(iter1)->equiv;
      }
    }
  }
}


void print_cycle_equiv(adapter_p ad) {
  int numclass = 0;
  aditer_p iter;
  int i,j;
  for (iter = adapter_iter(ad); !aditer_ended(iter) ; iter = aditer_next(iter)) {
    if (aditer_value(iter)->equiv > numclass)
      numclass = aditer_value(iter)->equiv;
      printf("Pour noeud %u classe %u\n", adnode_id(aditer_value(iter)), aditer_value(iter)->equiv);
  }
  numclass++;
  for (i = 0; i < numclass; i++) {
    printf("Class %u : ", i);
    for (iter = adapter_iter(ad); !aditer_ended(iter) ; iter = aditer_next(iter)) {
      if (aditer_value(iter)->equiv == i)
        printf("%u ", adnode_id(aditer_value(iter)));
    }
    printf("\n");
    
  }
}


#define N_NULL 0
#define N_MAX 1
#define N_PLUS 2
#define N_LEAF 3
#define N_LOOP 4

char *NODE_TYPE[] = {
	"NULL",
	"MAX",
	"PLUS",
	"LEAF",
	"LOOP",
};
struct tree_t {
  int type;
  int value; /* for N_LEAF: bb number, for N_PLUS: equiv class, for ther types: 0 */
#define ST_NOT_VISITED 0 
#define ST_VISITING 1
#define ST_VISITED 2 
  int visitstate;
  int loop_id;
  worklist_p succ;
  worklist_p ageing; /* of ageset_p */
  int is_endloop;
}; 

typedef struct tree_t tree_s;
typedef struct tree_t *tree_p;


tree_p build_node(int type, int value, int loop_id) { 
  tree_p tmp;
  tmp = (void*) malloc(sizeof(tree_s));
  tmp->succ = NULL;
  tmp->ageing = NULL;
  tmp->type = type;
  tmp->value = value;
  tmp->visitstate = 0;
  tmp->loop_id = loop_id;
  return tmp;
}


void dump_node(tree_p node) {
	worklist_p wl;
	fprintf(stderr, "l%u_%s_%u ", node->loop_id, NODE_TYPE[node->type], node->value);
	switch(node->type) {
		case N_NULL:
			fprintf(stderr, "[label=\"\"]"); 
			break;
		case N_MAX:
			fprintf(stderr, "[label=\"M\"]"); 
			break;
		case N_LOOP:
			fprintf(stderr, "[shape=diamond] [label=\"L%u\"]", node->value); 
			break;
		case N_PLUS:
			fprintf(stderr, "[label=\"+\"]"); 
			break;
		case N_LEAF:
			fprintf(stderr, "[shape=box] [label=\"%u\"]", node->value); 
			break;
	}
	for (wl = node->succ; wl; wl = wl->next) {
		tree_p subnode = (tree_p) wl->data;
		fprintf(stderr, " l%u_%s_%u -> l%u_%s_%u\n", node->loop_id, NODE_TYPE[node->type], node->value, subnode->loop_id, NODE_TYPE[subnode->type], subnode->value);
		dump_node(subnode);
	}
}
void dump_tree(tree_p *tloops, int loop_id) {
	fprintf(stderr, "digraph tree {\n");
	dump_node(tloops[loop_id]);
	fprintf(stderr, "}");
/*
  int i;
  worklist_p wl;

  for (i = 0; i < numclass; i++) {
    printf("N_PLUS %u\n", equiv_nodes[i]->value);
    assert(equiv_nodes[i]->type == N_PLUS);
	for (wl = equiv_nodes[i]->succ ; wl ; wl = wl->next) {
		tree_p s = (tree_p) wl->data;
		switch(s->type) {
			case N_PLUS:
				abort();
			case N_MAX:
				printf("...N_MAX\n");
				break;
			case N_NULL:
				printf("...N_NULL\n");
				break;
			case N_LEAF:
				printf("...N_LEAF %u\n", s->value);
				break;
		}
	}
  }
  */
}




int build_tree_visit(tree_p *equiv_nodes, adapter_p ad, tree_p node, int max_id, int loop_id) {
	int need_null;
	aditer_p iter,iter2;
	tree_p subnode;
	tcfg_edge_t *edge;
	node->visitstate = ST_VISITING;


	for (iter = adapter_iter(ad); !aditer_ended(iter) ; iter = aditer_next(iter)) {
		if (aditer_value(iter)->equiv != node->value)
			continue;

		if (aditer_value(iter)->type == ADNODE_TYPE_LOOP) {
			subnode = build_node(N_LOOP, aditer_value(iter)->loop_id, loop_id);
			addToWorkList(&node->succ, subnode);
		} else if (aditer_value(iter)->type == ADNODE_TYPE_NODE) {
			subnode = build_node(N_LEAF, adnode_id(aditer_value(iter)), loop_id);
			addToWorkList(&node->succ, subnode);
		}
		subnode = NULL;
		need_null = 0;
		for (iter2 = adnode_outedge(aditer_value(iter)); !aditer_ended(iter2); iter2 = aditer_next(iter2)) {
			switch(equiv_nodes[aditer_value(iter2)->equiv]->visitstate) {
				case ST_NOT_VISITED:
					max_id = build_tree_visit(equiv_nodes, ad, equiv_nodes[aditer_value(iter2)->equiv], max_id, loop_id);
				case ST_VISITED:
					if (!subnode) {
						subnode = build_node(N_MAX, max_id, loop_id);
						max_id++;
						addToWorkList(&node->succ, subnode);
					}
					addToWorkList(&subnode->succ, equiv_nodes[aditer_value(iter2)->equiv]);
					break;
				case ST_VISITING:
					need_null = 1;
					break;
				default:
					abort();
					break;
			}
			if (subnode && need_null && 0) {
				addToWorkList(&subnode->succ, build_node(N_NULL, max_id, loop_id));
				max_id++;
			}

			if (equiv_nodes[aditer_value(iter2)->equiv]->visitstate) {
			} else {
				addToWorkList(&subnode->succ, equiv_nodes[aditer_value(iter2)->equiv]);
			}
		}	
	}  
	node->visitstate = ST_VISITED;
	return max_id;
}


/*
 * Build sub-tree corresponding to loop-id.
 * Each of the tree node is corresponding to a cycle-equivalence class. 
 * Root is the equivalence class of the entry/exit nodes.
 */
tree_p build_tree(adapter_p ad, int loop_id) {
  tcfg_edge_t *edge;
  tree_p node;
  aditer_p iter;
  tree_p *equiv_nodes;
  int numclass = 0;
  tree_p root;

  int i,j;
  int first;
  for (iter = adapter_iter(ad); !aditer_ended(iter) ; iter = aditer_next(iter)) {
    if (aditer_value(iter)->equiv > numclass)
      numclass = aditer_value(iter)->equiv;
  }
  numclass ++;
  equiv_nodes = malloc(sizeof(tree_p) * numclass);
  for (i = 0; i < numclass; i++) {
      //printf("node+ cree pr classe %u\n", i);
	  equiv_nodes[i] = build_node(N_PLUS, i, loop_id);
  }


  root = equiv_nodes[adapter_entry(ad)->equiv];
  build_tree_visit(equiv_nodes,ad, root, 0, loop_id);
  return root;
}

void dump_adapted_graph(adapter_p ad) {
	aditer_p iter; 
	aditer_p iter2;
	int x;
	int y;
	int z;
   for (iter = adapter_iter(ad); !aditer_ended(iter) ; iter = aditer_next(iter)) {
//       printf("node %u, type %u, loopid %u\n", adnode_id(aditer_value(iter)), aditer_value(iter)->type, aditer_value(iter)->loop_id);
       for (iter2 = adnode_inedge(aditer_value(iter)); !aditer_ended(iter2); iter2 = aditer_next(iter2)) {
//         printf("  --> node %u, type %u, loopid %u\n", adnode_id(aditer_value(iter2)), aditer_value(iter2)->type, aditer_value(iter2)->loop_id);
         x = adnode_id(aditer_value(iter2));
         y = aditer_value(iter2)->type;
         z = aditer_value(iter2)->loop_id;
       }  
    }	   
}

/*
 * Build and return subtree associated with this loop-id
 */ 
tree_p compute_loop(int loop_id) {
	//fprintf(stderr, "Compute loop id %u\n", loop_id);
	/*
	 * This makes an adapted (virtual) subgraph containing only the blocks for the current loop, and 
	 * all the inner loops are reduced to a single node.
	 */
        adapter_p ad = new_adapter(tcfg, num_tcfg_nodes, num_tcfg_loops, loop_id);	
        dump_adapted_graph(ad);
        
        /* Compute dominance/etc on the adapted subgraph corresponding to the loop*/ 
        compute_dominance(ad);
        D_DMG(print_dominance(ad));
        compute_postdominance(ad);
        D_DMG(print_postdominance(ad));
        compute_cycle_equiv(ad);
        D_DMG(print_cycle_equiv(ad));
        return build_tree(ad, loop_id);
        
//        dump_tree();
}

void link_loop_trees_aux(tree_p *tloops, tree_p tree) {
	worklist_p wl;
	for (wl = tree->succ; wl; wl = wl->next) {
		tree_p subtree = (tree_p) wl->data;
		link_loop_trees_aux(tloops, subtree);
	}
	if (tree->type == N_LOOP) {
		for (wl = tloops[tree->value]->succ; wl; wl = wl->next) {
			addToWorkList(&tree->succ, wl->data);
		}
	}
}

void link_loop_trees(tree_p *tloops, int loop_id) {
	link_loop_trees_aux(tloops, tloops[loop_id]);
}

extern int X, Y, B, l1, l2; //X: cache-assoc, Y: # cache-set; B: block-size


/*
 * link TS info to tree nodes
 */
void link_temporal_scopes(tree_p tree) {
	worklist_p wl;
	tree->ageing = NULL;
	for (wl = tree->succ; wl; wl = wl->next) {
		tree_p subtree = (tree_p) wl->data;
		link_temporal_scopes(subtree);
	}
	if (tree->type == N_LEAF) {
		int i;
		tcfg_node_t *node = tcfg[tree->value];
		worklist_p addr_in;
		assert(node);
		if (node->address_set == NULL)
			return;
		for (i = 0; i < node->bb->num_d_inst ; i++) {
			if (node->address_set[i] == NULL)
				continue;
			for (addr_in = node->address_set[i]; addr_in; addr_in = addr_in->next) {
				worklist_p tslist;
				saddr_p saddr = addr_in->data;
				//printf("data %x inst %x set=%u\n", saddr->blkAddr, saddr->instAddr, GET_SET(saddr->blkAddr));
				if (saddr->blkAddr < 0)
					continue;
				assert(saddr->blkAddr >= 0);
				ageset_p as = malloc(sizeof(ageset_t));
				as->set = GET_SET(saddr->blkAddr);
				as->amount = 1;
				as->ts_list = saddr->tsList;
				addToWorkList(&tree->ageing, as);
				for (tslist = saddr->tsList; tslist; tslist = tslist->next) {
					ts_p ts = tslist->data;
					//printf("- scope loop %u (%u, %u)\n", ts->loop_id, ts->lw, ts->up);
				}
			}
		}
	}
}

void join_max(int *dst, int *src) {
	if (*dst < *src) {
		*dst = *src;
	}
}
void join_plus(int *dst, int *src) {
	*dst += *src;
}

typedef void (*amount_join_p)(int *, int *);

struct it_s {
	int loop_id;
	int it;
};
typedef struct it_s it_t;
typedef struct it_s* it_p;

void it_push(it_p it_tab, int *idx, int loop_id, int it) {
	it_tab[*idx].loop_id = loop_id;
	it_tab[*idx].it = it;
	(*idx)++;
}

int it_try_lookup(it_p it_tab, int *idx, int loop_id) {
	int i;
	for (i = 0; i < *idx; i++) {
		if (it_tab[i].loop_id == loop_id) {
			return (it_tab[i].it);
		}
	}
	return -1;
}
int it_lookup(it_p it_tab, int *idx, int loop_id) {
	int ret;
	ret = it_try_lookup(it_tab, idx, loop_id);
	assert(ret != -1);
	return(ret);
}

void it_pop(it_p it_tab, int *idx) {
	(*idx)--;
}

void it_print(it_p it_tab, int *idx) {
	int i;
	for (i = 0; i < *idx; i++) {
		printf("%u -> %u, ", it_tab[i].loop_id, it_tab[i].it);
	}
}

/*
 * Compute the bounds (min / max iteration number) for all the temporal scopes corresponding to "current" loop
 * Excludes temporal scopes not intersecting with currently-processed outer loops
 */
void compute_bounds(tree_p tree, it_p it_tab, int *idx, int current, int *lo, int *hi, bitset_p completed) {
	worklist_p wl;
					
	for (wl = tree->ageing; wl; wl = wl->next) {
		int tmp_lo = *lo;
		int tmp_hi = *hi;
		int ignore = 0;
		ageset_p as = wl->data;
		worklist_p ts_list;
		if (!as)
			continue;
		if (bs_test(completed, as->set)) {
			continue;
		}
		for  (ts_list = as->ts_list; ts_list; ts_list = ts_list->next) {
			int it;
			ts_p ts = ts_list->data;
			if (ts->loop_id == 0)
				continue;
			it = it_try_lookup(it_tab, idx, ts->loop_id);
			// exclude ts if current iteration of outer loop ts->loop_id is not within (ts->lw, ts->up)
			if ((it != -1) && (ts->lw < it) && (ts->up < it)) {
				ignore = 1;
				break;
			}
			if (ts->loop_id == current) {
				if (tmp_lo > ts->lw)
					tmp_lo = ts->lw;
				if (tmp_hi < ts->up)
					tmp_hi = ts->up;
			}
		}
		if (!ignore) {
			*lo = tmp_lo;
			*hi = tmp_hi;
		}
	}

	for (wl = tree->succ; wl; wl = wl->next) {
		tree_p subtree = wl->data;
		compute_bounds(subtree, it_tab, idx, current, lo, hi, completed);
	}
}

typedef struct {
        int ageing;
        int misses;
} missinfo_t;

void record_miss(int miss_num, int needed_ageing, worklist_p *miss_list) {
        missinfo_t *mi = ((*miss_list) ? (*miss_list)->data : NULL);
        worklist_p current;
        if (!mi || (miss_num * mi->ageing > mi->misses * needed_ageing)) {
                worklist_p newlist = malloc(sizeof(worklist_s));
                missinfo_t *new_mi = malloc(sizeof(missinfo_t));
                new_mi->misses = miss_num;
                new_mi->ageing = needed_ageing;
                newlist->next = *miss_list;
                newlist->data = new_mi;
                *miss_list = newlist;
                return;
        }
        // *miss_list != NULL
        for (current = *miss_list; current->next ; current = current->next) {
        	mi = current->next->data;
        	if (miss_num * mi->ageing > mi->misses * needed_ageing)
        		break;
        }
        
        missinfo_t *new_mi = malloc(sizeof(missinfo_t));
        new_mi->misses = miss_num;
        new_mi->ageing = needed_ageing;
        worklist_p newlist = malloc(sizeof(worklist_s));
        newlist->next = current->next;
        newlist->data = new_mi;
        current->next = newlist;       
}

int ageing2crpd(worklist_p miss_list, int num_ageing) {
	worklist_p cur;
	int result = 0;
	for (cur = miss_list; cur; cur = cur->next) {
		missinfo_t *mi = cur->data;
		if (mi->ageing > num_ageing)
			break;
		result += mi->misses;
		num_ageing -= mi->ageing;
	}
	
	return result;
}


void process_aux(tree_p tree, int itcount, loop_t *current, worklist_p *miss_list, int *misses, int set, int itnum) {
	worklist_p wl;
	int max_age = 0;
	int i;
	
	
	switch(tree->type) {
		case N_LEAF:
		{
			tcfg_node_t *node = tcfg[tree->value];
			assert(tree->value >= 0);
			assert(tree->value < num_tcfg_nodes);
			worklist_p addr_in;
			assert(node);
			if (node->address_set == NULL)
				return;
			for (i = 0; i < node->bb->num_d_inst ; i++) {
				if (node->address_set[i] == NULL)
					continue;
				for (addr_in = node->address_set[i]; addr_in; addr_in = addr_in->next) {
					saddr_p saddr = addr_in->data;
					
					
					//only take into account blocks for the correct set 
//					assert(saddr->blkAddr == 0x42);
					if (GET_SET(saddr->blkAddr) != set)
						continue;
//					printf("%x %u\n",saddr->psLoop, saddr->max_persist_age);

					//TODO: check inner temporal scope
					
					
					// Record "itcount" miss, iff a preemption causes more than (A - max_persist-age) misses
					//record_miss(itcount, X - saddr->max_persist_age, miss_list);
					//TODO : ajouter uniquement si temporal scope compatible

					//iterate over temporal scopes
					worklist_p tslist = saddr->tsList;
					for (; tslist; tslist = tslist->next) {
						ts_p ts = tslist->data;

						/*TODO: cleekee: Temporarily disable condition check due to:
						 * (1) current always NULL?
						 * (2) temp scope not right.
						 * We assume max damage for now.
						 */
						//if (current && (ts->loop_id == current->id) && (ts->lw <= itnum) && (ts->up >= itnum)) {
							misses[X - saddr->max_persist_age] += 1;
							break;
						//}
						
					}
				}                                                                 				
			}
			break;
		}
		case N_NULL:
			break;
		case N_MAX:
		case N_PLUS:
			for (wl = tree->succ; wl; wl = wl->next) {
				tree_p subtree = wl->data;
				process_aux(subtree, itcount, current, miss_list, misses, set, itnum);
			}
			break;
		case N_LOOP:
			abort();
			break;
		default:
			abort();
			break;
	}
}

void process_persistent_blocks(tree_p tree, int itcount, worklist_p *miss_list, int *misses, int set) {
	worklist_p wl;
	int i,j;
	int was_null = 0;
	int num_miss;
	int required_ageing;
	switch(tree->type) {
		case N_LEAF:
			break;
		case N_NULL:
			break;
		case N_MAX:
		case N_PLUS:
		case N_LOOP:
		{
			int bound = 1;
			if (tree->is_endloop == 0) {
				for (wl = tree->succ; wl; wl = wl->next) {
					tree_p subtree = wl->data;
					process_persistent_blocks(subtree, bound*itcount, miss_list, misses, set); // TODO: chercher le loopcount dans les flowfacts
				}
			} else {
				
				for (i = 0; i < bound; i++) {
					misses = malloc(sizeof(int) * X);
					memset(misses, 0, sizeof(int) * X );
					for (wl = tree->succ; wl; wl = wl->next) {
						tree_p subtree = wl->data;
						process_aux(subtree, itcount, NULL/*loops[subtree->loop_id]*/, miss_list, misses, set, i);
					}
					//record miss in list; 
					num_miss = 0;
					required_ageing = X;
					for (j = 0; j < X; j++) {
						if (misses[j] >0) {
							if (num_miss < misses[j])
								num_miss = misses[j];

							//cleekee:
							//if (required_ageing > j)
							//	required_ageing = j;
							if (required_ageing > j+1)
								required_ageing = j+1;
						}
					}
					
					if (num_miss) {
						assert(num_miss > 0);
						assert(required_ageing > 0);
						assert(required_ageing <= X);
					}

					for (i= 0; i < itcount; i++)  // TODO: test this
						record_miss(num_miss, required_ageing, miss_list);
					
					free(misses);				
				}  
			}
		}
			break;
		default:
			abort();
			break;
	}
}


int mark_endloops(tree_p tree) {
	int marked = 0;
	worklist_p wl;
	switch(tree->type) {
		case N_LEAF:
			break;
		case N_NULL:
			break;
		case N_MAX:
		case N_PLUS:
		case N_LOOP:
			for (wl = tree->succ; wl; wl = wl->next) {
				tree_p subtree = wl->data;
				marked |= mark_endloops(subtree);
			}
	}
	
	tree->is_endloop = 0;
	if (tree->type == N_LOOP) {
		if (!marked) {
			tree->is_endloop = 1;
			marked = 1;
		}
		//printf("Loop %u endloop=%u\n", tree->value, tree->is_endloop);
	}
	return marked;
}

void evaluate_tree(tree_p tree, it_p it_tab, int *idx) {
	worklist_p wl;
	amount_join_p join = NULL;
	/*
	 * The bitset holds sets that are fully aged (>= associativity-number), so to gain time we don't compute anything for them anymore. 
	 */ 
	bitset_p completed = bs_alloc(MAX_CACHE_SET);
	bs_empty(completed);
	for (wl = tree->ageing; wl; wl = wl->next) { 
		ageset_p as = wl->data;
		if (as->amount >= X) {
			bs_add(completed, as->set);
		}  
	}

	ageset_p *done = malloc(sizeof(ageset_p)*MAX_CACHE_SET);
	memset(done, 0, sizeof(ageset_p)*MAX_CACHE_SET);

	it_print(it_tab, idx);
//	printf("Evaluation du noeud: (L=%u,V=%u,T=%s)\n", tree->loop_id, tree->value, NODE_TYPE[tree->type]);
	switch(tree->type) {
		case N_LEAF:
			break;
		case N_NULL:
			break;
		case N_LOOP:
		case N_MAX:
		case N_PLUS:
			join = (tree->type == N_MAX) ? join_max : join_plus;
			{
				for (wl = tree->succ; wl; wl = wl->next) {
					tree_p subtree = wl->data;
					if (tree->type != N_LOOP) {
						//non-loop case: just evaluate child node
						evaluate_tree(subtree, it_tab, idx);
					} else {
							//loop case: try to find lower / higher iteration value for temporal-scopes corresponding to current loop
							int i, lo, hi;
							hi = 0;
							lo = 999999999;
//							printf("ajoute boucle: %u\n", tree->value);
							compute_bounds(subtree, it_tab, idx, tree->value, &lo, &hi, completed);
//							printf("BOUNDS: %d %d\n", lo, hi);

						/*
						 * Evaluate the ageing info, accounting for all iterations of the current loop. 
						 * TODO: correct (i believie), but horribly inefficient. 
						 * Should to detect "duplicate cases" (when two loop iteration triggers the same set of
						 * temporal scopes). Anyway I've not tried to optimize this yet, because the dominance stuff uses up most of the 
						 * computation time. 
						 */					
						for (i = lo; i < hi; i++) {
							printf("iter %u\n", i);
							it_push(it_tab, idx, tree->value, i);
							evaluate_tree(subtree, it_tab, idx);
							it_pop(it_tab, idx);
						}
						D_DMG(printf("enleve boucle: %u\n", tree->value));
					}
					worklist_p al;
//	printf("Cherche les TS pour le sous-noeud: (L=%u,V=%u,T=%s)\n", subtree->loop_id, subtree->value, NODE_TYPE[subtree->type]);
					for (al = subtree->ageing; al; al = al->next) {
						int to_add = 0;
					ageset_p as2 = al->data;

						// search for relevant temporal scope
						worklist_p ts_list;
						to_add = 1;
						if (as2 != NULL) {
							to_add = 1;
							for  (ts_list = as2->ts_list; ts_list; ts_list = ts_list->next) {
								int it;
								ts_p ts = ts_list->data;
								if (ts->loop_id == 0)
									continue;
								if ((tree->type != N_LOOP) || (tree->value != ts->loop_id)) {
									it = it_lookup(it_tab, idx, ts->loop_id);
									if ((ts->lw > it) || (ts->up < it)) {
										to_add = 0;
										break;
									}
								}
							}
						} else abort();

						if (to_add) {
							int i;
							if (!done[as2->set]) {
				//				ts_p newts = malloc(sizeof(struct temporal_scope));
								ageset_p as = malloc(sizeof(ageset_t));
								done[as2->set] = as;
								as->set = as2->set;
								as->amount = 0;
								as->ts_list = NULL; /* TODO/ mettre ce qui est dans le it_tab */ 
								
				//				addToWorkList(&as->ts_list, newts);
								addToWorkList(&tree->ageing, as);
							}
							for (i = 0; i < *idx; i++) {
								if ((tree->type == N_LOOP) && (it_tab[i].loop_id == tree->value))
									continue;
								ts_p ts = malloc(sizeof(struct temporal_scope));
								ts->lw = it_tab[i].it;
								ts->up = it_tab[i].it;
								ts->loop_id = it_tab[i].loop_id;
								addToWorkList(&done[as2->set]->ts_list, ts);
							}
							join(&done[as2->set]->amount, &as2->amount);
							if (done[as2->set]->amount > X) 
								done[as2->set]->amount = X;
						}
					}
				}
			}
			break;
		default:
			abort();
	}
	free(done);
	bs_free(completed);
	D_DMG(printf("back.\n"));
}

damage_p compute_damage(tree_p root) {
	damage_p dmg;
	worklist_p wl;
	//assert(Y == 16);
	dmg = malloc(sizeof(int)*Y);
	memset(dmg, 0, sizeof(int)*Y);
	for (wl = root->ageing; wl; wl = wl->next) {
		ageset_p as = wl->data;
		dmg[as->set] = as->amount;
	}
	return dmg;
}
void print_damage(damage_p dmg) {
	int i;
	printf("Max set: %u\n", Y);
	for (i = 0; i < Y; i++) {
		printf("Set %d, Ageing=%d\n", i, dmg[i]);
	}
}

damage_p load_damage(FILE *f) {
	int i;
	damage_p dmg;
	dmg = malloc(sizeof(int)*Y);
	for (i = 0; i < Y; i++) {
		dmg[i] = 0;
		fscanf(f, "%d", &dmg[i]);
	}
	return dmg;
}

void store_damage(FILE *f, damage_p dmg) {
	int i;
	for (i = 0; i < Y; i++) {
		fprintf(f, "%d\n", dmg[i]);
	}
}


damage_p compute_all(worklist_p **miss_info) {
	damage_p dmg;
	int has_inner;
	tree_p tloops[num_tcfg_loops];
	int idx = 0;
	it_p ittab = malloc(sizeof(it_t)*1024);
	int i;

	//cleekee: debug only, to check temp scope
	/*int j;
	for (i = 0; i < num_tcfg_nodes; i++) {
		tcfg_node_t *node = tcfg[i];
		for (j = 0; j < node->bb->num_d_inst ; j++) {
			worklist_p addr_in;
			for (addr_in = node->address_set[j]; addr_in; addr_in = addr_in->next) {
				worklist_p tslist;
				saddr_p saddr = addr_in->data;

				for (tslist = saddr->tsList; tslist; tslist = tslist->next) {
					ts_p ts = tslist->data;
					dat_inst_t *d_inst = ((dat_inst_t*) (node->d_instlist));
					d_inst += j;
					if (ts->up==0) ts->up+=1;
					printf("TCFG insn=%s, blkaddr=%u, id=%d , inst=%d,  scope loop %u (%u, %u)\n",
							((insn_t*)(d_inst->insn))->addr, saddr->blkAddr, i, j , ts->loop_id, ts->lw, ts->up);
				}
			}
		}
	}*/

	//DATA (tree) damage
	for (i = 0; i < num_tcfg_loops; i++) {
		tloops[i] = compute_loop(i);
	}
	for (i = 0 ; i < num_tcfg_loops; i++) {
		link_loop_trees(tloops, i);
	}
	//dump_tree(tloops, 0);
	link_temporal_scopes(tloops[0]);
	//printf("\nendloop: maintenant.\n");
	mark_endloops(tloops[0]);

	
	worklist_p miss_list = NULL;
	
	*miss_info = malloc(sizeof(worklist_p) * MAX_CACHE_SET);
	for (i = 0; i < MAX_CACHE_SET; i++) {
		(*miss_info)[i] = NULL;
		process_persistent_blocks(tloops[0], 1, &(*miss_info)[i], NULL, i);
	}
	//printf("\napres\n");

	// test
//	ageing2crpd(miss_list, 50);
	


	//fprintf(stderr, "Evaluating damage...\n");
	evaluate_tree(tloops[0], ittab, &idx);
	dmg = compute_damage(tloops[0]);
	FILE *d = fopen("cache.dmg", "w+");
	store_damage(d, dmg);
	fclose(d);
	//fprintf(stderr, "Damage written to cache.dmg\n");
	
	// INSTRUCTION (phase) damage
	
	return dmg;
}

/*
 * Put an edge between src and dst, to bypass some part of the CFG.
 * Remove from the TCFG the nodes that are between "src" and "dst". 
 * NOTE: I've modified some other parts of Chronos to support TCFGs with
 * some removed nodes (tcfg[x] == NULL) 
 */ 

void virtual_bypass(int src, int dst, damage_p dmg) {
	int i;
	bitset_p bs = bs_alloc(num_tcfg_nodes);
	worklist_p wl = NULL;
	worklist_p todel = NULL;
	tcfg_edge_t *cur;
	tcfg_node_t *node_src = tcfg[src];
	tcfg_node_t *node_dst = tcfg[dst];
	tcfg_edge_t *edge = malloc(sizeof(tcfg_edge_t));

	edge->src = node_src;
	edge->dst = node_dst;
	edge->next_out = NULL;
	edge->next_in = NULL;
	edge->flags = 0; /* gni? */
	edge->branch = TAKEN;
	edge->dmg = dmg;

	addToWorkList(&wl, node_src);
	bs_empty(bs);
	bs_add(bs, src);
	bs_add(bs, dst);
	while (wl != NULL) {
		tcfg_node_t *curnode = wl->data;
		worklist_p old = wl;
		wl = wl->next;
		free(old);
		for (cur = curnode->out; cur; cur = cur->next_out) {
			if (!bs_test(bs, cur->dst->id)) {
				addToWorkList(&wl, cur->dst);
				bs_add(bs, cur->dst->id);
				addToWorkList(&todel, cur->dst);
				cur->dst->marked = 1;
			}
		}
	}

//	if (only_mark == 0) {
		for (; todel; todel = todel->next) {
			tcfg_node_t *curnode = todel->data;
			printf("delete node %u (addr %p)\n", curnode->id, curnode);
			tcfg[curnode->id] = NULL;
			for (i = 0; i < num_tcfg_loops; i++) {
				if (loops[i]->head == curnode)

					loops[i]->head = NULL;
			}
			curnode->in = NULL;
			curnode->out = NULL;
			free(curnode);
		}

		node_src->out = edge;
		node_dst->in = edge;
//	}
	
	printf("bypass from %d to %d\n", src, dst);
}



