
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "cache.h"
#include "tcfg.h"
#include "address.h"
#include "common.h"
#include "bpred.h"
#include "loops.h"
#include "task.h"
#include "unicache.h"

#define CINFTY assoc_l2 

/* Set flag to display debug printf statements for cache analysis */
#if 0
#  define D_CACHE(x) x
#else
#  define D_CACHE(x) 
#endif

extern tcfg_node_t** tcfg;
extern int num_tcfg_nodes;
extern prog_t prog;

/* The instruction set */
extern isa_t* isa;
extern int gcd(int a, int b);

extern char repl_policy;
extern int nsets, bsize, assoc;
/* sudiptac ::: adding options for level 1 data cache */
extern int nsets_dl1, bsize_dl1, assoc_dl1, cache_dl1_lat;

/* sudiptac ::: adding options for level 2 cache 
 * (it could be unified or separate instruction cache) */
extern int nsets_l2, bsize_l2, assoc_l2, cache_dl2_lat, cache_il2_lat;
extern int mem_lat[2];
extern int enable_il2cache;
extern int enable_ul2cache;

/* cleekee: bpred info */
extern int bpred_scheme;
extern de_inst_t ***mp_insts;
extern int *num_mp_insts;

static acs_p Difference(acs_p acs, mem_blk_set_t* mem_blk);
static void build_graph(proc_t* proc);
static void change_layout();
static void change_address();
static int count;
static int getShiftedBlock(unsigned int tblock);
static int allocated(mem_blk_set_t* arg);
static int addMapping(mem_blk_set_t* arg, unsigned int naddr);
void apply_instdamage(instdamage_p dmg, fifo_must_t *fifo, acs_p **must);
void apply_instdamage_lru(int *dmgmust, acs_p **must);


#ifdef _DEBUG
static void dumpCacheBB(acs_p** acs, FILE* fp);
int analysis = 0;
int unified = 0;
#endif
int l1_d1_ps = 0;
int l1_i1_ps = 0;
int i1_u1_ps = 0;
int u1_d1_ps = 0;
int opt = 0;
extern ric_p getAddrBaseOffset(de_inst_t* inst, int base, int offset, int opt);
extern ric_p getAddrBaseIndex(de_inst_t* inst, int base, int index, int opt);
int X, Y, B;

/* Error message routine */
void prerr(char* msg) {
	printf("PANIC ***** %s. Exiting now.......\n", msg);
	exit(-1);
}

void printSet(acs_p * acs_set_print) {
  int i;
	mem_blk_set_t* mem_blk_h_print = NULL;
		for (i = 0; i <= CACHE_SET_SIZE; i++) {
			if (acs_set_print != NULL) {
				if (acs_set_print[i] != NULL
					)
					mem_blk_h_print = acs_set_print[i]->mem_blk_h;
				else
					printf("-");
			} else
				printf("N/A");

			while (mem_blk_h_print != NULL) {
				printf("%i ", mem_blk_h_print->block);
				mem_blk_h_print = mem_blk_h_print->next;
			}

			printf("\n");
		}
}


/* cleekee: Print ACS (debugging purpose) - can remove */
void printAcs(acs_p** acs_print) {
	int i, numset;
	acs_p* acs_set_print = NULL;
	mem_blk_set_t* mem_blk_h_print = NULL;

	for (numset = 0; numset < MAX_CACHE_SET; numset++) {
		printf("set %i:\n", numset);
		for (i = 0; i <= CACHE_SET_SIZE; i++) {
			acs_set_print = acs_print[numset];
			if (acs_set_print != NULL) {
				if (acs_set_print[i] != NULL
					)
					mem_blk_h_print = acs_set_print[i]->mem_blk_h;
				else
					printf("-");
			} else
				printf("N/A");

			while (mem_blk_h_print != NULL) {
				printf("%i ", mem_blk_h_print->block);
				mem_blk_h_print = mem_blk_h_print->next;
			}

			printf("\n");
		}
	}
	printf("\n");
}

/* Create a cache set */
static acs_p* makeCacheSet() {
	acs_p* ret;

	/* To accomodate the victim cache block, cache-set size is set
	 * one more than the given parameter */
	ret = (acs_p *) malloc((CACHE_SET_SIZE + 1) * sizeof(acs_p));
	CHECK_MEM(ret);
	memset(ret, 0, (CACHE_SET_SIZE + 1) * sizeof(acs_p));

	return ret;
}

/* Free a set of linked memory blocks */
static void freeMemBlock(mem_blk_set_t* head) {
	mem_blk_set_t* iter;

	if (!head)
		return;

	freeMemBlock(head->next);
	head->next = NULL;

#ifdef _DELETE
	printf("Freeing mem block = %x\n", head);
#endif

	free(head);
}

/* Free an abstract cache line */
static void freeCacheLine(acs_p acl) {
	if (!acl)
		return;

	freeMemBlock(acl->mem_blk_h);
	acl->mem_blk_h = NULL;

	free(acl);
}

/* Free an abstract cache set */
static void freeCacheSet(acs_p* acs) {
	int i;

	if (!acs)
		return;

	for (i = 0; i <= CACHE_SET_SIZE; i++) {
		freeCacheLine(acs[i]);
		acs[i] = NULL;
	}

	free(acs);
}

/* Free an abstract cache state */
static void freeCacheState(acs_p** acs) {
	int i;

	if (!acs)
		return;

	for (i = 0; i < MAX_CACHE_SET; i++) {
		freeCacheSet(acs[i]);
		acs[i] = NULL;
	}

	free(acs);
}

/* Get all memory referenced by this address range */
static mem_blk_set_t* getMemoryBlocks(ric_p addr) {
	mem_blk_set_t* mem_set = NULL;
	mem_blk_set_t* temp;
	int i;
	int prev = -1;
	int count = 0;

	for (i = addr->lower_bound; i <= addr->upper_bound; i += addr->stride) {
		if (prev == GET_MEM(i)
			)
			continue;
		prev = GET_MEM(i);
		temp = (mem_blk_set_t *) malloc(sizeof(mem_blk_set_t));
		CHECK_MEM(temp);

		/* Assume that all addresses are aligned */
		temp->block = prev;
		temp->next = mem_set;
		mem_set = temp;
		count++;
		/* if number of memory blocks for this address set is more than 
		 * the cache size then the cache is flushed */
		if (count > MAX_CACHE_SET * CACHE_SET_SIZE
			)
			break;
		if (!addr->stride)
			break;
	}

	return mem_set;
}

/* Make an empty cache block */
static acs_p makeEmpty() {
	acs_p ret;

	ret = (acs_p) malloc(sizeof(acs_s));
	CHECK_MEM(ret);
	/* Nothing in the cache block */
	ret->mem_blk_h = NULL;

	return ret;
}

/* Returns 1 if the memory block is present in the cache set, and its age is strictly less than the age passed as argument */

static int isBeforeAge(acs_p *acs, int block, int age) {
  mem_blk_set_t *iter;
  int way;
  if (age == 0)
    return 0;
  assert((age > 0) && (age <= CACHE_SET_SIZE));
  if (acs == NULL)
          return 0;
  for (way = 0; way < age; way++) {
  	if (acs[way] == NULL)
  	  continue;
  	for (iter = acs[way]->mem_blk_h; iter; iter = iter->next) {
  		if (iter->block == block)
  			return 1;
	}
  }
  return 0;
}
/* Returns 1 if a particular memory block is present in a given
 * Cache block */
static int isResident(mem_blk_set_t* mem_blk_h, mem_blk_set_t* item) {
	mem_blk_set_t* iter;

	for (iter = mem_blk_h; iter; iter = iter->next) {
		if (iter->block == item->block)
			return 1;
	}

	return 0;
}

/* Check for a given set of memory blocks whether its
 * superset (may not be proper) is present in any of 
 * the cache block */
static int checkForInclusionSingle(acs_p* acs_in, mem_blk_set_t* mem_blk_set) {
	mem_blk_set_t* iter;
	int i;

	for (i = 0; i <= CACHE_SET_SIZE; i++) {
		for (iter = mem_blk_set; iter; iter = iter->next) {
			if (!acs_in[i])
				continue;
			else if (isResident(acs_in[i]->mem_blk_h, iter))
				break;
		}
		if (iter) {
			return i;
		}
	}

	/* oops .. Not found memory block in the cache */
	return -1;
}

/* Count the number of memory blocks present */
static int getCardinality(mem_blk_set_t* mem_blk_set) {
	mem_blk_set_t* iter;
	int count = 0;

	for (iter = mem_blk_set; iter; iter = iter->next)
		count++;

	return count;
}

/* Make a copy of the cache block */
static acs_p makeCopy(acs_p acs_in) {
	acs_p ret;
	int i;
	mem_blk_set_t* iter;

	if (!acs_in)
		return NULL;

	ret = (acs_p) malloc(sizeof(acs_s));
	CHECK_MEM(ret);
	memset(ret, 0, sizeof(acs_s));
	ret->mem_blk_h = NULL;

	for (iter = acs_in->mem_blk_h; iter; iter = iter->next) {
		mem_blk_set_t* temp = (mem_blk_set_t *) malloc(sizeof(mem_blk_set_t));
		CHECK_MEM(temp);
		temp->block = iter->block;
		temp->next = ret->mem_blk_h;
		ret->mem_blk_h = temp;
	}

	return ret;
}

/* Make a cache block from a set of memory blocks */
static acs_p makeCacheBlock(mem_blk_set_t* mem_blk_set) {
	acs_p ret;
	int i;
	mem_blk_set_t* iter;

	if (!mem_blk_set)
		return NULL;

	ret = (acs_p) malloc(sizeof(acs_s));
	CHECK_MEM(ret);
	memset(ret, 0, sizeof(acs_s));
	ret->mem_blk_h = NULL;

	for (iter = mem_blk_set; iter; iter = iter->next) {
		mem_blk_set_t* temp = (mem_blk_set_t *) malloc(sizeof(mem_blk_set_t));
		CHECK_MEM(temp);
		temp->block = iter->block;
		temp->next = ret->mem_blk_h;
		ret->mem_blk_h = temp;
	}

	return ret;
}

/* Set intersection of the contents of two cache blocks */
/* and return the result */
static acs_p Intersect(acs_p acs1, acs_p acs2) {
	acs_p ret;
	int i;
	mem_blk_set_t* iter;

	if (!acs1 || !acs2)
		return NULL;

	ret = makeEmpty();

	for (iter = acs2->mem_blk_h; iter; iter = iter->next) {
		/* If the cache block is present in both of the 
		 * argument cache blocks, then it is present in 
		 * the return cache block. Useful for must 
		 * analysis */
		if (isResident(acs1->mem_blk_h, iter)) {
			mem_blk_set_t* temp = (mem_blk_set_t *) malloc(
					sizeof(mem_blk_set_t));
			CHECK_MEM(temp);
			temp->block = iter->block;
			temp->next = ret->mem_blk_h;
			ret->mem_blk_h = temp;
		}
	}

	return ret;
}

/* Set union of the contents of two cache blocks */
/* and return the result */
static acs_p Union(acs_p acs1, acs_p acs2) {
	acs_p ret;
	int i;
	mem_blk_set_t* iter;

	if (!acs1 && !acs2)
		return NULL;

	/* Copy the first cache block */
	ret = makeCopy(acs1);

	if (!acs2)
		return ret;

	for (iter = acs2->mem_blk_h; iter; iter = iter->next) {
		if (!ret) {
			ret = (acs_p) malloc(sizeof(acs_s));
			CHECK_MEM(ret);
			ret->mem_blk_h = NULL;
		}
		if (!isResident(ret->mem_blk_h, iter)) {
			mem_blk_set_t* temp = (mem_blk_set_t *) malloc(
					sizeof(mem_blk_set_t));
			CHECK_MEM(temp);
			temp->block = iter->block;
			temp->next = ret->mem_blk_h;
			ret->mem_blk_h = temp;
		}
	}

	return ret;
}

/* Set union of one cache block ans a set of memory blocks
 */
static acs_p UnionCacheMem(acs_p acs1, mem_blk_set_t* mem_blk_set) {
	acs_p ret = NULL;
	int i;
	mem_blk_set_t* iter;

	if (!mem_blk_set)
		return acs1;

	/* Copy the first cache block */
	ret = makeCopy(acs1);

	for (iter = mem_blk_set; iter; iter = iter->next) {
		if (!ret) {
			ret = (acs_p) malloc(sizeof(acs_s));
			memset(ret, 0, sizeof(acs_s));
		}
		if (!isResident(ret->mem_blk_h, iter)) {
			mem_blk_set_t* temp = (mem_blk_set_t *) malloc(
					sizeof(mem_blk_set_t));
			CHECK_MEM(temp);
			temp->block = iter->block;
			temp->next = ret->mem_blk_h;
			ret->mem_blk_h = temp;
		}
	}

	return ret;
}

/* check number of conflicting memory blocks with memory block containing 
 * instruction "cinst" inside basic block bbi */
static int checkForConflicts(tcfg_node_t* bbi, de_inst_t* cinst, int age,
		char first_word, int *conflicts) {
	cfg_node_t* bb = bbi->bb;
	de_inst_t* inst = bb->code;
	int prev_blk = -1;
	int blk;
	int cblk;
	int shift = 0;
	//printf("conflict check\n");

	if (first_word)
		cblk = GET_MEM(cinst->addr);
	else
		cblk = GET_MEM(cinst->addr + SIZE_OF_WORD);

	while (inst != cinst) {
		blk = GET_MEM(inst->addr);
		//printf("test instr addr 0x%x\n", inst->addr);

		if (GET_SET(blk) == GET_SET(cblk) && blk != cblk) {
			/* sudiptac: make sure that you do not count multiple 
			 * conflicts from the same memory block */
			if (blk != prev_blk) {
				shift++;
			}
			prev_blk = blk;
		} else if (blk == cblk) {
			age = shift = 0;
			break;
		}

		inst++;
	}

	*conflicts = age + shift;
	//printf("end.\n");
	if (age != -1 && age + shift < CACHE_SET_SIZE
		)
		return 1;

	return 0;
}

/* Check for a given set of memory blocks whether its
 * superset (may not be proper) is present in the entire 
 * cache */
static int checkForOnePresence(acs_p** acs_in, mem_blk_set_t* mem_blk_set) {
	mem_blk_set_t* iter;
	int i, k;

	for (iter = mem_blk_set; iter; iter = iter->next) {
		k = GET_SET(iter->block);
		for (i = 0; i < CACHE_SET_SIZE; i++) {
			if (acs_in[k][i] && isResident(acs_in[k][i]->mem_blk_h, iter))
				return 1;
		}
	}

	return 0;
}

/* Check for a given set of memory blocks whether its
 * superset (may not be proper) is present in the entire 
 * cache */
int checkForPresence(acs_p** acs_in, mem_blk_set_t* mem_blk_set) {
	mem_blk_set_t* iter;
	int i, k;

	for (iter = mem_blk_set; iter; iter = iter->next) {
		k = GET_SET(iter->block);

		for (i = 0; i < CACHE_SET_SIZE; i++) {
			if (acs_in[k] && acs_in[k][i]
					&& isResident(acs_in[k][i]->mem_blk_h, iter))
				break;
		}

		if (i == CACHE_SET_SIZE
			)
			return -1;
	}

	/* return the position of memory block in the corresponding cache set */
	return i;
}

/* Check for a given set of memory blocks whether its
 * superset (may not be proper) is present in the entire 
 * cache */
static int checkForVictim(acs_p** acs_in, mem_blk_set_t* mem_blk_set) {
	mem_blk_set_t* iter;
	int i, k;

	for (iter = mem_blk_set; iter; iter = iter->next) {
		for (i = 0; i < MAX_CACHE_SET; i++) {
			if (acs_in[i] && acs_in[i][PSEUDO]
					&& isResident(acs_in[i][PSEUDO]->mem_blk_h, iter))
				return 1;
		}
	}

	return 0;
}

/* Check for a given set of memory blocks whether its
 * superset (may not be proper) is present in any of 
 * the cache block */
static int checkForInclusion(acs_p* acs_in, mem_blk_set_t* mem_blk_set) {
	mem_blk_set_t* iter;
	int i;

	for (i = 0; i < CACHE_SET_SIZE; i++) {
		for (iter = mem_blk_set; iter; iter = iter->next) {
			if (acs_in[i] && !isResident(acs_in[i]->mem_blk_h, iter))
				break;
		}
		if (!iter)
			return i;
	}

	/* oops .. Not found memory block in the cache */
	return -1;
}

/* Get the memory blocks mapping to the same set */
static mem_blk_set_t* getMemoryBlockOfSet(mem_blk_set_t* mem_blk, int set) {
	mem_blk_set_t* head = NULL;
	mem_blk_set_t* iter;
	int st;

	if (!mem_blk)
		return NULL;

	for (iter = mem_blk; iter; iter = iter->next) {
		st = GET_SET(iter->block);
		if (st == set) {
			mem_blk_set_t* temp = (mem_blk_set_t *) malloc(
					sizeof(mem_blk_set_t));
			CHECK_MEM(temp);
			temp->block = iter->block;
			temp->next = head;
			head = temp;
		}
	}

	return head;
}

/* Take the set difference of a cache line and a memory block */
static acs_p Difference(acs_p acs, mem_blk_set_t* mem_blk) {
	mem_blk_set_t* iter;
	acs_p ret;

	if (!acs)
		return NULL;

	ret = (acs_p) malloc(sizeof(acs_s));
	CHECK_MEM(ret);
	ret->mem_blk_h = NULL;

	for (iter = acs->mem_blk_h; iter; iter = iter->next) {
		if (iter->block != mem_blk->block) {
			mem_blk_set_t* temp = (mem_blk_set_t *) malloc(
					sizeof(mem_blk_set_t));
			CHECK_MEM(temp);
			temp->block = iter->block;
			temp->next = ret->mem_blk_h;
			ret->mem_blk_h = temp;
		}
	}

	return ret;

}


/* Copy an abstract cache state */
static acs_p* copy_cache(acs_p* acs) {
	int i;
	acs_p* dest = NULL;

	if (!acs)
		return NULL;

	/* Allocate memory for the copying destination */
	dest = makeCacheSet();

	for (i = 0; i <= CACHE_SET_SIZE; i++) {
		dest[i] = makeCopy(acs[i]);
	}

	return dest;
}

static acs_p** copy_acs(acs_p **acs) {
  int i;
  acs_p** result = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
  CHECK_MEM(result);
  memset(result, 0, MAX_CACHE_SET * sizeof(acs_p *));
  for (i = 0; i < MAX_CACHE_SET; i++)
    result[i] = copy_cache(acs[i]);
  return result;                 
}



/* abstract cache set update for a singleton memory block */
static acs_p* update_singleton(acs_p* acs, mem_blk_set_t* mem_blk_set, acs_p *must_in) {
	int line = 0;
	int line_in_acs = 0;
	acs_p* ret;
	acs_p cur, cur2;
	int i, j;
	mem_blk_set_t* temp = NULL;

	temp = (mem_blk_set_t *) malloc(sizeof(mem_blk_set_t));
	CHECK_MEM(temp);
	temp->block = mem_blk_set->block;
	temp->next = NULL;

	ret = makeCacheSet();

	
	/* Check if the accessed block was already in the cache. 
	 * If must_in != NULL, then we are currently doing the Persistence analysis.
	 * In that case, to have a sound persistence analysis, we must check the
	 * presence of the block in the MUST acs that was previously stored in must_in.
	 */
//	must_in = NULL;
	
	if (must_in == NULL) {
		/* If must_in == NULL, then we are doing the musto or may analysis, 
		 * and so, we check presence using the ACS from the current analysis */
		must_in = acs;
	}
	
	for (line = 0; line < CACHE_SET_SIZE; line++) {
		/* block is not present in the cache line */
		if (!must_in) {
			line = CACHE_SET_SIZE;
			break;
		}
		/* block is present in the cache line */
		if (must_in[line] && isResident(must_in[line]->mem_blk_h, temp))
			break;
	}

	if (must_in == acs) {
		line_in_acs = line;
	} else {
		for (line_in_acs = 0; line_in_acs < CACHE_SET_SIZE; line_in_acs++) {
			/* block is not present in the cache line */
			if (!acs) {
				line_in_acs = CACHE_SET_SIZE;
				break;
			}
			/* block is present in the cache line_in_acs */
			if (acs[line_in_acs] && isResident(acs[line_in_acs]->mem_blk_h, temp))
				break;
			}
	}
	D_CACHE(printf("accessed block: %u\n", temp->block));
	D_CACHE(printf("MUST: \n"));
	D_CACHE(printSet(must_in));
	if (must_in && must_in[0] && must_in[1] && must_in[0]->mem_blk_h->block && must_in[1]->mem_blk_h && (must_in[0]->mem_blk_h->block == must_in[1]->mem_blk_h->block)) abort();

	D_CACHE(printf("PERS: \n"));
	D_CACHE(printSet(acs));
	D_CACHE(printf("%d %d \n", line_in_acs, line));
	if (line_in_acs > line) {
	abort();
	}
	
	/* The memory block is present in the cache */
	if (line < CACHE_SET_SIZE)
	{
		for (i = 1; i < line; i++) {
			if ((i - 1) == line_in_acs) {
				cur = Difference(acs[i-1], temp);
				ret[i] = makeCopy(cur);
				free(cur);
				cur = NULL;
			} else {
				ret[i] = makeCopy(acs[i - 1]);
			}
		}
		if (line > 0) {
			cur = Difference(acs[line], temp);
			ret[line] = Union(cur, acs[line - 1]);
#ifdef MEM_FREE
			freeCacheLine(cur);
			cur = NULL;
#endif
		}
		for (i = line + 1; i < CACHE_SET_SIZE; i++) {
			ret[i] = makeCopy(acs[i]);
		}
		D_CACHE(printf("rafraichissement du block %u\n", temp->block));
		ret[0] = makeCacheBlock(temp);

		/* For May analysis */
		if (line == 0 && acs[0]) {
			cur = Difference(acs[0], temp);
			ret[1] = Union(cur, ret[1]);
#ifdef MEM_FREE
			freeCacheLine(cur);
			cur = NULL;
#endif
		}
	}

	/* The memory block is not present in the cache */
	else {
		if (acs) {
			for (i = 1; i < CACHE_SET_SIZE; i++) {
				if ((i - 1) == line_in_acs) {
					cur = Difference(acs[i-1], temp);
					ret[i] = makeCopy(cur);
					free(cur);
					cur = NULL;
				} else {
					ret[i] = makeCopy(acs[i - 1]);
				}
			}
		}
		ret[0] = makeCacheBlock(temp);
		D_CACHE(printf("insertion du block %u\n", temp->block));
		if (acs) {
			/* For persistence analysis collect the victim cache 
			 * blocks */
			cur = Difference(acs[PSEUDO], temp);
			cur2 = Difference(acs[CACHE_SET_SIZE - 1], temp);
			ret[PSEUDO] = Union(cur, cur2);
			free(cur);
			free(cur2);
			cur = NULL;
			cur2 = NULL;
		}
	}

#ifdef MEM_FREE
	free(temp);
#endif
	D_CACHE(printf("PERS (retour): \n"));
	D_CACHE(printSet(ret));
	if (ret && ret[0] && ret[1] && ret[0]->mem_blk_h->block && ret[1]->mem_blk_h && (ret[0]->mem_blk_h->block == ret[1]->mem_blk_h->block)) abort();

	return ret;
}

/* Must join of cache analysis */
static acs_p* joinCacheMust(acs_p* acs1, acs_p* arg) {
	acs_p temp = NULL;
	acs_p val = NULL;
	acs_p val2 = NULL;
	int i, j;
	acs_p* acs;
	mem_blk_set_t* iter;

	if (!acs1)
		return copy_cache(arg);

	acs = makeCacheSet();

	/* Do cache join for all the abstract cache sets of a cache */
	for (i = 0; i < CACHE_SET_SIZE; i++) {
		temp = NULL;
		val = NULL;

		/* If one memory block is present in more than one cache
		 * block take the cache block which is older in age. 
		 * Following code take care of this in case of must 
		 * analysis. */
		for (j = i; j >= 0; j--) {
			val = temp;
			val2 = Intersect(acs1[i], arg[j]);
			temp = Union(temp, val2);
#ifdef MEM_FREE
			freeCacheLine(val);
			freeCacheLine(val2);
			val = NULL;
#endif
		}

		for (j = i; j >= 0; j--) {
			val = temp;
			val2 = Intersect(acs1[i], arg[j]);
			temp = Union(temp, val2);
#ifdef MEM_FREE
			freeCacheLine(val);
			freeCacheLine(val2);
			val = NULL;
#endif
		}

		acs[i] = temp;
	}

	return acs;
}

/* May join of cache analysis */
static acs_p* joinCacheMay(acs_p* acs1, acs_p* arg) {
	acs_p temp = NULL;
	acs_p val = NULL;
	int i, j;
	acs_p* acs;
	mem_blk_set_t* iter;

	if (!acs1)
		return copy_cache(arg);

	acs = makeCacheSet();

	for (i = 0; i < CACHE_SET_SIZE; i++) {
		temp = NULL;
		val = NULL;

		/* PRESENT in ACS-0 but not in ACS-1 */
		if (acs1[i]) {
			for (iter = acs1[i]->mem_blk_h; iter; iter = iter->next) {
				for (j = i - 1; j >= 0; j--) {
					if (arg[j] && isResident(arg[j]->mem_blk_h, iter)) {
						break;
					}
				}
				if (j < 0) {
					mem_blk_set_t* cur = (mem_blk_set_t *) malloc(
							sizeof(mem_blk_set_t));
					cur->block = iter->block;
					cur->next = NULL;
					val = temp;
					temp = UnionCacheMem(temp, cur);
#ifdef MEM_FREE
					freeCacheLine(val);
					free(cur);
					cur = NULL;
					val = NULL;
#endif
				}
			}
		}
		/* PRESENT in ACS-1 but not in ACS-0 */
		if (arg[i]) {
			for (iter = arg[i]->mem_blk_h; iter; iter = iter->next) {
				for (j = i - 1; j >= 0; j--) {
					if (acs1[j] && isResident(acs1[j]->mem_blk_h, iter)) {
						break;
					}
				}
				if (j < 0) {
					mem_blk_set_t* cur = (mem_blk_set_t *) malloc(
							sizeof(mem_blk_set_t));
					cur->block = iter->block;
					cur->next = NULL;
					val = temp;
					temp = UnionCacheMem(temp, cur);
#ifdef MEM_FREE
					freeCacheLine(val);
					free(cur);
					val = NULL;
					cur = NULL;
#endif
				}
			}
		}
		acs[i] = temp;
	}

	return acs;
}

/* Persistence join of cache analysis */
static acs_p* joinCachePS(acs_p* acs1, acs_p* arg) {
	acs_p temp = NULL;
	acs_p val = NULL;
	int i, j;
	acs_p* acs;
	mem_blk_set_t* iter;

	if (!acs1)
		return copy_cache(arg);

	acs = makeCacheSet();

	for (i = 0; i <= CACHE_SET_SIZE; i++) {
		temp = NULL;
		val = NULL;

		if (acs1[i]) {
			for (iter = acs1[i]->mem_blk_h; iter; iter = iter->next) {
				for (j = i + 1; j <= CACHE_SET_SIZE; j++) {
					if (arg[j] && isResident(arg[j]->mem_blk_h, iter)) {
						break;
					}
				}
				if (j > CACHE_SET_SIZE)
				{
					mem_blk_set_t* cur = (mem_blk_set_t *) malloc(
							sizeof(mem_blk_set_t));
					cur->block = iter->block;
					cur->next = NULL;
					val = temp;
					temp = UnionCacheMem(temp, cur);
#ifdef MEM_FREE
					freeCacheLine(val);
					free(cur);
					val = NULL;
					cur = NULL;
#endif
				}
			}
		}
		if (arg[i]) {
			for (iter = arg[i]->mem_blk_h; iter; iter = iter->next) {
				for (j = i + 1; j <= CACHE_SET_SIZE; j++) {
					if (acs1[j] && isResident(acs1[j]->mem_blk_h, iter)) {
						break;
					}
				}
				if (j > CACHE_SET_SIZE)
				{
					mem_blk_set_t* cur = (mem_blk_set_t *) malloc(
							sizeof(mem_blk_set_t));
					cur->block = iter->block;
					cur->next = NULL;
					val = temp;
					temp = UnionCacheMem(temp, cur);
#ifdef MEM_FREE
					freeCacheLine(val);
					free(cur);
					val = NULL;
					cur = NULL;
#endif
				}
			}
		}
		acs[i] = temp;
	}

	return acs;
}

/* Join function during data cache update. Depends on the
 * direction of analysis (Must, May and Persistence) */
static acs_p* joinCache(acs_p* acs1, acs_p* arg, ANALYSIS_T type) {
	acs_p temp = NULL;
	acs_p val = NULL;
	int i, j;
	acs_p* acs;
	mem_blk_set_t* iter;

	/* Join operation for for must cache analysis */
	if (type == MUST) {
		return joinCacheMust(acs1, arg);
	}

	/* Join function for MAY analysis */
	else if (type == MAY) {
		return joinCacheMay(acs1, arg);
	}

	/* Join function for persistence analysis */
	else if (type == PERSISTENCE) {
		return joinCachePS(acs1, arg);
	}

	return NULL;
}

/* Instruction cache update function */
static acs_p** update_abs_inst(acs_p** acs_in, unsigned addr, acs_p** must_in) {
	mem_blk_set_t mem_blk, *iter;
	acs_p* temp;
	acs_p* cur;
	acs_p** acs_out;
	int sameset;
	int set, i;

	mem_blk.block = GET_MEM(addr);
	mem_blk.next = NULL;

	acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
	memset(acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));

	if (acs_in) {
		for (i = 0; i < MAX_CACHE_SET; i++)
			acs_out[i] = copy_cache(acs_in[i]);
	}

	/* Each instruction corresponds to two addresses. Update
	 * the instruction cache state accordingly */
	set = GET_SET(mem_blk.block);
	D_CACHE(printf("update singleton %d (set %d)\n", mem_blk.block, set)); 

	if (GET_MEM(addr + SIZE_OF_WORD) != mem_blk.block) {
		sameset = (set == GET_SET(GET_MEM(addr + SIZE_OF_WORD)));
		if (sameset) {
			temp = update_singleton(acs_in ? acs_in[set] : NULL, &mem_blk, must_in ? must_in[set] : NULL);
		} else {
			temp = NULL;
			acs_out[set] = update_singleton(acs_in ? acs_in[set] : NULL, &mem_blk, must_in ? must_in[set] : NULL);
			
		}
		mem_blk.block = GET_MEM(addr + SIZE_OF_WORD);
		set = GET_SET(mem_blk.block);
		cur = acs_out[set];
		D_CACHE(printf("update singleton %d (set %d)\n", mem_blk.block, set)); 
		if (sameset) {
			acs_out[set] = update_singleton(temp, &mem_blk, must_in ? must_in[set] : NULL);
		} else {
			acs_out[set] = update_singleton(acs_in ? acs_in[set] : NULL, &mem_blk, must_in ? must_in[set] : NULL);
		}
	} else   {
		cur = acs_out[set];
		temp = NULL;
		acs_out[set] = update_singleton(acs_in ? acs_in[set] : NULL, &mem_blk, must_in ? must_in[set] : NULL);
	}
	

	/* free up memory */
#ifdef MEM_FREE
	freeCacheSet(temp);
	freeCacheSet(cur);
	temp = cur = NULL;
#endif

//	printAcs(acs_out);
	return acs_out;
}

/* Check whether two abstract cache blocks are same or not */
static int is_same_cache_block(acs_p acs1, acs_p acs2) {
	mem_blk_set_t* iter;

	if (!acs1 && !acs2)
		return 1;
	if (!acs1 || !acs2)
		return 0;

	for (iter = acs1->mem_blk_h; iter; iter = iter->next) {
		if (!isResident(acs2->mem_blk_h, iter))
			return 0;
	}

	if (getCardinality(acs1->mem_blk_h) == getCardinality(acs2->mem_blk_h))
		return 1;

	return 0;
}

/* Check whether two abstract cache states are identical */
static int checkEquality(acs_p* acs1, acs_p* acs2) {
	int i;

	if (!acs1 && !acs2)
		return 1;
	if (!acs1 || !acs2)
		return 0;
	for (i = 0; i <= CACHE_SET_SIZE; i++) {
		if (!acs1[i] && !acs2[i])
			continue;
		else if (!acs1[i] || !acs2[i])
			return 0;
		else if (!is_same_cache_block(acs1[i], acs2[i]))
			return 0;
	}

	return 1;
}

/* Transfer/update function for l2 instruction abstract cache */
static void transforml2InstCacheState(tcfg_node_t* bbi, int* change_flag,
		ANALYSIS_T type, int iteration_count) {
	de_inst_t* inst;
	int n_inst, bbi_n_inst;
	int base, imm, index;
	char* isa_name;
	ric_p addr;
	acs_p** acs_out = NULL;
	acs_p** cur_acs = NULL;
	acs_p** prev_acs;
	acs_p** acs_spec;
	acs_p* cur_acs_set;
	tcfg_node_t* bbi_spec;
	tcfg_edge_t* bbi_in_p;
	int k;
	char is_backedge = 0;

	assert(bbi);
	assert(bbi->bb);
	inst = bbi->bb->code;

	/* save a copy to check the change in abstract cache states */
	acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
	CHECK_MEM(acs_out);
	memset(acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));

	if (!bbi->acs_out) {
		bbi->acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
		CHECK_MEM(bbi->acs_out);
		memset(bbi->acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));
	}

	for (k = 0; k < MAX_CACHE_SET; k++) {
		acs_out[k] = copy_cache(bbi->acs_out[k]);

		if (!bbi->acs_out[k])
			bbi->acs_out[k] = makeCacheSet();
	}
#if 0
	/* cleekee: account for cache changes by mispredicted instructions */
	if (bpred_scheme != NO_BPRED) {
		acs_spec = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
		CHECK_MEM(acs_spec);
		memset(acs_spec, 0, MAX_CACHE_SET * sizeof(acs_p *));

		CHECK_MEM(num_mp_insts);
		CHECK_MEM(mp_insts);

		bbi_in_p = bbi->in;

		/* iterate through every incoming edge*/
		while (bbi_in_p != NULL) {
			/* sudiptac: check whether first iteration, skip the speculation 
			 * through back edge in that case */
			is_backedge =
					((loop_map[bbi->id] == loop_map[bbi_in_p->src->id])
							&& (bbi == loop_map[bbi->id]->head)) ? 1 : 0;
			if (iteration_count == 0 && is_backedge) {
				bbi_in_p = bbi_in_p->next_in;
				continue;
			}

			/*ipdate incoming ACS with mispredicated instructions */
			for (n_inst = 0; n_inst < num_mp_insts[bbi_in_p->id]; n_inst++) {
				bbi_n_inst = 0;

				/* identify tcfg node that contains mp_insts */
				if (!n_inst) {
					if (bbi_in_p == bbi_in_p->src->out)
						bbi_spec = bbi_in_p->next_out->dst;
					else
						bbi_spec = bbi_in_p->src->out->dst;
				}

				/* point bbi_spec to next tcfg node containing mp_insts */
				if (bbi_n_inst >= bbi_spec->bb->num_inst) {
					bbi_spec = bbi_spec->out->dst;
					bbi_n_inst = 0;
				}

				/* Use cache access classification method */
				if (inst_chmc_l1[bbi_spec->id][bbi_n_inst] == ALL_MISS) {
					if (!n_inst)
						cur_acs = update_abs_inst(bbi_in_p->src->acs_out,
								mp_insts[bbi_in_p->id][n_inst]->addr, NULL);
					else
						cur_acs = update_abs_inst(cur_acs,
								mp_insts[bbi_in_p->id][n_inst]->addr, NULL);
				} else if (inst_chmc_l1[bbi_spec->id][bbi_n_inst]
						== NOT_CLASSIFIED
						|| inst_chmc_l1[bbi_spec->id][bbi_n_inst] == PS) {
					if (!n_inst) {
						prev_acs = bbi_in_p->src->acs_out;
						cur_acs = update_abs_inst(bbi_in_p->src->acs_out,
								mp_insts[bbi_in_p->id][n_inst]->addr, NULL);
					} else {
						prev_acs = cur_acs;
						cur_acs = update_abs_inst(cur_acs,
								mp_insts[bbi_in_p->id][n_inst]->addr, NULL);
					}

					for (k = 0; k < MAX_CACHE_SET; k++) {
						cur_acs_set = cur_acs[k];
						cur_acs[k] = joinCache(prev_acs[k], cur_acs[k], type);
#ifdef MEM_FREE
						freeCacheSet(cur_acs_set);
						cur_acs_set = NULL;
#endif
					}
				}

				bbi_n_inst++;
			}

			/* join all incoming ACS as speculated ACS */
			if (cur_acs && num_mp_insts[bbi_in_p->id] > 0) {
				for (k = 0; k < MAX_CACHE_SET; k++) {
					cur_acs_set = acs_spec[k];

					acs_spec[k] = joinCache(acs_spec[k], cur_acs[k], type);
#ifdef MEM_FREE
					freeCacheSet(cur_acs_set);
					cur_acs_set = NULL;
#endif
				}
			}
#ifdef MEM_FREE
			freeCacheState(cur_acs);
			cur_acs = NULL;
#endif
			bbi_in_p = bbi_in_p->next_in;
		}

		/* join speculated ACS with ACS without branch prediction */
		for (k = 0; k < MAX_CACHE_SET; k++) {
			if (!bbi->acs_in || !bbi->acs_in[k])
				continue;

			cur_acs_set = bbi->acs_out[k];

			bbi->acs_out[k] = joinCache(acs_spec[k], bbi->acs_in[k], type);
#ifdef MEM_FREE
			freeCacheSet(cur_acs_set);
			cur_acs_set = NULL;
#endif
		}
#ifdef MEM_FREE
		freeCacheState(acs_spec);
		acs_spec = NULL;
#endif
	}
#endif
	for (n_inst = 0; n_inst < bbi->bb->num_inst; n_inst++) {
#if 0
		isa_name = isa[inst->op_enum].name;

		/* Save a copy to check the equality between the 
		 * previous state and the updated state. This is 
		 * important for the iterative computaion */
		acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
		memset(acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));

		if(!inst->acs_out)
		{
			inst->acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
			memset(inst->acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));
		}

		for(k = 0; k < MAX_CACHE_SET; k++)
		{
			acs_out[k] = copy_cache(inst->acs_out[k]);

			if(!inst->acs_out[k])
			{
				inst->acs_out[k] = makeCacheSet();
			}
		}
#endif

		/* Use cache access classification method */
		if (inst_chmc_l1[bbi->id][n_inst] == ALL_MISS) {
#if 0
			cur_acs = inst->acs_out;
			inst->acs_out = update_abs_inst(inst->acs_in, inst->addr, NULL);
#endif

			cur_acs = bbi->acs_out;

			/* cleekee: bbi->acs_out is already updated if bpred is enabled */
			if (!n_inst && bpred_scheme == NO_BPRED)
				bbi->acs_out = update_abs_inst(bbi->acs_in, inst->addr, NULL);
			else
				bbi->acs_out = update_abs_inst(bbi->acs_out, inst->addr, NULL);

#ifdef MEM_FREE
			freeCacheState(cur_acs);
			cur_acs = NULL;
#endif
		} else if (inst_chmc_l1[bbi->id][n_inst] == NOT_CLASSIFIED
				|| inst_chmc_l1[bbi->id][n_inst] == PS) {
#if 0
			cur_acs = inst->acs_out;
			inst->acs_out = update_abs_inst(inst->acs_in, inst->addr, NULL);
#endif
			cur_acs = bbi->acs_out;

			/* cleekee: bbi->acs_out is already updated if bpred is enabled */
			if (!n_inst && bpred_scheme == NO_BPRED) {
				prev_acs = bbi->acs_in;
				bbi->acs_out = update_abs_inst(bbi->acs_in, inst->addr, NULL);
			} else {
				prev_acs = bbi->acs_out;
				bbi->acs_out = update_abs_inst(bbi->acs_out, inst->addr, NULL);
			}

#if 0
			for(k = 0; k < MAX_CACHE_SET; k++) {
				cur_acs_set = inst->acs_out[k];
				inst->acs_out[k] = joinCache(inst->acs_in[k], inst->acs_out[k], type);
#ifdef MEM_FREE
				freeCacheSet(cur_acs_set);
				cur_acs_set = NULL;
#endif
			}
#endif

			for (k = 0; k < MAX_CACHE_SET; k++) {
				cur_acs_set = bbi->acs_out[k];
				bbi->acs_out[k] = joinCache(prev_acs[k], bbi->acs_out[k], type);
#ifdef MEM_FREE
				freeCacheSet(cur_acs_set);
				cur_acs_set = NULL;
#endif
			}

#ifdef MEM_FREE
			freeCacheState(cur_acs);
			cur_acs = NULL;
#endif
		}
		/* L2 cache is not accessed at all. So no change in abstract cache state, 
		 * just copy the incoming abstract cache state */
		else {
			if (!n_inst) {
				for (k = 0; k < MAX_CACHE_SET; k++) {
					cur_acs_set = bbi->acs_out[k];
					bbi->acs_out[k] = copy_cache(bbi->acs_in[k]);
#ifdef MEM_FREE
					freeCacheSet(cur_acs_set);
					cur_acs_set = NULL;
#endif
				}
			} else {
				/* Do nothing */
			}
		}
#if 0
		else
		{
			for(k = 0; k < MAX_CACHE_SET; k++) {
				cur_acs_set = inst->acs_out[k];
				inst->acs_out[k] = copy_cache(inst->acs_in[k]);
#ifdef MEM_FREE
				freeCacheSet(cur_acs_set);
				cur_acs_set = NULL;
#endif
			}
		}
		/* check whether the abstract cache states change or not */
		for(k = 0; k < MAX_CACHE_SET; k++)
		*change_flag |= !checkEquality(inst->acs_out[k], acs_out[k]);

		/* Linking incoming and outgoing abstract cache states of two 
		 * consecutive basic blocks */
		if(n_inst < bbi->bb->num_inst - 1)
		{
			de_inst_t* next_inst = inst + 1;
			for(k = 0; k < MAX_CACHE_SET; k++)
			next_inst->acs_in[k] = inst->acs_out[k];
			inst++;
		}
#endif
		inst++;
	}

	/* check whether abstract cache states change or not at the end of basic 
	 * block "bbi" */
	for (k = 0; k < MAX_CACHE_SET; k++)
		*change_flag |= !checkEquality(bbi->acs_out[k], acs_out[k]);

	/* sudiptac : free running acs_out */
	freeCacheState(acs_out);
}


static int debug = 0;

/* Transfer/update function for instruction abstract cache */
static void transformInstCacheState(tcfg_node_t* bbi, int* change_flag,
		ANALYSIS_T type, int iteration_count) {
	de_inst_t* inst;
	int n_inst;
	int base, imm, index;
	char* isa_name;
	ric_p addr;
	acs_p** acs_out;
	acs_p** cur_acs;
	acs_p** acs_spec;
	acs_p* cur_acs_set;
	tcfg_edge_t* bbi_in_p;
	int k;
	char is_backedge = 0;

	assert(bbi);
	assert(bbi->bb);
	inst = bbi->bb->code;


	/* save a copy to check changes in abstract state value */
	acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
	CHECK_MEM(acs_out);
	memset(acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));

	if (!bbi->acs_out) {
		bbi->acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
		CHECK_MEM(bbi->acs_out);
		memset(bbi->acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));
	}

	for (k = 0; k < MAX_CACHE_SET; k++) {
		acs_out[k] = copy_cache(bbi->acs_out[k]);

		if (!bbi->acs_out[k]) {
			bbi->acs_out[k] = makeCacheSet();
		}
	}

	/* cleekee: account for cache changes by mispredicted instructions */
	if (bpred_scheme != NO_BPRED) {
		if (type == PERSISTENCE) {
		  printf("[WARNING] Due to unfinished code, PERS analysis may be unsound if branch prediction is enabled. \n");
		}
		acs_spec = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
		CHECK_MEM(acs_spec);
		memset(acs_spec, 0, MAX_CACHE_SET * sizeof(acs_p *));

		CHECK_MEM(num_mp_insts);
		CHECK_MEM(mp_insts);

		bbi_in_p = bbi->in;

		/* iterate through every incoming edge */
		while (bbi_in_p != NULL) {
			/* sudiptac: check whether first iteration, skip the speculation 
			 * through back edge in that case */
			is_backedge =
					((loop_map[bbi->id] == loop_map[bbi_in_p->src->id])
							&& (bbi == loop_map[bbi->id]->head)) ? 1 : 0;
			if (iteration_count == 0 && is_backedge) {
				bbi_in_p = bbi_in_p->next_in;
				continue;
			}
			/* update incoming ACS with mispredicted instructions */
			for (n_inst = 0; n_inst < num_mp_insts[bbi_in_p->id]; n_inst++) {
				if (!n_inst)
					cur_acs = update_abs_inst(bbi_in_p->src->acs_out,
							mp_insts[bbi_in_p->id][n_inst]->addr, NULL);
				else
					cur_acs = update_abs_inst(cur_acs,
							mp_insts[bbi_in_p->id][n_inst]->addr, NULL);
			}

			/* join all incoming ACS as speculated ACS */
			if (num_mp_insts[bbi_in_p->id] > 0) {
				for (k = 0; k < MAX_CACHE_SET; k++) {
					cur_acs_set = acs_spec[k];

					/* sudiptac: sanity check required ? */
					if (!cur_acs || !cur_acs[k])
						continue;

					acs_spec[k] = joinCache(acs_spec[k], cur_acs[k], type);
#ifdef MEM_FREE
					freeCacheSet(cur_acs_set);
					cur_acs_set = NULL;
#endif
				}
			}
#ifdef MEM_FREE
			freeCacheState(cur_acs);
			cur_acs = NULL;
#endif
			bbi_in_p = bbi_in_p->next_in;
		}

		/* join speculated ACS with ACS without branch prediction */
		for (k = 0; k < MAX_CACHE_SET; k++) {
			if (!bbi->acs_in || !bbi->acs_in[k])
				continue;

			cur_acs_set = bbi->acs_out[k];

			bbi->acs_out[k] = joinCache(acs_spec[k], bbi->acs_in[k], type);
#ifdef MEM_FREE
			freeCacheSet(cur_acs_set);
			cur_acs_set = NULL;
#endif
		}
#ifdef MEM_FREE
		freeCacheState(acs_spec);
		acs_spec = NULL;
#endif
	}

	for (n_inst = 0; n_inst < bbi->bb->num_inst; n_inst++) {
		isa_name = isa[inst->op_enum].name;

#if 0
		/* Save a copy to check the equality between the 
		 * previous state and the updated state. This is 
		 * important for the iterative computation */
		acs_out = (acs_p **)malloc(MAX_CACHE_SET * sizeof(acs_p *));
		memset(acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));

		if(!inst->acs_out)
		{
			inst->acs_out = (acs_p **)malloc(MAX_CACHE_SET * sizeof(acs_p *));
			memset(inst->acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));
		}

		for(k = 0; k < MAX_CACHE_SET; k++) {
			acs_out[k] = copy_cache(inst->acs_out[k]);

			if(!inst->acs_out[k])
			{
				inst->acs_out[k] = makeCacheSet();
			}
		}

		cur_acs = inst->acs_out;
		inst->acs_out = update_abs_inst(inst->acs_in, inst->addr, NULL);
#endif
		/* updating cache states through each instruction */
		cur_acs = bbi->acs_out;

		/* cleekee: bbi->acs_out is already updated if bpred is enabled */
		D_CACHE(printf("updateblock: %u, ", bbi->id));

		if ((type == MUST) && (bbi->must_inst[n_inst] != NULL))
			freeCacheState(bbi->must_inst[n_inst]);
		if ((type == MAY) && (bbi->may_inst[n_inst] != NULL))
			freeCacheState(bbi->may_inst[n_inst]);
		if ((type == MAY) && (bbi->may_inst_out[n_inst] != NULL))
			freeCacheState(bbi->may_inst_out[n_inst]);
		if (!n_inst && (bpred_scheme == NO_BPRED)) {
			if (type == MAY) 
				bbi->may_inst[n_inst] = copy_acs(bbi->acs_in);
			if (type == MUST) 
				bbi->must_inst[n_inst] = copy_acs(bbi->acs_in);
			bbi->acs_out = update_abs_inst(bbi->acs_in, inst->addr,  (type==PERSISTENCE) ? bbi->must_inst[n_inst] : NULL);
			if (type == MAY) 
				bbi->may_inst_out[n_inst] = copy_acs(bbi->acs_in);
		} else {
			if (type == MUST)
				bbi->must_inst[n_inst] = copy_acs(bbi->acs_out);
			if (type == MAY)
				bbi->may_inst[n_inst] = copy_acs(bbi->acs_out);
			bbi->acs_out = update_abs_inst(bbi->acs_out, inst->addr,  (type==PERSISTENCE) ? bbi->must_inst[n_inst] : NULL);
			if (type == MAY)
				bbi->may_inst_out[n_inst] = copy_acs(bbi->acs_out);
	        }
#ifdef MEM_FREE
        	freeCacheState(cur_acs);
		cur_acs = NULL;
#endif

#if 0
		/* checking equality of two abstract cache states */
		for(k = 0; k < MAX_CACHE_SET; k++)
		*change_flag |= !checkEquality(inst->acs_out[k], acs_out[k]);

		/* linking incoming and outgoing abstract cache states of two 
		 * consecutive basic blocks */
		if(n_inst < bbi->bb->num_inst - 1)
		{
			de_inst_t* next_inst = inst + 1;

			for(k = 0; k < MAX_CACHE_SET; k++)
			next_inst->acs_in[k] = inst->acs_out[k];

			inst++;
		}
#endif
		inst++;
	}

	/* checking equality between two abstract cache states for this 
	 * basic block */
	for (k = 0; k < MAX_CACHE_SET; k++)
		*change_flag |= !checkEquality(bbi->acs_out[k], acs_out[k]);

	/* free running acs_out */
	freeCacheState(acs_out);
}

/* Perform join operation on two abstract cache states */
/* Join operation depends on the direction of cache analysis 
 * (may, must, persistence). This direction of cache analysis 
 * is provided through the argument "type" */
static void JoinCacheState(tcfg_node_t* pred, tcfg_node_t* bbi, int type, int *lru_damage) {
	de_inst_t* f_inst, *l_inst;
	acs_p* free_p = NULL;
	acs_p **out;
	int i, j, k;

	if (!pred)
		return;
		
	out = copy_acs(pred->acs_out);

	if (lru_damage)
		apply_instdamage_lru(lru_damage, out);

#if 0
	/* First instruction of this basic block */
	f_inst = bbi->bb->code;

	/* Last instruction of predecessor basic block */
	l_inst = pred->bb->code + pred->bb->num_inst - 1;
#endif

	/* Get the abstract cache state at the exit point of the last
	 * instruction of the predecessor basic block */
	for (i = 0; i < MAX_CACHE_SET; i++) {
#if 0
		/* no cache state out of the last instruction of an input basic 
		 * block */
		if(!l_inst->acs_out[i])
		continue;
		free_p = f_inst->acs_in[i];
		f_inst->acs_in[i] = joinCache(f_inst->acs_in[i], l_inst->acs_out[i], type);
#endif

		if (!out || !out[i])
			continue;

		free_p = bbi->acs_in[i];
		if (i == 2) {
			if (bbi->id == 6) {
			debug++;
			if (debug == 8) {
			  //printf("maintenant\n");
			}
			}
			D_CACHE(printf("set2 predecessor acs: \n"));
			D_CACHE(printSet(out[i]));
			D_CACHE(printf("preexisting state: \n"));
			D_CACHE(printSet(bbi->acs_in[i]));
		}
		bbi->acs_in[i] = joinCache(bbi->acs_in[i], out[i], type);
		if (i == 2) {
		D_CACHE(printf("join result for blk %u: \n", bbi->id));
		D_CACHE(printSet(bbi->acs_in[i]));
		}

		/* Free the old abstract cache state */
#ifdef MEM_FREE
		freeCacheSet(free_p);
		free_p = NULL;
#endif
	}

	freeCacheState(out);
}

/* This procedure is called before doing each cache analysis. Cache 
 * analysis is initialized with allocating memory for abstract cache 
 * states. After finishing one analysis and before starting a next 
 * one, this allocated memory should be reclaimed for better memory 
 * management. "cleanupCache" procedure is used for this purpose 
 */
static void initializeCache(tcfg_node_t* bbi) {
	de_inst_t* inst = bbi->bb->code;
	int i, j;

	bbi->acs_in = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
	CHECK_MEM(bbi->acs_in);
	memset(bbi->acs_in, 0, MAX_CACHE_SET * sizeof(acs_p *));
	bbi->acs_out = (acs_p **) malloc(MAX_CACHE_SET * sizeof(acs_p *));
	CHECK_MEM(bbi->acs_out);
	memset(bbi->acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));

	//for (i = 0; i < MAX_CACHE_SET; i++)
	//bbi->acs_in[i] = makeCacheSet();
#if 0		  
	assert(inst);

	for(j = 0; j < bbi->bb->num_inst; j++)
	{
		inst->acs_in = (acs_p **)malloc(MAX_CACHE_SET * sizeof(acs_p *));
		memset(inst->acs_in, 0, MAX_CACHE_SET * sizeof(acs_p *));
		inst->acs_out = (acs_p **)malloc(MAX_CACHE_SET * sizeof(acs_p *));
		memset(inst->acs_out, 0, MAX_CACHE_SET * sizeof(acs_p *));
		inst++;
	}
	inst = bbi->bb->code;
	for(i = 0; i < MAX_CACHE_SET; i++)
	{
		inst->acs_in[i] = makeCacheSet();
	}
#endif

}

/* Allocate memory for storing CHMC information of an instruction */
/* inst_chmc_l1[i][j] = CHMC in L1 cache of j-th instruction at i-th tcfg block */
/* inst_chmc_l2[i][j] = CHMC in L2 cache of j-th instruction at i-th tcfg block */
/* also record the age of the memory block in the abstract must cache. This will 
 * later be used for CRPD analysis and the shared cache conflict analysis */
static void initialize_CHMC() {
	int i, j;

	inst_chmc_l1 = (char **) malloc(num_tcfg_nodes * sizeof(char *));
	CHECK_MEM(inst_chmc_l1);
	inst_age_l1 = (char **) malloc(num_tcfg_nodes * sizeof(char *));
	CHECK_MEM(inst_age_l1);
	inst_chmc_l2 = (char **) malloc(num_tcfg_nodes * sizeof(char *));
	CHECK_MEM(inst_chmc_l2);
	inst_age_l2 = (char **) malloc(num_tcfg_nodes * sizeof(char *));
	CHECK_MEM(inst_age_l2);

	for (i = 0; i < num_tcfg_nodes; i++) {
		if (tcfg[i] == NULL)
			continue;
		inst_chmc_l1[i] = (char *) malloc(tcfg[i]->bb->num_inst * sizeof(char));
		CHECK_MEM(inst_chmc_l1[i]);
		inst_age_l1[i] = (char *) malloc(tcfg[i]->bb->num_inst * sizeof(char));
		CHECK_MEM(inst_age_l1[i]);
		inst_chmc_l2[i] = (char *) malloc(tcfg[i]->bb->num_inst * sizeof(char));
		CHECK_MEM(inst_chmc_l2[i]);
		inst_age_l2[i] = (char *) malloc(tcfg[i]->bb->num_inst * sizeof(char));
		CHECK_MEM(inst_age_l2[i]);

		for (j = 0; j < tcfg[i]->bb->num_inst; j++) {
			//inst_chmc_l1[i][j] = NOT_CLASSIFIED;
			//inst_chmc_l2[i][j] = NOT_CLASSIFIED;
			inst_chmc_l1[i][j] = CINFTY;
			inst_chmc_l2[i][j] = CINFTY;
			inst_age_l1[i][j] = CINFTY;
			inst_age_l2[i][j] = CINFTY;
		}
	}
}

/* For a given instruction addresse classify it as persistence
 * or non-classified (L2 cache) */
static void categorize_l2_inst_X_PS_NC(tcfg_node_t* bbi, de_inst_t* inst,
		int inst_id) {
	mem_blk_set_t temp;
	int h1, h2;

	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

#if 0
	h1 = checkForVictim(inst->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForVictim(inst->acs_in, &temp);
#endif
	h1 = checkForVictim(bbi->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForVictim(bbi->acs_in, &temp);

	if (inst_chmc_l1[bbi->id][inst_id] == ALL_HIT) {
		inst_chmc_l2[bbi->id][inst_id] = ALL_X;
	} else if ((!h1 && !h2)
			&& inst_chmc_l2[bbi->id][inst_id] == NOT_CLASSIFIED) {
		inst_chmc_l2[bbi->id][inst_id] = PS;
	}
	/* Rest are already L2 classified or L2 not classified */
}

/* For a given instruction addresse classify it as all hit
 * or non-classified (L2 cache) */
static void categorize_l2_inst_X_AH_NC(tcfg_node_t* bbi, de_inst_t* inst,
		int inst_id) {
	mem_blk_set_t temp;
	int h1, h2;
	int hc1, hc2;
	int conflicts = 0;

	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

#if 0
	h1 = checkForPresence(inst->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(inst->acs_in, &temp);
#endif

	h1 = checkForPresence(bbi->acs_in, &temp);
	hc1 = checkForConflicts(bbi, inst, h1, 1, &conflicts);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(bbi->acs_in, &temp);
	hc2 = checkForConflicts(bbi, inst, h2, 0, &conflicts);

	if (inst_chmc_l1[bbi->id][inst_id] == ALL_HIT) {
		inst_chmc_l2[bbi->id][inst_id] = ALL_X;

		if (hc1 && hc2)
			inst_age_l2[bbi->id][inst_id] = conflicts;
		else
			inst_age_l2[bbi->id][inst_id] = CINFTY;

	} else if (hc1 && hc2) {
		inst_chmc_l2[bbi->id][inst_id] = ALL_HIT;
		inst_age_l2[bbi->id][inst_id] = conflicts;
	} else {
		inst_chmc_l2[bbi->id][inst_id] = NOT_CLASSIFIED;
		inst_age_l2[bbi->id][inst_id] = CINFTY;
	}
}

/* For a given instruction address classify it as all miss
 * or non-classified (L2 cache) */
static void categorize_l2_inst_X_AM_NC(tcfg_node_t* bbi, de_inst_t* inst,
		int inst_id) {
	mem_blk_set_t temp;
	int h1, h2;
	int hc1, hc2;
	int conflicts = 0;

	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

#if 0
	h1 = checkForPresence(inst->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(inst->acs_in, &temp);
#endif

	h1 = checkForPresence(bbi->acs_in, &temp);
	hc1 = checkForConflicts(bbi, inst, h1, 1, &conflicts);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(bbi->acs_in, &temp);
	hc2 = checkForConflicts(bbi, inst, h2, 0, &conflicts);

	if (inst_chmc_l1[bbi->id][inst_id] == ALL_HIT) {
		inst_chmc_l2[bbi->id][inst_id] = ALL_X;
	} else if ((!hc1 || !hc2)
			&& inst_chmc_l2[bbi->id][inst_id] == NOT_CLASSIFIED) {
		inst_chmc_l2[bbi->id][inst_id] = ALL_MISS;
	}
	/* Rest are already classified or kept non-classified */
}

/* For a given instruction addresse classify it as all hit
 * or non-classified (L1 cache) */
static void categorize_inst_AH_NC(tcfg_node_t* bbi, de_inst_t* inst,
		int inst_id) {
	mem_blk_set_t temp;
	int h1, h2;
	int hc1, hc2;
	int conflicts = 0;

	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

#if 0
	h1 = checkForPresence(inst->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(inst->acs_in, &temp);
#endif

	h1 = checkForPresence(bbi->acs_in, &temp);
	hc1 = checkForConflicts(bbi, inst, h1, 1, &conflicts);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(bbi->acs_in, &temp);
	hc2 = checkForConflicts(bbi, inst, h2, 0, &conflicts);

	if (hc1 && hc2) {
		inst_chmc_l1[bbi->id][inst_id] = ALL_HIT;
		inst_age_l1[bbi->id][inst_id] = conflicts;
	} else {
		inst_chmc_l1[bbi->id][inst_id] = NOT_CLASSIFIED;
		inst_age_l1[bbi->id][inst_id] = CINFTY;
	}
	/* Rest are already non-classified */
}

/* For a given instruction address classify it as persistence
 * or non-classified (L1 cache) */
static void categorize_inst_PS_NC(tcfg_node_t* bbi, de_inst_t* inst,
		int inst_id) {
	mem_blk_set_t temp;
	int h1, h2;

	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

#if 0
	h1 = checkForVictim(inst->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForVictim(inst->acs_in, &temp);
#endif		  
	h1 = checkForVictim(bbi->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForVictim(bbi->acs_in, &temp);

	if (!h1 && !h2 && inst_chmc_l1[bbi->id][inst_id] == NOT_CLASSIFIED) {
		inst_chmc_l1[bbi->id][inst_id] = PS;
	}
	
	/* debugging purpose only */
	switch(inst_chmc_l1[bbi->id][inst_id]) {
	case ALL_HIT:
	  	D_CACHE(printf("alwayshit!\n"));
	  	break;
	case ALL_MISS:
		D_CACHE(printf("alwaysmiss!\n"));
		break;
	case PS:
		D_CACHE(printf("persistent!\n"));
		break;
	case NOT_CLASSIFIED:
		D_CACHE(printf("notclassified!\n"));
		break;
	}

	/* Rest are already L1 classified or kept non-classified */
}

/* For a given instruction address classify it as all miss
 * or non-classified (L1 cache) */
static void categorize_inst_AM_NC(tcfg_node_t* bbi, de_inst_t* inst,
		int inst_id) {
	mem_blk_set_t temp;
	int h1, h2;
	int hc1, hc2;
	int conflicts = 0;

	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

#if 0
	h1 = checkForPresence(inst->acs_in, &temp);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(inst->acs_in, &temp);
#endif

	h1 = checkForPresence(bbi->acs_in, &temp);
	hc1 = checkForConflicts(bbi, inst, h1, 1, &conflicts);
	temp.block = GET_MEM(inst->addr + SIZE_OF_WORD);
	h2 = checkForPresence(bbi->acs_in, &temp);
	hc2 = checkForConflicts(bbi, inst, h2, 0, &conflicts);

	if ((!hc1 || !hc2) && inst_chmc_l1[bbi->id][inst_id] == NOT_CLASSIFIED) {
		inst_chmc_l1[bbi->id][inst_id] = ALL_MISS;
	}
	/* Rest are already L1 classified or kept L1 non-classified */
}

/* Categorize instruction at L2 cache after analysis */
static void categorize_L2(int type) {
	tcfg_node_t* bbi;
	de_inst_t* inst;
	char* isa_name;
	int i, j;

	for (i = 0; i < num_tcfg_nodes; i++) {
		inst = tcfg[i]->bb->code;

		for (j = 0; j < tcfg[i]->bb->num_inst; j++) {
			isa_name = isa[inst->op_enum].name;
			switch (type) {
			case MUST:
				categorize_l2_inst_X_AH_NC(tcfg[i], inst, j);
				break;
			case MAY:
				categorize_l2_inst_X_AM_NC(tcfg[i], inst, j);
				break;
			case PERSISTENCE:
				categorize_l2_inst_X_PS_NC(tcfg[i], inst, j);
				/* set the latency for instruction cache miss */
				/* if (inst->l2_inst_access == NOT_CLASSIFIED && (inst->inst_access == NOT_CLASSIFIED ||
				 inst->inst_access == ALL_MISS))
				 tcfg[i]->inst_cache_delay += L2_MISS_PENALTY; */
				/* Increment number of persistence memory blocks in L2 instruction cache */
				/* else if (inst->l2_inst_access == PS || inst->inst_access == PS)
				 tcfg[i]->n_l2_persistence += 1;	*/
				break;
			default:
				prerr("Unknown analysis type");
			};

			inst++;
		}
	}
}

static void categorize_inst_fifo_pers(tcfg_node_t* bbi, de_inst_t* inst, int inst_id) {
	int dbg = 0;
	int set = GET_SET(GET_MEM(inst->addr));
	int c;
	int i;
	int h1, hc1;
	int pers_ratio = 0;
	int conflicts;
	mem_blk_set_t temp;
	
	temp.block = GET_MEM(inst->addr);
	temp.next = NULL;

	c = 0;
	for (i = 0; i <= CACHE_SET_SIZE; i++) {
		if (bbi->acs_in[set] && bbi->acs_in[set][i])
			c += getCardinality(bbi->acs_in[set][i]->mem_blk_h);
	}

	if (inst_chmc_l1[bbi->id][inst_id] == NOT_CLASSIFIED) {
		h1 = checkForPresence(bbi->acs_in, &temp);
		hc1 = checkForConflicts(bbi, inst, h1, 1, &conflicts);
/*
		if ((h1 == -1) && (c < CACHE_SET_SIZE)) {
		 printf("accessed bb %u, inst %u, block %u set %u\n", bbi->id, inst_id, temp.block, set);
		  printAcs(bbi->acs_in);
		  abort();
		} */
		if (h1 != -1) {
		  if (c < CACHE_SET_SIZE) { /*
		  	if (!hc1) {
		  		printf("accessed bb %u, inst %u (%x), block %u set %u\n", bbi->id, inst_id, inst->addr, temp.block, set);
		  		printAcs(bbi->acs_in);
		  		abort();
		  	}
		  	*/
		  	inst_chmc_l1[bbi->id][inst_id] = PS;
		  	pers_ratio = 0;
		  } else if (hc1) {
		  	if (conflicts == 0) {
/*
		  printf("accessed bb %u, inst %u, block %u set %u\n", bbi->id, inst_id, temp.block, set);
		  printAcs(bbi->acs_in);
		  printf("conflicts: %d\n", conflicts);
		  abort();
		*/
		  	} else {
		  	pers_ratio = (CACHE_SET_SIZE % conflicts) ? ((CACHE_SET_SIZE/conflicts)+1) : (CACHE_SET_SIZE/conflicts);
		  	inst_chmc_l1[bbi->id][inst_id] = PS;
		  	assert(h1 != 0);
		  	}
		  }
		}
	}
	
	/* debugging purpose only */
	if (dbg) {
		printf("inst_chmc_l1[%i][%i]: ", bbi->id, inst_id);
		switch(inst_chmc_l1[bbi->id][inst_id]) {
		case ALL_HIT:
			printf("alwayshit!\n");
			break;
		case ALL_MISS:
			printf("alwaysmiss!\n");
			break;
		case PS:
			printf("persistent! ( %u ) c=%d h1=%d hc1=%d conflicts=%d\n", pers_ratio, c, h1, hc1, conflicts);
			break;
		case NOT_CLASSIFIED:
			printf("notclassified!\n");
			break;
		}
	}

	/* Rest are already L1 classified or kept non-classified */
}

static void categorize_fifo_pers() {
	tcfg_node_t* bbi;
	de_inst_t* inst;
	int i, j;

	for (i = 0; i < num_tcfg_nodes; i++) {
		inst = tcfg[i]->bb->code;
		for (j = 0; j < tcfg[i]->bb->num_inst; j++) {
			categorize_inst_fifo_pers(tcfg[i], inst, j);
			inst++;
		}
	}
}

/* Categorize instruction access patterns. For must analysis :: all hits 
 * or non-classified */
static void categorize_L1(int type) {
	tcfg_node_t* bbi;
	de_inst_t* inst;
	char* isa_name;
	int i, j;

	for (i = 0; i < num_tcfg_nodes; i++) {
		inst = tcfg[i]->bb->code;

		for (j = 0; j < tcfg[i]->bb->num_inst; j++) {
			isa_name = isa[inst->op_enum].name;

			switch (type) {
			case MUST:
				categorize_inst_AH_NC(tcfg[i], inst, j);
				break;
			case MAY:
				categorize_inst_AM_NC(tcfg[i], inst, j);
				/* if(inst->inst_access == ALL_MISS)
				 tcfg[i]->inst_cache_delay += L1_MISS_PENALTY; */
				break;
			case PERSISTENCE:
				categorize_inst_PS_NC(tcfg[i], inst, j);
				/* set the latency for instruction cache miss */
				/* if(inst->inst_access == NOT_CLASSIFIED)
				 tcfg[i]->inst_cache_delay += L1_MISS_PENALTY; */
				/* Increase the no. of persistence instructions by one */
				/* else if(inst->inst_access == PS)
				 tcfg[i]->n_persistence += 1; */
				break;
			default:
				prerr("Unknown analysis type");
			};

			inst++;
		}

	}
}

/* clean up the cache, this section is called before starting 
 * a new cache analysis */
static void cleanupCache() {
	int i, j, k;
	de_inst_t* inst;
	de_inst_t* prev_inst;

	for (i = 0; i < num_tcfg_nodes; i++) {
		if (tcfg[i] == NULL)
			continue;
#if 0
		inst = tcfg[i]->bb->code;
		prev_inst = NULL;
#endif

#ifdef MEM_FREE
		freeCacheState(tcfg[i]->acs_in);
		if (tcfg[i]->out != NULL)
			freeCacheState(tcfg[i]->acs_out);
#endif

#if 0
		for(j = 0; j < tcfg[i]->bb->num_inst; j++)
		{
#ifdef MEM_FREE
			freeCacheState(inst->acs_out);
#endif
			inst->acs_out = NULL;
			/* CAUTION :::: since the input abstract cache state of previous 
			 * instruction and the output abstract cache state of this 
			 * instruction are connected, there is no need to do a double 
			 * memory free */
			if (prev_inst)
			prev_inst->acs_in = NULL;
			prev_inst = inst;
			inst++;
		}
#endif
	}
}

/* Analysis of l2 instruction cache */
static void analyze_abs_l2_instr_cache(int type) {
	int change_flag = 1;
	tcfg_edge_t* edge;
	int i, j, topoidx;
	int temp = 0;
	int iter = 0;

	tcfg_node_t* bbi = tcfg[0];

	/* Initialize the abstract cache at the entry of the 
	 * program */
	for (i = 0; i < num_tcfg_nodes; i++) {
		if (!tcfg[i])
			continue;
		initializeCache(tcfg[i]);
	}

	while (change_flag) {
		change_flag = 0;

		for (i = 0; i < num_tcfg_nodes; i++) {
			/* sudiptac: get the topological index */
			topoidx = topo_tcfg[i];

			/* Join cache states from predecessor basic 
			 * blocks */
			for (edge = tcfg[topoidx]->in; edge; edge = edge->next_in) {
				JoinCacheState(edge->src, tcfg[topoidx], type, NULL);
			}
			transforml2InstCacheState(tcfg[topoidx], &change_flag, type, iter);
		}
		iter++;
	}

	/* Dump the cache state at end of each basic block */
#ifdef _DEBUG_CRPD
	dumpInstCache();
	printf("L2 Iteration count = %d\n", iter);
#endif

	/* CHMC classification of instructions after analysis */
	categorize_L2(type);

	/* Clean up the cache before the analysis starts */
	cleanupCache();
}

int fifo_maximal_concrete_substack(acs_p *acs, int p) {
  int n;
  int result = 0;
  int num_blocks = 0;
  if (acs == NULL)
    return 0;
  for (n = 1; n <= p; n++) {
    if (acs[n-1] != NULL)
      num_blocks += getCardinality(acs[n-1]->mem_blk_h);
    if (num_blocks == n)
      result = n;
  }
  return result;
}


void fifo_copy_set(fifo_must_t *src, fifo_must_t *dst) {
  int j;
  for (j = 0; j < CACHE_SET_SIZE; j++) {
    dst->pp[j] = src->pp[j];
    dst->pc[j] = src->pc[j];
  }
}
void fifo_copy(fifo_must_t *src, fifo_must_t *dst) {
  int i, j;
  for (i = 0; i < MAX_CACHE_SET; i++) {
    for (j = 0; j < CACHE_SET_SIZE; j++) {
      dst[i].pp[j] = src[i].pp[j];
      dst[i].pc[j] = src[i].pc[j];
    }
  }
}


/* bb->fifo_in = bb->fifo_in U pred->fifo_out */
void fifo_join(tcfg_node_t *pred, tcfg_node_t *bb, int type, instdamage_p dmg) {
  int tmp_pc, tmp_pp, n, i, num_blocks, set;
  fifo_must_t *out;
  
  if (dmg == NULL) {
          out = pred->fifo_out;
          } else {
        	  out = alloc_fifo_must(CACHE_SET_SIZE, MAX_CACHE_SET);
          fifo_copy(pred->fifo_out, out);
          apply_instdamage(dmg, out, pred->acs_out);
        }
  if (bb->fifo_in->bottom) {
    fifo_copy(out, bb->fifo_in);
    D_CACHE(printf("copie!\n"));
    return;
  }
  
 
 
 
  for (set = 0; set < MAX_CACHE_SET; set++) {
    for (n = 0; n < CACHE_SET_SIZE; n++) {
      /* Count blocks in the N youngest ages */
      num_blocks = 0;
      for (i = 0; i <= n; i++) {
        if ((bb->acs_in[set] != NULL) && (bb->acs_in[set][i] != NULL))
          num_blocks += getCardinality(bb->acs_in[set][i]->mem_blk_h);
      }
      
      if (num_blocks == (n + 1)) {
        /* same phase-blocks ? then merge the two sequences, taking the minimum number of phase iteration */
        tmp_pc = (out[set].pc[n] > bb->fifo_in[set].pc[n]) ? bb->fifo_in[set].pc[n] : out[set].pc[n];
      } else {
        /* not same phase-block, too bad we have to restart to phase count 0 */
        tmp_pc = 0;
      }
      /* Compute new phase-progress based on which branch we used for the phase-count value */
      if (out[set].pc[n] < bb->fifo_in[set].pc[n]) {
        tmp_pp = out[set].pp[n];
      } else if (out[set].pc[n] > bb->fifo_in[set].pc[n]) {
        tmp_pp = bb->fifo_in[set].pp[n];
      } else { /* equal */
 	tmp_pp = (out[set].pp[n] > bb->fifo_in[set].pp[n]) ? bb->fifo_in[set].pp[n] : out[set].pp[n];       
      }

      /* If we are starting a new sequence (num_blocks != n) then make phase progress as bigger as possible, else limit to phase progress value computed
       * above using the phase progress of the two joined branches */
      tmp_pp = fifo_maximal_concrete_substack(bb->acs_in[set], (num_blocks == (n+1)) ? tmp_pp : (n+1));
      
      bb->fifo_in[set].pp[n] = tmp_pp;
      bb->fifo_in[set].pc[n] = tmp_pp;
      
    }
  }
}


int fifo_is_equal(fifo_must_t *a, fifo_must_t *b) {
    int j;
    for (j = 0; j < CACHE_SET_SIZE; j++) {
      if (a->pp[j] != b->pp[j])
        return 0;
      if (a->pc[j] != b->pc[j])
        return 0;
    }
    return 1;
}

void fifo_update_singleton(fifo_must_t *fifo, acs_p *acs, int block) {
  int n;
  D_CACHE(printf("update fifo w/addr %x\n", fifo));
  for (n = 0; n < CACHE_SET_SIZE; n++) {
    D_CACHE(printf("before: pp[%u]=%u pc[%u]=%u\n", n, fifo->pp[n], n, fifo->pc[n]));
    if (acs && isBeforeAge(acs, block, fifo->pp[n])) {
    	/* Block already acceded in current phase, no phase progress */
    	/* Do nothing */   
    	D_CACHE(printf("do nothing\n"));
    } else if (acs && isBeforeAge(acs, block, n+1)) {
    	/* Blocks has been accessed in current sequence, but not yet accessed in current phase : phase progress by 1 */
    	D_CACHE(printf("phase progess before: %u\n", fifo->pp[n]));
    	fifo->pp[n]++;
    	D_CACHE(printf("increase current phase (%u)\n", fifo->pp[n]));
    } else {
        /* Block have not been accessed in this sequence, give up on this sequence and restart with a new one */
        D_CACHE(printSet(acs));
        fifo->pp[n] = fifo_maximal_concrete_substack(acs, n + 1);
        if ((fifo->pp[n] < (n+1)) && (!isBeforeAge(acs, block, fifo->pp[n]))) {
          /* attempt to take into account current block */
          fifo->pp[n] ++;
        }
        fifo->pc[n] = 0;
        D_CACHE(printf("reset phase, mcs(%u) = %u\n", n+1, fifo->pp[n]));
    }    
    if (fifo->pp[n] == (n+1)) {
    	D_CACHE(printf("phase finished.\n"));
    	/* Current phase finished, move to next phase */
    	fifo->pp[n] = 0;
    	fifo->pc[n] ++;
    	if (fifo->pc[n] > (n+1)) {
    	  /* It's useless to have a phase count greater than N */
    	  fifo->pc[n] = (n+1);
    	}
    }
    D_CACHE(printf("after: pp[%u]=%u pc[%u]=%u\n", n, fifo->pp[n], n, fifo->pc[n]));
  }
}


void fifo_update(tcfg_node_t *node, int *change_flag, int type, int iter) {
  int n_inst;
  de_inst_t *inst;
  int block1, block2, i;
  fifo_must_t *tmp;
  D_CACHE(printf("update le block %u\n", node->id));
  tmp=alloc_fifo_must(CACHE_SET_SIZE, MAX_CACHE_SET);
  fifo_copy(node->fifo_out,tmp);
  fifo_copy(node->fifo_in, node->fifo_out);
  
  for (n_inst = 0; n_inst < node->bb->num_inst; n_inst++) {
    inst = node->bb->code + n_inst;
    D_CACHE(printf("===\n"));
    D_CACHE(printf("0x%x: %s; ", inst->addr, isa[inst->op_enum].name));
    block1 = GET_MEM(inst->addr);
    block2 = GET_MEM(inst->addr + SIZE_OF_WORD - 1);
    D_CACHE(printf("block1 %u block2 %u; ", block1, block2));
    D_CACHE(printf("set1 %u set2 %u\n", GET_SET(block1), GET_SET(block2)));

    D_CACHE(printf("ACS ADDR: %x\n", node->must_inst[n_inst][GET_SET(block1)]));
    D_CACHE(printSet(node->must_inst[n_inst][GET_SET(block1)])); 
    fifo_copy(node->fifo_out, node->fifo_inst[n_inst]);
    fifo_update_singleton(&node->fifo_out[GET_SET(block1)], node->must_inst[n_inst][GET_SET(block1)], block1);
    if (block1 != block2) {
 /*
      fifo_copy(&node->fifo_in[GET_SET(block1)], &node->fifo_out[GET_SET(block1)]);
    fifo_update_singleton(&node->fifo_out[GET_SET(block2)], node->must_inst[n_inst][GET_SET(block2)], block2);  */ 
    printf("not implented yet\n");
    abort();
    }
 }
  for (i = 0; i < MAX_CACHE_SET;i++) {
    *change_flag |= (!fifo_is_equal(&(node->fifo_out[i]), &(tmp[i])));
  }
  
  free_fifo_must(tmp,MAX_CACHE_SET);
  
}


int categorize_fifo(de_inst_t *inst, acs_p **acs, fifo_must_t *fifo) {
  int dbg = 0;
  int block = GET_MEM(inst->addr);
  int set = GET_SET(block);
  int hit = 0;
  int n;
  int i;
/*  printAcs(acs); */
  
//  printf("pc[0] = %u, Cache Set = ", fifo[set].pc[0]);
//  printSet(acs[set]);
  D_CACHE(printf("0x%x: %s; [set=%u/block=%u] ", inst->addr, isa[inst->op_enum].name, set,block));

  for (i = 0; i < CACHE_SET_SIZE; i++) {
    if ((fifo[set].pc[i] == (i+1)) && isBeforeAge(acs[set],block, i + 1)) {
      hit = 1+i;
      break;
    }
  }

  if (dbg) printf("%s(%u)\n\n", (hit ? "HIT" : "NC"), hit);

  return(hit);
}

/*

static void count_component_blocks() { 
	mem_blk_set_t **blocks;
	int i;
	
	blocks = (void*) malloc(sizeof(void*) * MAX_CACHE_SET);
	for (i = 0; i < MAX_CACHE_SET; i++) {
		blocks[i] = NULL;
	}
	for (i = 0; i < num_tcfg_nodes; i++) {
		for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
		de_inst_t *inst = tcfg[i]->bb->code + n_inst;
		int block = GET_MEM(inst->addr);
		int set = GET_SET(block);
		mem_blk_set_t *tmp = (void*) malloc(sizeof(mem_blk_set_t));
		tmp->block = block;
		tmp->next = blocks[set];
		blocks[set] = tmp;
	}
}
*/

// Compute max instruction cache ageing over all paths

static void lru_damage_join(tcfg_node_t *pred, tcfg_node_t *node) {
        int i;
        for (i = 0; i < MAX_CACHE_SET; i++) {
                if (node->mustdamage_in[i] < pred->mustdamage_out[i])
                        node->mustdamage_in[i] = pred->mustdamage_out[i];
        }
}


static void lru_damage_update(tcfg_node_t *node, int *change_flag, int iter) {
  int n_inst;
  de_inst_t *inst;
  int block1, block2, i;
  int *tmp = malloc(sizeof(int) * MAX_CACHE_SET);
  int set;
  

  for (set = 0; set < MAX_CACHE_SET; set++) {
          tmp[set] = node->mustdamage_in[set];
  }
  for (n_inst = 0; n_inst < node->bb->num_inst; n_inst++) {
    inst = node->bb->code + n_inst;
    block1 = GET_MEM(inst->addr);
    block2 = GET_MEM(inst->addr + SIZE_OF_WORD - 1);
    set = GET_SET(block1);
    assert(set < MAX_CACHE_SET);
    
    if (block1 != block2) {
            printf("not implemented yet\n");
            abort();
    }
    //increase age only if the accessed block is not in the must (ageing)
    if (!isBeforeAge(node->must_inst[n_inst][set], block1, CACHE_SET_SIZE)) {
            tmp[set] ++;
            if (node->mustdamage_out[set] < tmp[set])
                    node->mustdamage_out[set] = tmp[set];
    }
  }
  free(tmp);

}

static void analyze_lru_damage(void) {
  int change_flag = 1;
  int i, topoidx, n_inst, iter = 0;
  tcfg_edge_t *edge;
  for (i = 0; i < num_tcfg_nodes; i++) {
	  if (tcfg[i] == NULL)
		  continue;
	tcfg[i]->mustdamage_out = malloc(sizeof(int)*MAX_CACHE_SET);
	memset(tcfg[i]->mustdamage_out, 0, sizeof(int)*MAX_CACHE_SET);
	tcfg[i]->mustdamage_in = malloc(sizeof(int)*MAX_CACHE_SET);
	memset(tcfg[i]->mustdamage_in, 0, sizeof(int)*MAX_CACHE_SET);
  }

  while (change_flag) {
        change_flag = 0;

        for (i = 0; i < num_tcfg_nodes; i++) {
                topoidx = topo_tcfg[i];
                if (tcfg[topoidx] == NULL)
                        continue;

                for (edge = tcfg[topoidx]->in; edge; edge = edge->next_in) {
				lru_damage_join(edge->src, tcfg[topoidx]);
                }
			lru_damage_update(tcfg[topoidx], &change_flag, iter);
        }
		iter++;
  }
}

/* 
 * Perform phase analysis for FIFO caches
 * as described by: 
 * D. Grund, J. Reineke, "Precise and efficient FIFO-replacement analysis based on static phase detection", in ECRTS 2010
 */
static void analyze_phase_fifo(int type) {
  int change_flag = 1;
  int i, topoidx, n_inst, iter = 0;
  tcfg_edge_t *edge;
  de_inst_t *inst;
  FILE *cfg;
  cfg = fopen("tcfg.map", "w");
  dump_tcfg(cfg);
  fclose(cfg);
  //printf("CACHE: %u sets, %u ways\n", MAX_CACHE_SET, CACHE_SET_SIZE);
  /* Allocate acs for BB and instructions */
  for (i = 0; i < num_tcfg_nodes; i++) {
	  if (tcfg[i] == NULL)
		  continue;
	tcfg[i]->fifo_in = alloc_fifo_must(CACHE_SET_SIZE, MAX_CACHE_SET);
	tcfg[i]->fifo_out = alloc_fifo_must(CACHE_SET_SIZE, MAX_CACHE_SET); 
	tcfg[i]->fifo_inst = malloc(sizeof(void*) * tcfg[i]->bb->num_inst);
	CHECK_MEM(tcfg[i]->fifo_inst);
	for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
	  tcfg[i]->fifo_inst[n_inst] = alloc_fifo_must(CACHE_SET_SIZE, MAX_CACHE_SET); 
	}
  }

  /* Abs. Int. processing */
  //printf("\nassociativity: %d\n", CACHE_SET_SIZE);
	while (change_flag) {
		change_flag = 0;

		for (i = 0; i < num_tcfg_nodes; i++) {
			topoidx = topo_tcfg[i];
			if (tcfg[topoidx] == NULL)
				continue;
#ifdef _DEBUG_CRPD
			fprintf(stdout, "(FIFO phase) computing cache states of bbi %d (%d.%d)\n", topoidx, tcfg[topoidx]->bb->proc->id, \
					tcfg[topoidx]->bb->id);
#endif
			/* Join cache states from predecessor basic 
			 * blocks */
			for (edge = tcfg[topoidx]->in; edge; edge = edge->next_in) {
#ifdef _DEBUG_CRPD
				fprintf(stdout, "(FIFO phase) join cache states from bbi %d (%d.%d)\n", edge->src->id,
						tcfg[edge->src->id]->bb->proc->id, tcfg[edge->src->id]->bb->id);
#endif
/*
                                if (edge->dmg) {
                                        apply_instdamage(edge->inst_dmg, edge->src->fifo_out, edge->src->acs_out);
                                } */ 
/*				JoinCacheState(edge->src, tcfg[topoidx], type); */
				fifo_join(edge->src, tcfg[topoidx], type, edge->inst_dmg);
			}
			fifo_update(tcfg[topoidx], &change_flag, type, iter);
/*			transformInstCacheState(tcfg[topoidx], &change_flag, type, iter); */
		}
		iter++;
		//fprintf(stdout, "(FIFO phase) iteration count = %d\n", iter);
	}
  //printf("counting blocks in loops for persistence...\n");
  for (i = 0; i < num_tcfg_loops; i++) {
  	int j;
  	loops[i]->blocks = (void*) malloc(sizeof(void*) * MAX_CACHE_SET);
  	loops[i]->numblocks = (void*) malloc(sizeof(int) * MAX_CACHE_SET);
  	for (j = 0; j < MAX_CACHE_SET; j++) {
  		loops[i]->blocks[j] = NULL;
  		loops[i]->numblocks[j] = 0;
  	}
  }
  
  for (i = 0; i < num_tcfg_nodes; i++) {
	  if (tcfg[i] == NULL)
		  continue;
  	loop_t *loop = loop_map[tcfg[i]->bb->id];
  	if (loop) {
	    for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
	    	int set, block;
	    	mem_blk_set_t *oldhead;
	    	mem_blk_set_t item;
	    	inst = tcfg[i]->bb->code + n_inst;
	    	block = GET_MEM(inst->addr);
	    	set = GET_SET(block);
	    	item.next = NULL;
	    	item.block = block;
	    	if (!isResident(loop->blocks[set], &item)) {
	      	    oldhead = loop->blocks[set];
	    	    loop->blocks[set] = (void*) malloc(sizeof(mem_blk_set_t));
	    	    loop->blocks[set]->block = block;
		    loop->blocks[set]->next = oldhead;
		    loop->numblocks[set]++;
		    D_CACHE(printf("CPERS: loop%u (%u/%u) new block node %u inst %u nblocks %u\n", loop->id, set, block, tcfg[i]->bb->id, n_inst, loop->numblocks[set]));
	    	} else {
		    D_CACHE(printf("CPERS: loop%u (%u/%u) existing block node %u inst %u nblocks %u\n", loop->id, set, block , tcfg[i]->bb->id, n_inst, loop->numblocks[set]));
	    	}
	    }	    
  	} else {
		D_CACHE(printf("CPERS: pas dans une boucle\n"));
  	}
  }
  //printf("categorization...\n");
  /* Categorize, then free ACS for BB and instructions */
  for (i = 0; i < num_tcfg_nodes; i++) {
	  if (tcfg[i] == NULL)
		  continue;
  	loop_t *loop = loop_map[tcfg[i]->bb->id];
  	free_fifo_must(tcfg[i]->fifo_in, MAX_CACHE_SET);
  	if (tcfg[i]->out != NULL)
  		free_fifo_must(tcfg[i]->fifo_out, MAX_CACHE_SET);
	//printf("--\n");
	for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
	  inst = tcfg[i]->bb->code + n_inst;
	  //printf("BLK %u, ", i);
	  assert(tcfg[i]->id == i);
	  inst_chmc_l1[tcfg[i]->id][n_inst] = categorize_fifo(inst, tcfg[i]->must_inst[n_inst], tcfg[i]->fifo_inst[n_inst]) ? ALL_HIT : NOT_CLASSIFIED;
	  
	  if (loop && (loop->id != 0) && (loop->numblocks[GET_SET(GET_MEM(inst->addr))] < CACHE_SET_SIZE)) {
	  	inst_chmc_l1[tcfg[i]->id][n_inst] = PS;
	  }
	  
	  free_fifo_must(tcfg[i]->fifo_inst[n_inst], MAX_CACHE_SET);
	}
	free(tcfg[i]->fifo_inst);
  } 
}

/* Analysis of l1 instruction cache */
static void analyze_abs_instr_cache(int type, int backwards, int only_acs) {
	int change_flag = 1;
	tcfg_edge_t* edge;
	int i, j, topoidx;
	int temp = 0;
	int iter = 0;
	int n_inst;

	tcfg_node_t* bbi = tcfg[0];

	
	/* Initialize the abstract cache at the entry of the 
	 * program */
	for (i = 0; i < num_tcfg_nodes; i++) {
		if (!tcfg[i])
			continue;
	  initializeCache(tcfg[i]);
	  /* If we are doing must analysis, prepare storage for MUST acs */
	  if (type == MUST) { 
          	tcfg[i]->must_inst = malloc(sizeof(void*) * tcfg[i]->bb->num_inst);
          	for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
          		tcfg[i]->must_inst[n_inst] = NULL;
          	}
	  }
	  if (type == MAY) { 
          	tcfg[i]->may_inst = malloc(sizeof(void*) * tcfg[i]->bb->num_inst);
          	tcfg[i]->may_inst_out = malloc(sizeof(void*) * tcfg[i]->bb->num_inst);
          	for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
          		tcfg[i]->may_inst[n_inst] = NULL;
          		tcfg[i]->may_inst_out[n_inst] = NULL;
          	}
	  }

        }

	while (change_flag) {
		change_flag = 0;

		for (i = (backwards ? (num_tcfg_nodes - 1) : 0); (backwards ? i : (num_tcfg_nodes - i - 1)) >= 0 ; i = i + (backwards ? -1 : +1)) {
			/* sudiptac: get the topological index */
			topoidx = topo_tcfg[i];

#ifdef _DEBUG_CRPD
			fprintf(stdout, "computing cache states of bbi %d (%d.%d)\n", topoidx, tcfg[topoidx]->bb->proc->id, \
					tcfg[topoidx]->bb->id);
#endif
			/* Join cache states from predecessor basic 
			 * blocks */
			if (tcfg[topoidx] == NULL)
				continue;
			for (edge = (backwards ? tcfg[topoidx]->out : tcfg[topoidx]->in); edge; edge = (backwards ? edge->next_out : edge->next_in)) {
#ifdef _DEBUG_CRPD
				fprintf(stdout, "join cache states from bbi %d (%d.%d)\n", edge->src->id,
						tcfg[edge->src->id]->bb->proc->id, tcfg[edge->src->id]->bb->id);
#endif
				D_CACHE(printf("join edge before block %u\n", topoidx));
				JoinCacheState(backwards ? edge->dst : edge->src, tcfg[topoidx], type, edge->inst_dmg ? edge->inst_dmg->lru_damage : NULL);
			}
			transformInstCacheState(tcfg[topoidx], &change_flag, type, iter);
		}
		iter++;
#ifdef _DEBUG_WCET
		fprintf(stdout, "iteration count = %d\n", iter);
#endif
	}

	/* Dump the cache state at end of each basic block */
#ifdef _DEBUG_CRPD
	dumpInstCache();
	fprintf(stdout, "Iteration count = %d\n", iter);
#endif	

	D_CACHE(printf("lru analysis end\n"));
	if (only_acs == 1)
		return;
	/* LRU May/Must analysis is done at this point, regardless of the actual replacement policy of the cache.
	 * Now, we do additionnal processing in case of FIFO cache. --clement */
	if (repl_policy == 'f') {
		/* FIFO mode */
		if (type == MUST) {
			analyze_phase_fifo(type);
		} else if (type == PERSISTENCE) {
			categorize_fifo_pers();
		}
		/* we don't do FIFO may yet */
	} else {
		/* Other (LRU) mode, categorize for must/may/pers */
	 
		categorize_L1(type);
	}

	/* Clean up the cache before another analysis */
	cleanupCache();
}

/* abstract interpretation based analysis for instruction cache */
void analyze_abs_instr_cache_all() {
	/* set advanced L1 cache features */
	set_cache();
	/* Set L1 cache parameters */
	X = assoc;
	Y = nsets;
	B = bsize;
	/* initialize hit-miss classification */
	initialize_CHMC();

	/* Do must analysis in instruction cache */
	analyze_abs_instr_cache(MUST, 0, 0);
	/* Do may analysis in instruction cache */
	//analyze_abs_instr_cache(MAY, 0, 0);
	/* Do persistence analysis for instruction cache */
	analyze_abs_instr_cache(PERSISTENCE, 0, 0);
	if (enable_il2cache) {
		/* set advanced L2 cache features */
		set_cache_l2();
		/* Set L2 cache parameters */
		X = assoc_l2;
		Y = nsets_l2;
		B = bsize_l2;
		/* Do must analysis in L2 instruction cache */
		analyze_abs_l2_instr_cache(MUST);
		/* Do may analysis in L2 instruction cache */
		analyze_abs_l2_instr_cache(MAY);
		/* Do persistence analysis in L2 instruction cache */
		//analyze_abs_l2_instr_cache(PERSISTENCE);
	}

#if 0
	/* Print categorization of all instructions */
	print_classification();
#endif
}




int * compute_ucb(void) {
	int i, age, set, age2, inst;
	mem_blk_set_t *iter, *iter2;
	acs_p ****acs_forward;
	char old_policy = repl_policy;
	int n_inst;

	int *temp_ucb = malloc(sizeof(int) * MAX_CACHE_SET);
	int *ucb = malloc(sizeof(int) * MAX_CACHE_SET);
	memset(ucb, 0, sizeof(int)*MAX_CACHE_SET);
	
	
	//forward
	repl_policy = 'l';
	analyze_abs_instr_cache(MAY, 0, 1);
	cleanupCache();
	
	acs_forward = malloc(sizeof(acs_p ***) * num_tcfg_nodes);
	for (i = 0; i < num_tcfg_nodes; i++) {
		assert(i == tcfg[i]->id);
		acs_forward[i] = tcfg[i]->may_inst;

		/* cleekee: free unused cache states */
		for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
			freeCacheState(tcfg[i]->may_inst_out[n_inst]);
			tcfg[i]->may_inst_out[n_inst] = NULL;
		}
	}
	// reference acs_in for forward results
	
	//reverse
	analyze_abs_instr_cache(MAY, 1, 1);
	cleanupCache();
	repl_policy = old_policy;
	/* compute ucb */
	for (i = 0; i < num_tcfg_nodes; i++) {
	    for (inst = 0; inst < tcfg[i]->bb->num_inst; inst++) {
			memset(temp_ucb, 0, sizeof(int)*MAX_CACHE_SET);
			for (set = 0; set < MAX_CACHE_SET; set++) {
				for (age = 0; age < CACHE_SET_SIZE; age++) {
					if (!acs_forward[i] || !acs_forward[i][inst] || !acs_forward[i][inst][set] || !acs_forward[i][inst][set][age])
						continue;
					acs_p tmp = acs_forward[i][inst][set][age];
	
					for (iter = tmp->mem_blk_h; iter; iter = iter->next) {
						for (age2 = 0; age2 < CACHE_SET_SIZE; age2++) {
							assert(set == GET_SET(iter->block));
							if (!tcfg[i]->may_inst_out || !tcfg[i]->may_inst_out[inst] || !tcfg[i]->may_inst_out[inst][set] || !tcfg[i]->may_inst_out[inst][set][age2])
								continue;
							acs_p tmp2 = tcfg[i]->may_inst_out[inst][set][age2];
							for (iter2 = tmp2->mem_blk_h; iter2; iter2 = iter2->next) {
								if (iter2->block == iter->block ) {
									if (age2 + age < CACHE_SET_SIZE) {
										/* only add to ucb if it exists a path from 1st access (before preemption)
										 * to 2nd access (after preemption) without the block being evicted */

										temp_ucb[set]++;
									}
									continue;
								}
							}
						}
					}
				}
				if (temp_ucb[set] > ucb[set])
					ucb[set] = temp_ucb[set];
			}

			freeCacheState(acs_forward[i][inst]);
			acs_forward[i][inst] = NULL;
	    }				
/*		for (blocs dans acs_forward[i]) {
			if (isResident(le meme block dans acs_out)) {
				if (les deux ages > CACHE_SET_SIZE) {
					ucb[truc] ++;
				}
			}
		}
*/
		/* acs_forward[i] */
		/* tcfg[i]->acs_out */
	}
	free(temp_ucb);
	free(acs_forward);
	return ucb;
}

bitset_p compute_ecb() {
	bitset_p bs;
	int i, topoidx, lastidx, set, age;
	char old_policy = repl_policy;
	repl_policy = 'l';
	analyze_abs_instr_cache(MAY, 0, 1);
	repl_policy = old_policy;
	for (i = 0; i < num_tcfg_nodes; i++) {
		topoidx = topo_tcfg[i];
		if (tcfg[topoidx] == NULL)
			continue;
		lastidx = topoidx;
	}
	 /* bitset of all cache sets that may be accessed within the task */
	bs = bs_alloc(MAX_CACHE_SET);
	for (set = 0; set < MAX_CACHE_SET; set++)
		for (age = 0; age < CACHE_SET_SIZE; age++) {
			if (tcfg[lastidx]->acs_out[set] == NULL)
				continue;
			acs_p acs = tcfg[lastidx]->acs_out[set][age];
			if (acs && (acs->mem_blk_h != NULL)) {
				bs_add(bs, set);
			}
		}
	return bs;
}




instdamage_p compute_instdamage() {
	int i;
	de_inst_t *inst;
	int n_inst;
	analyze_lru_damage();
	instdamage_p dmg = malloc(sizeof(instdamage_t));
	//look for exit node 
	dmg->must_exit = NULL;
	dmg->blocks = NULL;
	dmg->block_count = 0;
	for (i = 0; i < num_tcfg_nodes; i++) {
                if (tcfg[i]->out == NULL) {
                        assert(dmg->must_exit == NULL);
                        dmg->must_exit = tcfg[i]->acs_out;
                        dmg->fifo_exit = tcfg[i]->fifo_out;
                        dmg->lru_damage = tcfg[i]->mustdamage_out;
                }
                  for (n_inst = 0; n_inst < tcfg[i]->bb->num_inst; n_inst++) {
                          inst = tcfg[i]->bb->code + n_inst;
                          int *block = malloc(sizeof(int));
                          *block = GET_MEM(inst->addr);                          
                          addToWorkList(&dmg->blocks, block);                          
                          dmg->block_count++;
                }
        }
        assert(dmg->must_exit != NULL);
        return dmg;
}

                                                                                        
static void age_acs(acs_p *acs, int value) {
        int i;
        
        // free acs[i] ? 
        for (i = CACHE_SET_SIZE - 1; i > 0; i--) {
                acs[i] = acs[i-1];
        }
        acs[0] = NULL;
}

// Apply LRU-only instruction must-damage
void apply_instdamage_lru(int *dmgmust, acs_p **must) {
        int set;
        int i;
        for (set = 0; set < MAX_CACHE_SET; set++) {
        	if (must[set])
                age_acs(must[set],dmgmust[set]);
        }

}

// Apply FIFO-only instruction must-damage
void apply_instdamage(instdamage_p dmg, fifo_must_t *fifo, acs_p **must) {
        int n;
        int set;
        int included;
        worklist_p blocks;
        for (set = 0; set < MAX_CACHE_SET; set++) {
                for (n = 0; n < CACHE_SET_SIZE; n++) {
                       included = 1;
                       for (blocks = dmg->blocks; blocks; blocks = blocks->next) {
                               int *block = blocks->data;
                               if ((GET_SET(*block) == set) && !isBeforeAge(must[set], *block, CACHE_SET_SIZE))
                                       included = 0;
                       }                       
                       if (included) {
                              fifo[set].pc[n] += dmg->fifo_exit[set].pc[n];
                              fifo[set].pp[n] = dmg->fifo_exit[set].pp[n];
                       } else {
                              fifo[set].pc[n] = 0;
                              fifo[set].pp[n] = 0;
                       }
                }
        }
//        apply_instdamage_lru(dmg->lru_damage, must);
}

int compute_crpd_fifo_with_way(int preempted, int preempting, int L) {
        int fifo_crpd, lru_crpd;
        int K = X;
        
        assert((L > 0) && (L <= K));
        X = L;
        lru_crpd = compute_crpd(preempted, preempting);
        fifo_crpd = lru_crpd * K/(K-L+1) + L;
        X = K;
        return fifo_crpd;
}

int compute_crpd_fifo(int preempted, int preempting) {
        int min_crpd = -1;
        int cur_crpd;
        int start;
        int L;

// Enable this to find the L value giving the best (lower) crpd
//        start = 1;
        start = X;
        
        for (L = start ; L <= X; L++) {
                cur_crpd = compute_crpd_fifo_with_way(preempted, preempting, L);
                if ((min_crpd == -1) || (min_crpd > cur_crpd))
                        min_crpd = cur_crpd;
        }
        return (min_crpd);
}

int compute_crpd(int preempted, int preempting) {
	int i, set;
	int temp;
	int crpd = 0 ;
	int *ucb;
	bitset_p ecb;
	
	if (preempted >= 0)
		set_task(tasks[preempted]);

	ucb = compute_ucb();

	set_task(tasks[preempting]);
	ecb = compute_ecb();
/*	for (i = 0; i < num_tcfg_nodes; i++) {
		if (tcfg[i]->task != preempting)
			continue;
		
		for (set = 0; set < MAX_CACHE_SET; set++) {
			temp = 0;
			if (bs_test(ecb, set)) {
				temp = tcfg[i]->ucb[set];
				if (temp > CACHE_SET_SIZE)
					temp = CACHE_SET_SIZE;
			}
			if (crpd[set] < temp)
				crpd[set] = temp;
			
		}
		
	}
	
*/
	for (set = 0; set < MAX_CACHE_SET; set++) {
		if (bs_test(ecb, set)) {
			/*if (tasks[preempting]->priority==1)
				crpd += 1;
			else*/
				crpd += ucb[set];
		}		
	}

	return crpd;
}
